/*
 *      Routines to handle MIB operation
 *
 *      Authors: David Hsu	<davidhsu@realtek.com.tw>
 *
 *      $Id: apmib.c,v 1.4 2005/05/13 07:53:08 tommy Exp $
 *
 */

// include file
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "apmib.h"
#include "mibtbl.h"
char _WAN_IF_[10];
char _LAN_IF_[10];
// MAC address filtering
typedef struct _filter {
	struct _filter *prev, *next;
	char val[1];
} FILTER_T, *FILTER_Tp;

typedef struct _linkChain {
	FILTER_Tp pUsedList, pFreeList;
	int size, num, usedNum, compareLen,appendNumber;
	char *buf;
} LINKCHAIN_T, *LINKCHAIN_Tp;


// macro to remove a link list entry
#define REMOVE_LINK_LIST(entry) { \
	if ( entry ) { \
		if ( entry->prev ) \
			entry->prev->next = entry->next; \
		if ( entry->next ) \
			entry->next->prev = entry->prev; \
	} \
}

// macro to add a link list entry
#define ADD_LINK_LIST(list, entry) { \
	if ( list == NULL ) { \
		list = entry; \
		list->prev = list->next = entry; \
	} \
	else { \
		entry->prev = list; \
		entry->next = list->next; \
		list->next = entry; \
		entry->next->prev = entry; \
	} \
}

// local routine declaration
static int flash_read(char *buf, int offset, int len);
static int flash_write(char *buf, int offset, int len);
static int init_linkchain(LINKCHAIN_Tp pLinkChain, int size, int num);
static int add_linkchain(LINKCHAIN_Tp pLinkChain, char *val);
static int delete_linkchain(LINKCHAIN_Tp pLinkChain, char *val);
static void delete_all_linkchain(LINKCHAIN_Tp pLinkChain);
static int edit_linkchain(LINKCHAIN_Tp pLinkChain, char *val);
static int moveUp_linkchain(LINKCHAIN_Tp pLinkChain, char *val);
static int moveDown_linkchain(LINKCHAIN_Tp pLinkChain, char *val);
static int get_linkchain(LINKCHAIN_Tp pLinkChain, char *val, int index);
int wlan_idx=0;	// interface index 
int vwlan_idx=0;	// initially set interface index to root

// local & global variable declaration
APMIB_Tp pMib=NULL;
//APMIB_Tp pMibDef;	//removed by Erwin
PARAM_HEADER_T hsHeader, dsHeader, csHeader;
HW_SETTING_Tp pHwSetting;

static LINKCHAIN_T wlanMacChain;
#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_ )
static LINKCHAIN_T wlanMacChain2;
static LINKCHAIN_T wlanMacChain3;
static LINKCHAIN_T wlanMacChain4;
static LINKCHAIN_T wlanMacChain5;
#endif

static LINKCHAIN_T wdsListChain;

//vance 2008.11.20
#if defined(_WIRELESS_SCHEDULE_)
static LINKCHAIN_T wlanSchChain;
#endif

//Felix 20101222
#if defined(_PROFILE_LIST_)
static LINKCHAIN_T profilelistChain;
#endif

// Static DHCP Lease --------------
static LINKCHAIN_T staticDhcpChain;
//---------------------------------

#ifdef _EZVIEW_
static LINKCHAIN_T EZVIEWChain;
#endif

#ifdef _STARCRAFT_
static LINKCHAIN_T STARCRAFTChain;
#endif

#ifdef _QoS_MODE_SELECT_
static LINKCHAIN_T wanQosCONChain;
#endif

#ifdef _VLAN_CONFIG_SUPPORTED_
static LINKCHAIN_T vlanConfigChain;
#endif

#ifdef HOME_GATEWAY
#if defined(_MPPPOE_)
static LINKCHAIN_T portFwChain, ipFilterChain, portFilterChain, macFilterChain, triggerPortChain, virtSvChain, dmzChain, URLBChain, ACPCChain, sroutChain, rserCltChain, rserUsrChain, wan1QosChain, wan2QosChain;
#else
static LINKCHAIN_T portFwChain, ipFilterChain, portFilterChain, macFilterChain, triggerPortChain, virtSvChain, dmzChain, URLBChain, ACPCChain, sroutChain, rserCltChain, rserUsrChain, wan1QosChain;
#endif
#else
static LINKCHAIN_T rserCltChain, rserUsrChain;
#endif
void dump_mem(void *ptr, int size)
{
    int i;
    printf("\n");
    for(i=0; i < size; i++)
	{
        printf("%4.2x", (unsigned )(((unsigned char*)ptr)[i]));
		if((i%16 == 0) && (i!=0))
			printf("\n");
	}
	printf("\n");
}

////////////////////////////////////////////////////////////////////////////////
char *apmib_hwconf(void)
{
	int ver;
	char *buff;
	// Read hw setting
	if ( flash_read((char *)&hsHeader, HW_SETTING_OFFSET, sizeof(hsHeader))==0 ) {
		printf("Read hw setting header failed!, Offset=%x\n", HW_SETTING_OFFSET);
		return NULL;
	}


	if ( sscanf(&hsHeader.signature[TAG_LEN], "%02d", &ver) != 1)
		ver = -1;
#ifdef __TARGET_BOARD__
	if ( memcmp(hsHeader.signature, HW_SETTING_HEADER_TAG, TAG_LEN) || // invalid signatur
		(ver != HW_SETTING_VER) || // version is less than current
		(hsHeader.len < (sizeof(HW_SETTING_T)+1)) ) { // length is less than current
		printf("Invalid hw setting signature or version number [sig=%c%c != sig%c%c, ver=%d != Ver=%d, len=%d < len= %d]!\n", hsHeader.signature[0],
			hsHeader.signature[1],HW_SETTING_HEADER_TAG[0],HW_SETTING_HEADER_TAG[1], ver, HW_SETTING_VER,hsHeader.len,sizeof(HW_SETTING_T)+1);
		return NULL;
	}
//	if (ver > HW_SETTING_VER)
//		printf("HW setting version is greater than current [f:%d, c:%d]!\n", ver, HW_SETTING_VER);
#endif
	buff = calloc(1, hsHeader.len);
	if ( buff == 0 ) {
		printf("Allocate buffer failed!\n");
		return NULL;
	}
	if ( flash_read(buff, HW_SETTING_OFFSET+sizeof(hsHeader), hsHeader.len)==0 ) {
		printf("Read hw setting failed!\n");
		free(buff);
		return NULL;
	}
	if ( !CHECKSUM_OK(buff, hsHeader.len) ) {
		printf("Invalid checksum of hw setting!\n");
		free(buff);
		return NULL;
	}
	return buff;
}

////////////////////////////////////////////////////////////////////////////////
char *apmib_dsconf(void)
{
	int ver;
	char *buff;

	// Read default s/w mib
	if ( flash_read((char *)&dsHeader, DEFAULT_SETTING_OFFSET, sizeof(dsHeader))==0 ) {
		printf("Read default setting header failed!\n");
		return NULL;
	}

	if ( sscanf(&dsHeader.signature[TAG_LEN], "%02d", &ver) != 1)
		ver = -1;

	if ( memcmp(dsHeader.signature, DEFAULT_SETTING_HEADER_TAG, TAG_LEN) || // invalid signatur
		(ver != DEFAULT_SETTING_VER) || // version is less than current
		(dsHeader.len < (sizeof(APMIB_T))) ) { // length is less than current
		printf("Invalid default setting signature or version number [sig=%c%c != sig:%c%c, ver=%d != ver:%d, len=%d , APMIB_T=%d]!\n",
			dsHeader.signature[0], dsHeader.signature[1],DEFAULT_SETTING_HEADER_TAG[0],DEFAULT_SETTING_HEADER_TAG[1], ver,DEFAULT_SETTING_VER, dsHeader.len, sizeof(APMIB_T));
		return NULL;
	}

	if (ver > DEFAULT_SETTING_VER)
		printf("Default setting version is greater than current [f:%d, c:%d]!\n", ver, DEFAULT_SETTING_VER);

	buff = calloc(1, dsHeader.len + 1);
	if ( buff == 0 ) {
		printf("Allocate buffer failed!\n");
		return NULL;
	}

	if ( flash_read(buff, DEFAULT_SETTING_OFFSET+sizeof(dsHeader), dsHeader.len)==0 ) {
		printf("Read default setting failed!\n");
		free(buff);
		return NULL;
	}
	if ( !CHECKSUM_OK(buff, dsHeader.len) ) {
		printf("Invalid checksum of current setting!\n");
		free(buff);
		return NULL;
	}

	return buff;
}

////////////////////////////////////////////////////////////////////////////////
/*char *apmib_csconf(void)		//removed by Erwin
{
	int ver;
	char *buff;

	// Read current s/w mib
	if ( flash_read((char *)&csHeader, CURRENT_SETTING_OFFSET, sizeof(csHeader))==0 ) {
//		printf("Read current setting header failed!\n");
		return NULL;
	}

	if ( sscanf(&csHeader.signature[TAG_LEN], "%02d", &ver) != 1)
		ver = -1;

	if ( memcmp(csHeader.signature, CURRENT_SETTING_HEADER_TAG, TAG_LEN) || // invalid signatur
		(ver != CURRENT_SETTING_VER) || // version is less than current
			(csHeader.len < (sizeof(APMIB_T)+1)) ) { // length is less than current
		printf("Invalid current setting signature or version number [sig=%c%c, ver=%d, len=%d]!\n",
			csHeader.signature[0], csHeader.signature[1], ver, csHeader.len);
		return NULL;
	}

//	if (ver > CURRENT_SETTING_VER)
//		printf("Current setting version is greater than current [f:%d, c:%d]!\n", ver, CURRENT_SETTING_VER);

	buff = calloc(1, csHeader.len);
	if ( buff == 0 ) {
//		printf("Allocate buffer failed!\n");
		return NULL;
	}

	if ( flash_read(buff, CURRENT_SETTING_OFFSET+sizeof(csHeader), csHeader.len)==0 ) {
//		printf("Read current setting failed!\n");
		free(buff);
		return NULL;
	}

	if ( !CHECKSUM_OK(buff, csHeader.len) ) {
//		printf("Invalid checksum of current setting!\n");
		free(buff);
		return NULL;
	}

	return buff;
}
*/
////////////////////////////////////////////////////////////////////////////////
int apmib_init(void)
{
	int i;
	char *buff;
	if ( pMib != NULL )	// has been initialized
		return 1;

	if ((buff=apmib_hwconf()) == NULL){
		return 0;
	}
	pHwSetting = (HW_SETTING_Tp)buff;

	if ((buff=apmib_dsconf()) == NULL) {
		free(pHwSetting);
		return 0;
	}
	pMib = (APMIB_Tp)buff;

	// initialize MAC access control list
	if ( !init_linkchain(&wlanMacChain, sizeof(MACFILTER_T), MAX_WLAN_AC_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}

	for (i=0; i<pMib->acNum; i++) {
		if ( !add_linkchain(&wlanMacChain, (char *)&pMib->acAddrArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	wlanMacChain.compareLen = sizeof(MACFILTER_T) - COMMENT_LEN;


#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_ )

	if ( !init_linkchain(&wlanMacChain2, sizeof(MACFILTER_T), MAX_WLAN_AC_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->acNum2; i++) {
		if ( !add_linkchain(&wlanMacChain2, (char *)&pMib->acAddrArray2[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	wlanMacChain2.compareLen = sizeof(MACFILTER_T) - COMMENT_LEN;

	if ( !init_linkchain(&wlanMacChain3, sizeof(MACFILTER_T), MAX_WLAN_AC_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->acNum3; i++) {
		if ( !add_linkchain(&wlanMacChain3, (char *)&pMib->acAddrArray3[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	wlanMacChain3.compareLen = sizeof(MACFILTER_T) - COMMENT_LEN;

	if ( !init_linkchain(&wlanMacChain4, sizeof(MACFILTER_T), MAX_WLAN_AC_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->acNum4; i++) {
		if ( !add_linkchain(&wlanMacChain4, (char *)&pMib->acAddrArray4[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	wlanMacChain4.compareLen = sizeof(MACFILTER_T) - COMMENT_LEN;

	if ( !init_linkchain(&wlanMacChain5, sizeof(MACFILTER_T), MAX_WLAN_AC_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->acNum5; i++) {
		if ( !add_linkchain(&wlanMacChain5, (char *)&pMib->acAddrArray5[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	wlanMacChain5.compareLen = sizeof(MACFILTER_T) - COMMENT_LEN;
#endif

        if ( !init_linkchain(&wdsListChain, sizeof(WDSList_T), MAX_WDS_NUM)) {
                free(pMib);
                free(pHwSetting);
                return 0;
        }
        for (i=0; i<pMib->WDSNum; i++) {
                if ( !add_linkchain(&wdsListChain, (char *)&pMib->WDSList[i]) ) {
                        free(pMib);
                        free(pHwSetting);
                        return 0;
                }
        }
        wdsListChain.compareLen = sizeof(WDSList_T) - COMMENT_LEN;

//vance 2008.11.20
#ifdef _WIRELESS_SCHEDULE_
	if ( !init_linkchain(&wlanSchChain, sizeof(SCHEDULE_T), MAX_WLAN_SCH_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->WlanSchNum; i++) {
		if ( !add_linkchain(&wlanSchChain, (char *)&pMib->WlanSchArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	wlanSchChain.compareLen = sizeof(SCHEDULE_T) - COMMENT_LEN;
#endif

//Felix 2012222
#ifdef _PROFILE_LIST_
	if ( !init_linkchain(&profilelistChain, sizeof(PROFILELIST_T), MAX_PROFILE_LIST_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->ProfileListNum; i++) {
		if ( !add_linkchain(&profilelistChain, (char *)&pMib->ProfileListArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	profilelistChain.compareLen = sizeof(PROFILELIST_T);
#endif

	// initialize Static DHCP Lease list
	if ( !init_linkchain(&staticDhcpChain, sizeof(SDHCP_T), MAX_STATIC_DHCP_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}

	for (i=0; i<pMib->SDHCPNum; i++) {
		if ( !add_linkchain(&staticDhcpChain, (char *)&pMib->SDHCPArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	staticDhcpChain.compareLen = sizeof(SDHCP_T);

	//----------------------------------

#ifdef _EZVIEW_
	if ( !init_linkchain(&EZVIEWChain, sizeof(EZVIEW_T), MAX_EZVIEW_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->EZVIEWNum; i++) {
		if ( !add_linkchain(&EZVIEWChain, (char *)&pMib->EZVIEWArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	EZVIEWChain.compareLen = sizeof(EZVIEW_T);
#endif

#ifdef _STARCRAFT_
	if ( !init_linkchain(&STARCRAFTChain, sizeof(STARCRAFT_T), MAX_STARCRAFT_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->STARCRAFTNum; i++) {
		if ( !add_linkchain(&STARCRAFTChain, (char *)&pMib->STARCRAFTArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	STARCRAFTChain.compareLen = sizeof(STARCRAFT_T);
#endif


#ifdef _QoS_MODE_SELECT_
		if ( !init_linkchain(&wanQosCONChain, sizeof(QOSCON_T), MAX_QOSCON_NUM)) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
		for (i=0; i<pMib->QosCONNum; i++) {
			if ( !add_linkchain(&wanQosCONChain, (char *)&pMib->QosCONArray[i]) ) {
				free(pMib);
				free(pHwSetting);
				return 0;
			}
		}
		wanQosCONChain.compareLen = sizeof(QOSCON_T);
#endif

#ifdef HOME_GATEWAY
///////////////////////////////////////////////////////////////////////////////////////////
	// initialize port forwarding table
	if ( !init_linkchain(&virtSvChain, sizeof(PORTFW_T), MAX_FILTER_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->vserNum; i++) {
		if ( !add_linkchain(&virtSvChain, (char *)&pMib->vserArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	virtSvChain.compareLen = sizeof(PORTFW_T) - COMMENT_LEN;
///////////////////////////////////////////////////////////////////////////////////////////
	// initialize port forwarding table
	if ( !init_linkchain(&portFwChain, sizeof(PORTFW_T), MAX_FILTER_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->portFwNum; i++) {
		if ( !add_linkchain(&portFwChain, (char *)&pMib->portFwArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	portFwChain.compareLen = sizeof(PORTFW_T) - COMMENT_LEN;
/////////////////////////////////////////////////////////////////////////////////////////////
	// initialize DMZ table
	if ( !init_linkchain(&dmzChain, sizeof(DMZ_T), MAX_FILTER_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->DMZNum; i++) {
		if ( !add_linkchain(&dmzChain, (char *)&pMib->DMZArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	dmzChain.compareLen = sizeof(DMZ_T);
	// initialize Static Routing table
	if ( !init_linkchain(&sroutChain, sizeof(SROUT_T), MAX_FILTER_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->sroutNum; i++) {
		if ( !add_linkchain(&sroutChain, (char *)&pMib->sroutArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	sroutChain.compareLen = sizeof(SROUT_T);
	// initialize URLB table
	if ( !init_linkchain(&URLBChain, sizeof(URLB_T), MAX_FILTER_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->URLBNum; i++) {
		if ( !add_linkchain(&URLBChain, (char *)&pMib->URLBArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	//kyle 07.02.13
	//URLBChain.compareLen = sizeof(URLB_T) - COMMENT_LEN;
	URLBChain.compareLen = sizeof(URLB_T) ;
	// initialize ACPC table
	if ( !init_linkchain(&ACPCChain, sizeof(ACPC_T), MAX_FILTER_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->ACPCNum; i++) {
		if ( !add_linkchain(&ACPCChain, (char *)&pMib->ACPCArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	ACPCChain.compareLen = sizeof(ACPC_T) - COMMENT_LEN;
/////////////////////////////////////////////////////////////////////////////////////////////
	// initialize ip-filter table
//	if ( !init_linkchain(&ipFilterChain, sizeof(IPFILTER_T), MAX_FILTER_NUM)) {
	if ( !init_linkchain(&ipFilterChain, sizeof(IPFILTER_T), 1)) {//Erwin Modify 02.09.2003
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->ipFilterNum; i++) {
		if ( !add_linkchain(&ipFilterChain, (char *)&pMib->ipFilterArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	ipFilterChain.compareLen = sizeof(IPFILTER_T) - COMMENT_LEN;
	// initialize port-filter table
//	if ( !init_linkchain(&portFilterChain, sizeof(PORTFILTER_T), MAX_FILTER_NUM)) {
	if ( !init_linkchain(&portFilterChain, sizeof(PORTFILTER_T), 1)) {//Erwin Modify 09.24
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->portFilterNum; i++) {
		if ( !add_linkchain(&portFilterChain, (char *)&pMib->vserArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	portFilterChain.compareLen = sizeof(PORTFILTER_T) - COMMENT_LEN;

	// initialize mac-filter table
	if ( !init_linkchain(&macFilterChain, sizeof(MACFILTER_T), MAX_FILTER_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->macFilterNum; i++) {
		if ( !add_linkchain(&macFilterChain, (char *)&pMib->macFilterArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	macFilterChain.compareLen = sizeof(MACFILTER_T) - COMMENT_LEN;

	// initialize trigger-port table
	if ( !init_linkchain(&triggerPortChain, sizeof(TRIGGERPORT_T), MAX_FILTER_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->triggerPortNum; i++) {
		if ( !add_linkchain(&triggerPortChain, (char *)&pMib->triggerPortArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	triggerPortChain.compareLen = 5;	// length of trigger port range + proto type
	// initialize QoS of WAN1 table
	if ( !init_linkchain(&wan1QosChain, sizeof(QOS_T), MAX_QOS_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->QosNum[0]; i++) {
		if ( !add_linkchain(&wan1QosChain, (char *)&pMib->QosArray[0][i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	wan1QosChain.compareLen = sizeof(QOS_T) - COMMENT_LEN;

#if defined(_MPPPOE_)
	// initialize QoS of WAN2 table
	if ( !init_linkchain(&wan2QosChain, sizeof(QOS_T), MAX_QOS_NUM)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->QosNum[1]; i++) {
		if ( !add_linkchain(&wan2QosChain, (char *)&pMib->QosArray[1][i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	wan2QosChain.compareLen = sizeof(QOS_T) - COMMENT_LEN;
#endif

/////////////////////////////////////////////////////////////////////////////////////////////
#else	// Radius server
	// initialize Radius client table
	if ( !init_linkchain(&rserCltChain, sizeof(RSER_CLT_T), 16)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->rserCltNum; i++) {
		if ( !add_linkchain(&rserCltChain, (char *)&pMib->rserCltArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	rserCltChain.compareLen = sizeof(RSER_CLT_T);
	// initialize Radius user table
	if ( !init_linkchain(&rserUsrChain, sizeof(RSER_USR_T), 96)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	for (i=0; i<pMib->rserUsrNum; i++) {
		if ( !add_linkchain(&rserUsrChain, (char *)&pMib->rserUsrArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	}
	rserUsrChain.compareLen = sizeof(RSER_USR_T);
#endif
//printf("bbbb 999\n");


#ifdef _VLAN_CONFIG_SUPPORTED_
	if ( !init_linkchain(&vlanConfigChain, sizeof(VLAN_CONFIG_T), MAX_IFACE_VLAN_CONFIG)) {
		free(pMib);
		free(pHwSetting);
		return 0;
	}
	//printf("vlanconfigNum = %d \n" , pMib->VlanConfigNum);
	for (i=0; i<pMib->VlanConfigNum; i++)
	if ( !add_linkchain(&vlanConfigChain, (char *)&pMib->VlanConfigArray[i]) ) {
			free(pMib);
			free(pHwSetting);
			return 0;
		}
	vlanConfigChain.compareLen = sizeof(VLAN_CONFIG_T);

#endif

#ifdef _WISP_
	apmib_get( MIB_OP_MODE, (void *)&i);
	if(i==0)
	{
#endif
		#ifdef _DEFAULT_WAN_IF_
			strcpy(_WAN_IF_,_DEFAULT_WAN_IF_);
		#else
			strcpy(_WAN_IF_,"eth1");
		#endif
#ifdef _WISP_
	}
	else if(i==1)
	{
		strcpy(_WAN_IF_,"apcli0");
	}
	else if(i==2)
	{
		strcpy(_WAN_IF_,"wlan0");
	}
#endif


#ifdef _DEFAULT_LAN_IF_
	strcpy(_LAN_IF_,_DEFAULT_LAN_IF_);
#else
	strcpy(_LAN_IF_,"eth0");
#endif

	return 1;
}


///////////////////////////////////////////////////////////////////////////////
int apmib_reinit(void)
{

	printf("ReInit APMIB\n");
	if (pMib == NULL)	// has not been initialized
		return 0;

	free(pMib);
//	free(pMibDef);		//removed by Erwin
	free(pHwSetting);
	free(wlanMacChain.buf);
#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_ )
	free(wlanMacChain2.buf);
	free(wlanMacChain3.buf);
	free(wlanMacChain4.buf);
	free(wlanMacChain5.buf);
#endif

//vance 2008.11.20
#if defined(_WIRELESS_SCHEDULE_)
	free(wlanSchChain.buf);
#endif

//Felix 20101222
#if defined(_PROFILE_LIST_)
	free(profilelistChain.buf);
#endif

	free(staticDhcpChain.buf); // Static DHCP Lease

#ifdef _EZVIEW_
	free(EZVIEWChain.buf);
#endif

#ifdef _STARCRAFT_
	free(STARCRAFTChain.buf);
#endif

#ifdef _QoS_MODE_SELECT_
	free(wanQosCONChain.buf);
#endif

#ifdef HOME_GATEWAY
	free(virtSvChain.buf);////////////////////////////////////////////////////////
	free(dmzChain.buf);///////////////////////////////////////////////////
	free(URLBChain.buf);///////////////////////////////////////////////////
	free(ACPCChain.buf);///////////////////////////////////////////////////
	free(sroutChain.buf);///////////////////////////////////////////////////
	free(portFwChain.buf);
	free(ipFilterChain.buf);
	free(portFilterChain.buf);
	free(macFilterChain.buf);
	free(triggerPortChain.buf);
#else
	free(rserUsrChain.buf);
	free(rserCltChain.buf);
#endif
#ifdef _VLAN_CONFIG_SUPPORTED_
	free(vlanConfigChain.buf);
#endif

	pMib=NULL;
//	pMibDef=NULL;	//removed by Erwin

	return apmib_init();
}


////////////////////////////////////////////////////////////////////////////////
int apmib_get(int id, void *value)
{
	int i, index;
	void *pMibTbl;
	mib_table_entry_T *pTbl;
	unsigned char ch;
	unsigned short wd;
	unsigned long dwd;


	// search current setting mib table
	for (i=0; mib_table[i].id; i++) {
		if ( mib_table[i].id == id )
			break;
	}

	if ( mib_table[i].id != 0 ) {
		pMibTbl = (void *)pMib;
		pTbl = mib_table;
	}
	else {
		// if not found, search hw setting table
		for (i=0; hwmib_table[i].id; i++) {
			if ( hwmib_table[i].id == id )
				break;
		}
		if (hwmib_table[i].id == 0) // not found
			return 0;
		pMibTbl = (void *)pHwSetting;
		pTbl = hwmib_table;
	}


	switch (pTbl[i].type) {
	case BYTE_T:
//		*((int *)value) =(int)(*((unsigned char *)(((long)pMibTbl) + pTbl[i].offset)));
		memcpy((char *)&ch, ((char *)pMibTbl) + pTbl[i].offset, 1);
		*((int *)value) = (int)ch;
		break;

	case WORD_T:
//		*((int *)value) =(int)(*((unsigned short *)(((long)pMibTbl) + pTbl[i].offset)));
		memcpy((char *)&wd, ((char *)pMibTbl) + pTbl[i].offset, 2);
		*((int *)value) = (int)wd;
		break;

	case STRING_T:
		strcpy( (char *)value, (const char *)(((long)pMibTbl) + pTbl[i].offset) );
		break;

	case BYTE5_T:
		memcpy( (unsigned char *)value, (unsigned char *)(((long)pMibTbl) + pTbl[i].offset), 5);
		break;

	case BYTE6_T:
		memcpy( (unsigned char *)value, (unsigned char *)(((long)pMibTbl) + pTbl[i].offset), 6);
		break;

	case BYTE8_T:
		memcpy( (unsigned char *)value, (unsigned char *)(((long)pMibTbl) + pTbl[i].offset), 8);
		break;

	case BYTE13_T:
		memcpy( (unsigned char *)value, (unsigned char *)(((long)pMibTbl) + pTbl[i].offset), 13);
		break;

	case BYTE16_T:
		memcpy( (unsigned char *)value, (unsigned char *)(((long)pMibTbl) + pTbl[i].offset), 16);
		break;

	case DWORD_T:
		memcpy((char *)&dwd, ((char *)pMibTbl) + pTbl[i].offset, 4);
		*((int *)value) = (int)dwd;
		break;

	case INDEX_T:
		memcpy((char *)&dwd, ((char *)pMibTbl) + pTbl[i].offset, 4);
		*((int *)value) = (int)dwd;
		break;

	case INDEXDOS_T://////////////////////////////////////////////////////////////
		memcpy((char *)&dwd, ((char *)pMibTbl) + pTbl[i].offset, 4);
		*((int *)value) = (int)dwd;
		break;

#ifdef WLAN_WPA
#if 0
	case BYTE32_T:
		memcpy( (unsigned char *)value, (unsigned char *)(((long)pMibTbl) + pTbl[i].offset), 32);
		break;
#endif
#endif

	case BYTE_ARRAY_T:
		memcpy( (unsigned char *)value, (unsigned char *)(((long)pMibTbl) + pTbl[i].offset), pTbl[i].size);
		break;

	case IA_T:
		memcpy( (unsigned char *)value, (unsigned char *)(((long)pMibTbl) + pTbl[i].offset), 4);
		break;

	case WLAC_ARRAY_T:
		//printf("DEBUG0\n");
		index = (int)( *((unsigned char *)value));
		return get_linkchain(&wlanMacChain, (char *)value, index );

#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_ )
	case WLAC_ARRAY2_T:
		index = (int)( *((unsigned char *)value));
		return get_linkchain(&wlanMacChain2, (char *)value, index );
	case WLAC_ARRAY3_T:
		//printf("DEBUG2\n");
		index = (int)( *((unsigned char *)value));
		return get_linkchain(&wlanMacChain3, (char *)value, index );
	case WLAC_ARRAY4_T:
		//printf("DEBUG3\n");
		index = (int)( *((unsigned char *)value));
		return get_linkchain(&wlanMacChain4, (char *)value, index );
	case WLAC_ARRAY5_T:
		//printf("DEBUG3\n");
		index = (int)( *((unsigned char *)value));
		return get_linkchain(&wlanMacChain5, (char *)value, index );
#endif

//vance 2008.11.20
#if defined(_WIRELESS_SCHEDULE_)
	case WLAN_SCH_ARRAY_T:
			//printf("DEBUG0\n");
			index = (int)( *((unsigned char *)value));
			return get_linkchain(&wlanSchChain, (char *)value, index );
#endif

//Felix 21011222
#if defined(_PROFILE_LIST_)
	case PROFILE_LIST_ARRAY_T:
			//printf("DEBUG0\n");
			index = (int)( *((unsigned char *)value));
			return get_linkchain(&profilelistChain, (char *)value, index );
#endif

// Static DHCP Lease ---------------------------------------------------------------
	case SDHCP_ARRAY_T:
		index = (int)( *((unsigned char *)value));
		return get_linkchain(&staticDhcpChain, (char *)value, index );
//-----------------------------------------------------------------------------------

#ifdef _EZVIEW_
	case EZVIEW_ARRAY_T:
                index = (int)( *((unsigned char *)value));
 		return get_linkchain(&EZVIEWChain, (char *)value, index );
#endif

#ifdef _STARCRAFT_
	case STARCRAFT_ARRAY_T:
                index = (int)( *((unsigned char *)value));
 		return get_linkchain(&STARCRAFTChain, (char *)value, index );
#endif

#ifdef _QoS_MODE_SELECT_
	case WAN_QOSCON_ARRAY_T:
	            index = (int)( *((unsigned char *)value));
	   return get_linkchain(&wanQosCONChain, (char *)value, index );
#endif

#ifdef HOME_GATEWAY
/////////////////////////////////////////////////////////////////////////////////////////
	case VSER_ARRAY_T:
                index = (int)( *((unsigned char *)value));
 		return get_linkchain(&virtSvChain, (char *)value, index );
	case PORTFW_ARRAY_T:
                index = (int)( *((unsigned char *)value));
 		return get_linkchain(&portFwChain, (char *)value, index );
	case DMZ_ARRAY_T:
		index = (int)( *((unsigned char *)value));
 		return get_linkchain(&dmzChain, (char *)value, index );
	case URLB_ARRAY_T:
		index = (int)( *((unsigned char *)value));
 		return get_linkchain(&URLBChain, (char *)value, index );
	case ACPC_ARRAY_T:
		index = (int)( *((unsigned char *)value));
 		return get_linkchain(&ACPCChain, (char *)value, index );
	case SROUT_ARRAY_T:
		index = (int)( *((unsigned char *)value));
 		return get_linkchain(&sroutChain, (char *)value, index );
	case IPFILTER_ARRAY_T:
		index = (int)( *((unsigned char *)value));
 		return get_linkchain(&ipFilterChain, (char *)value, index );

	case PORTFILTER_ARRAY_T:
                index = (int)( *((unsigned char *)value));
 		return get_linkchain(&portFilterChain, (char *)value, index );

	case MACFILTER_ARRAY_T:
                index = (int)( *((unsigned char *)value));
 		return get_linkchain(&macFilterChain, (char *)value, index );

	case TRIGGERPORT_ARRAY_T:
                index = (int)( *((unsigned char *)value));
 		return get_linkchain(&triggerPortChain, (char *)value, index );

	case WAN1_QOS_ARRAY_T:
                index = (int)( *((unsigned char *)value));
 		return get_linkchain(&wan1QosChain, (char *)value, index );

#if  defined(_MPPPOE_)
	case WAN2_QOS_ARRAY_T:
                index = (int)( *((unsigned char *)value));
 		return get_linkchain(&wan2QosChain, (char *)value, index );
#endif

///////////////////////////////////////////////////////////////////////////////////////////
#else
	case RSER_USR_ARRAY_T:
		index = (int)( *((unsigned char *)value));
 		return get_linkchain(&rserUsrChain, (char *)value, index );
	case RSER_CLT_ARRAY_T:
		index = (int)( *((unsigned char *)value));
 		return get_linkchain(&rserCltChain, (char *)value, index );
#endif
#ifdef _VLAN_CONFIG_SUPPORTED_
	case VLANCONFIG_ARRAY_T:	

		index = (int)( *((unsigned char *)value));
 		return get_linkchain(&vlanConfigChain, (char *)value, index ); 		
#endif 
/////////////////////////////////////////////////////////////////////////////////////////
        case WDSLIST_T:
                index = (int)( *((unsigned char *)value));
                return get_linkchain(&wdsListChain, (char *)value, index );

	}
	return 1;
}


////////////////////////////////////////////////////////////////////////////////
/*int apmib_getDef(int id, void *value)		//removed by Erwin
{
	int ret;
	APMIB_Tp saveMib=pMib;

	pMib = pMibDef;
	ret = apmib_get(id, value);
	pMib = saveMib;
	return ret;
}
*/

////////////////////////////////////////////////////////////////////////////////
int apmib_set(int id, void *value)
{
	int i, ret;
	void *pMibTbl;
	mib_table_entry_T *pTbl;
	unsigned char ch;
	unsigned short wd;
	unsigned long dwd;
    unsigned char* tmp;
    int max_chan_num;

        if (id == MIB_WDS_LIST_ADD) {
                ret = add_linkchain(&wdsListChain, (char *)value);
                if ( ret )
                        pMib->WDSNum++;

                return ret;
        }
        if (id == MIB_WDS_LIST_DEL) {
                ret = delete_linkchain(&wdsListChain, (char *)value);
                if ( ret )
                        pMib->WDSNum--;

                return ret;
        }
        if (id == MIB_WDS_LIST_DELALL) {
                delete_all_linkchain(&wdsListChain);
                pMib->WDSNum = 0;
                return 1;
        }


	if (id == MIB_WLAN_AC_ADDR_ADD) {
		ret = add_linkchain(&wlanMacChain, (char *)value);
		if ( ret )
			pMib->acNum++;

		return ret;
	}
	if (id == MIB_WLAN_AC_ADDR_DEL) {
		ret = delete_linkchain(&wlanMacChain, (char *)value);
		if ( ret )
			pMib->acNum--;

		return ret;
	}
	if (id == MIB_WLAN_AC_ADDR_DELALL) {
		delete_all_linkchain(&wlanMacChain);
		pMib->acNum = 0;
		return 1;
	}


#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_ )
	if (id == MIB_WLAN_AC2_ADDR_ADD) {
		ret = add_linkchain(&wlanMacChain2, (char *)value);
		if ( ret )
			pMib->acNum2++;

		return ret;
	}
	if (id == MIB_WLAN_AC2_ADDR_DEL) {
		ret = delete_linkchain(&wlanMacChain2, (char *)value);
		if ( ret )
			pMib->acNum2--;

		return ret;
	}
	if (id == MIB_WLAN_AC2_ADDR_DELALL) {
		delete_all_linkchain(&wlanMacChain2);
		pMib->acNum2 = 0;
		return 1;
	}

	if (id == MIB_WLAN_AC3_ADDR_ADD) {
		ret = add_linkchain(&wlanMacChain3, (char *)value);
		if ( ret )
			pMib->acNum3++;

		return ret;
	}
	if (id == MIB_WLAN_AC3_ADDR_DEL) {
		ret = delete_linkchain(&wlanMacChain3, (char *)value);
		if ( ret )
			pMib->acNum3--;

		return ret;
	}
	if (id == MIB_WLAN_AC3_ADDR_DELALL) {
		delete_all_linkchain(&wlanMacChain3);
		pMib->acNum3 = 0;
		return 1;
	}

	if (id == MIB_WLAN_AC4_ADDR_ADD) {
		ret = add_linkchain(&wlanMacChain4, (char *)value);
		if ( ret )
			pMib->acNum4++;

		return ret;
	}
	if (id == MIB_WLAN_AC4_ADDR_DEL) {
		ret = delete_linkchain(&wlanMacChain4, (char *)value);
		if ( ret )
			pMib->acNum4--;

		return ret;
	}
	if (id == MIB_WLAN_AC4_ADDR_DELALL) {
		delete_all_linkchain(&wlanMacChain4);
		pMib->acNum4 = 0;
		return 1;
	}

	if (id == MIB_WLAN_AC5_ADDR_ADD) {
		ret = add_linkchain(&wlanMacChain5, (char *)value);
		if ( ret )
			pMib->acNum5++;

		return ret;
	}
	if (id == MIB_WLAN_AC5_ADDR_DEL) {
		ret = delete_linkchain(&wlanMacChain5, (char *)value);
		if ( ret )
			pMib->acNum5--;

		return ret;
	}
	if (id == MIB_WLAN_AC5_ADDR_DELALL) {
		delete_all_linkchain(&wlanMacChain5);
		pMib->acNum5 = 0;
		return 1;
	}
#endif

//vance 2008.11.20
#if defined(_WIRELESS_SCHEDULE_)
	if (id == MIB_WLAN_SCH_ADD) {
		ret = add_linkchain(&wlanSchChain, (char *)value);
		if ( ret )
			pMib->WlanSchNum++;
		return ret;
	}
	if (id == MIB_WLAN_SCH_DEL) {
		ret = delete_linkchain(&wlanSchChain, (char *)value);
		if ( ret )
			pMib->WlanSchNum--;

		return ret;
	}
	if (id == MIB_WLAN_SCH_DELALL) {
		delete_all_linkchain(&wlanSchChain);
		pMib->WlanSchNum = 0;
		return 1;
	}
#endif

//Felix 21011222
#if defined(_PROFILE_LIST_)
	if (id == MIB_PROFILE_LIST_ADD) {
		ret = add_linkchain(&profilelistChain, (char *)value);
		if ( ret )
			pMib->ProfileListNum++;
		return ret;
	}
	if (id == MIB_PROFILE_LIST_DEL) {
		ret = delete_linkchain(&profilelistChain, (char *)value);
		if ( ret )
			pMib->ProfileListNum--;

		return ret;
	}
	if (id == MIB_PROFILE_LIST_EDIT) {
		ret = edit_linkchain(&profilelistChain, (char *)value);
		return ret;
	}
	if (id == MIB_PROFILE_LIST_DELALL) {
		delete_all_linkchain(&profilelistChain);
		pMib->ProfileListNum = 0;
		return 1;
	}
#endif


// Static DHCP Lease -------------------------------------------
	if (id == MIB_STATIC_DHCP_ADD) {
		ret = add_linkchain(&staticDhcpChain, (char *)value);
		if ( ret )
			pMib->SDHCPNum++;

		return ret;
	}
	if (id == MIB_STATIC_DHCP_DEL) {
		ret = delete_linkchain(&staticDhcpChain, (char *)value);
		if ( ret )
			pMib->SDHCPNum--;

		return ret;
	}
	if (id == MIB_STATIC_DHCP_DELALL) {
		delete_all_linkchain(&staticDhcpChain);
		pMib->SDHCPNum = 0;
		return 1;
	}
// ---------------------------------------------------------------

#ifdef _EZVIEW_
	if (id == MIB_EZVIEW_ADD) {
		ret = add_linkchain(&EZVIEWChain, (char *)value);
		if ( ret )
			pMib->EZVIEWNum++;
		return ret;
	}
	if (id == MIB_EZVIEW_DEL) {
		ret = delete_linkchain(&EZVIEWChain, (char *)value);
		if ( ret )
			pMib->EZVIEWNum--;
		return ret;
	}
	if (id == MIB_EZVIEW_DELALL) {
		delete_all_linkchain(&EZVIEWChain);
		pMib->EZVIEWNum = 0;
		return 1;
	}
#endif

#ifdef _STARCRAFT_
	if (id == MIB_STARCRAFT_ADD) {
		ret = add_linkchain(&STARCRAFTChain, (char *)value);
		if ( ret )
			pMib->STARCRAFTNum++;
		return ret;
	}
	if (id == MIB_STARCRAFT_DEL) {
		ret = delete_linkchain(&STARCRAFTChain, (char *)value);
		if ( ret )
			pMib->STARCRAFTNum--;
		return ret;
	}
	if (id == MIB_STARCRAFT_DELALL) {
		delete_all_linkchain(&STARCRAFTChain);
		pMib->STARCRAFTNum = 0;
		return 1;
	}
#endif

#ifdef _QoS_MODE_SELECT_
	if (id == MIB_WANCON_QOS_ADD) {
			ret = add_linkchain(&wanQosCONChain, (char *)value);
			if ( ret )
				(pMib->QosCONNum)++;
			return ret;
		}
		if (id == MIB_WANCON_QOS_EDIT) {
			ret = edit_linkchain(&wanQosCONChain, (char *)value);
			return ret;
		}
		if (id == MIB_WANCON_QOS_UP) {
			ret = moveUp_linkchain(&wanQosCONChain, (char *)value);
			return ret;
		}
		if (id == MIB_WANCON_QOS_DOWN) {
			ret = moveDown_linkchain(&wanQosCONChain, (char *)value);
			return ret;
		}
		if (id == MIB_WANCON_QOS_DEL) {
			ret = delete_linkchain(&wanQosCONChain, (char *)value);
			if ( ret )
				pMib->QosCONNum--;
			return ret;
		}
		if (id == MIB_WANCON_QOS_DELALL) {
			delete_all_linkchain(&wanQosCONChain);
			pMib->QosCONNum = 0;
			return 1;
		}
#endif


#ifdef _VLAN_CONFIG_SUPPORTED_
	if (id == MIB_VLANCONFIG_ADD1) {		//Default VLAN 
		int entryNum=0,i;
		VLAN_CONFIG_T entry;
		VLAN_CONFIG_T entry_new2;
		VLAN_CONFIG_Tp entry_new;
		entry_new = (VLAN_CONFIG_Tp)value;
		apmib_get(MIB_VLANCONFIG_NUM, (void *)&entryNum);
		
		/*for (i=1; i<=entryNum; i++) {
		*((char *)&entry) = (char)i;
			apmib_get(MIB_VLANCONFIG, (void *)&entry);
			memcpy(&entry_new, &entry, sizeof(VLAN_CONFIG_T));
			entry_new2.enabled=0;
			update_linkchain(VLANCONFIG_ARRAY_T, &entry, &entry_new, sizeof(VLAN_CONFIG_T));
		}*/

		for (i=1; i<=entryNum; i++) {
		
		*((char *)&entry) = (char)i;
			apmib_get(MIB_VLANCONFIG, (void *)&entry);
			if(i==1)
			{
			strcpy(entry_new->netIface, "eth0");
			entry_new->enabled = 0;
			entry_new->tagged = 0;
			entry_new->priority=7;
			entry_new->cfi=1;
			entry_new->vlanId=3022;			//note.   only Customer default Vid value=0 , if realtek vid default value = 1
				update_linkchain(VLANCONFIG_ARRAY_T, &entry, entry_new, sizeof(VLAN_CONFIG_T));
			}
			if(i==2)
			{
			strcpy(entry_new->netIface, "eth2");
			entry_new->enabled = 0;
			entry_new->tagged = 0;
			entry_new->priority=0;
			entry_new->cfi=1;
			entry_new->vlanId=3030;
				update_linkchain(VLANCONFIG_ARRAY_T, &entry, entry_new, sizeof(VLAN_CONFIG_T));
			}
			if(i==3)
			{
			strcpy(entry_new->netIface, "eth3");
			entry_new->enabled = 0;
			entry_new->tagged = 0;
			entry_new->priority=3;
			entry_new->cfi=1;
			entry_new->vlanId=500;
				update_linkchain(VLANCONFIG_ARRAY_T, &entry, entry_new, sizeof(VLAN_CONFIG_T));
			}
			if(i==4)
			{
			strcpy(entry_new->netIface, "eth4");
			entry_new->enabled = 0;
			entry_new->tagged = 0;
			entry_new->priority=0;
			entry_new->cfi=1;
			entry_new->vlanId=1;
				update_linkchain(VLANCONFIG_ARRAY_T, &entry, entry_new, sizeof(VLAN_CONFIG_T));
			}
			if(i==5)
			{
			strcpy(entry_new->netIface, "wlan0");
			entry_new->enabled = 0;
			entry_new->tagged = 0;
			entry_new->priority=0;
			entry_new->cfi=1;
			entry_new->vlanId=1;
				update_linkchain(VLANCONFIG_ARRAY_T, &entry, entry_new, sizeof(VLAN_CONFIG_T));
			}
			if(i==6)
			{
			strcpy(entry_new->netIface, "wlan0-va0");
			entry_new->enabled = 0;
			entry_new->tagged = 0;
			entry_new->priority=0;
			entry_new->cfi=1;
			entry_new->vlanId=1;
				update_linkchain(VLANCONFIG_ARRAY_T, &entry, entry_new, sizeof(VLAN_CONFIG_T));
			}
			if(i==7)
			{
			strcpy(entry_new->netIface, "wlan0-va1");
			entry_new->enabled = 0;
			entry_new->tagged = 0;
			entry_new->priority=0;
			entry_new->cfi=1;
			entry_new->vlanId=1;
				update_linkchain(VLANCONFIG_ARRAY_T, &entry, entry_new, sizeof(VLAN_CONFIG_T));
			}
			if(i==8)
			{
			strcpy(entry_new->netIface, "wlan0-va2");
			entry_new->enabled = 0;
			entry_new->tagged = 0;
			entry_new->priority=0;
			entry_new->cfi=1;
			entry_new->vlanId=1;
				update_linkchain(VLANCONFIG_ARRAY_T, &entry, entry_new, sizeof(VLAN_CONFIG_T));
			}
			if(i==9)
			{
			strcpy(entry_new->netIface, "wlan0-va3");
			entry_new->enabled = 0;
			entry_new->tagged = 0;
			entry_new->priority=0;
			entry_new->cfi=1;
			entry_new->vlanId=1;
				update_linkchain(VLANCONFIG_ARRAY_T, &entry, entry_new, sizeof(VLAN_CONFIG_T));
			}
			if(i==10)
			{
			strcpy(entry_new->netIface, "eth1");
			entry_new->enabled = 0;
			entry_new->tagged = 0;
			entry_new->priority=0;
			entry_new->cfi=1;
			entry_new->vlanId=1;
				update_linkchain(VLANCONFIG_ARRAY_T, &entry, entry_new, sizeof(VLAN_CONFIG_T));
			}
		}
		return 1;
	}
	
	if (id == MIB_VLANCONFIG_ADD || id == MIB_VLANCONFIG_DEL) {
		int entryNum=0,i;
		VLAN_CONFIG_T entry;
		VLAN_CONFIG_Tp entry_new;
		entry_new = (VLAN_CONFIG_Tp)value;
		apmib_get(MIB_VLANCONFIG_NUM, (void *)&entryNum);
		
		//printf("entry_new->netIface = %s\n" , entry_new->netIface);
		for (i=1; i<=entryNum; i++) {
		*((char *)&entry) = (char)i;
			apmib_get(MIB_VLANCONFIG, (void *)&entry);
			//printf("entry.netIface = %s\n" , entry.netIface);
			if(!strcmp(entry.netIface, entry_new->netIface)){
				update_linkchain(VLANCONFIG_ARRAY_T, &entry, entry_new, sizeof(VLAN_CONFIG_T));
				break;
			}
		}
		return 1;
	}

	if (id == MIB_VLANCONFIG_DELALL) {
		int entryNum=0,i;
		VLAN_CONFIG_T entry;
		VLAN_CONFIG_T entry_new;
		apmib_get(MIB_VLANCONFIG_NUM, (void *)&entryNum);

		for (i=1; i<=entryNum; i++) {
		*((char *)&entry) = (char)i;
			apmib_get(MIB_VLANCONFIG, (void *)&entry);
			memcpy(&entry_new, &entry, sizeof(VLAN_CONFIG_T));
			entry_new.enabled=0;
			update_linkchain(VLANCONFIG_ARRAY_T, &entry, &entry_new, sizeof(VLAN_CONFIG_T));
		}
		

		return 1;
	}

#endif


#ifdef HOME_GATEWAY
//////////////////////////////////////////////////////////////////////////////////////////////
	if (id == MIB_VSER_ADD) {
		ret = add_linkchain(&virtSvChain, (char *)value);
		if ( ret )
			pMib->vserNum++;
		return ret;
	}
	if (id == MIB_VSER_DEL) {
		ret = delete_linkchain(&virtSvChain, (char *)value);
		if ( ret )
			pMib->vserNum--;

		return ret;
	}
	if (id == MIB_VSER_DELALL) {
		delete_all_linkchain(&virtSvChain);
		pMib->vserNum = 0;
		return 1;
	}
//////////////////////////////////////////////////////////////////////////////////////////////
	if (id == MIB_PORTFW_ADD) {
		ret = add_linkchain(&portFwChain, (char *)value);
		if ( ret )
			pMib->portFwNum++;

		return ret;
	}
	if (id == MIB_PORTFW_DEL) {
		ret = delete_linkchain(&portFwChain, (char *)value);
		if ( ret )
			pMib->portFwNum--;

		return ret;
	}
	if (id == MIB_PORTFW_DELALL) {
		delete_all_linkchain(&portFwChain);
		pMib->portFwNum = 0;
		return 1;
	}
///////////////////////////////////////////////////////////////////////////////////////////
                                //DMZ
	if (id == MIB_DMZ_ADD) {

		ret = add_linkchain(&dmzChain, (char *)value);
		if ( ret )
			pMib->DMZNum++;

		return ret;
	}
	if (id == MIB_DMZ_DEL) {
		ret = delete_linkchain(&dmzChain, (char *)value);
		if ( ret )
			pMib->DMZNum--;

		return ret;
	}
	if (id == MIB_DMZ_EDIT) {
		ret = edit_linkchain(&dmzChain, (char *)value);

		return ret;
	}
	if (id == MIB_DMZ_DELALL) {
		delete_all_linkchain(&dmzChain);
		pMib->DMZNum = 0;
		return 1;
	}
                                //Static Routing
	if (id == MIB_SROUT_ADD) {
		ret = add_linkchain(&sroutChain, (char *)value);
		if ( ret )
			pMib->sroutNum++;

		return ret;
	}
	if (id == MIB_SROUT_DEL) {
		ret = delete_linkchain(&sroutChain, (char *)value);
		if ( ret )
			pMib->sroutNum--;

		return ret;
	}
	if (id == MIB_SROUT_DELALL) {
		delete_all_linkchain(&sroutChain);
		pMib->sroutNum = 0;
		return 1;
	}
                                //URLB
	if (id == MIB_URLB_ADD) {
		ret = add_linkchain(&URLBChain, (char *)value);
		if ( ret )
			pMib->URLBNum++;
		return ret;
	}
	if (id == MIB_URLB_DEL) {
		ret = delete_linkchain(&URLBChain, (char *)value);
		if ( ret )
			pMib->URLBNum--;
		return ret;
	}
	if (id == MIB_URLB_DELALL) {
		delete_all_linkchain(&URLBChain);
		pMib->URLBNum = 0;
		return 1;
	}
                                //ACPC
	if (id == MIB_ACPC_ADD) {
		ret = add_linkchain(&ACPCChain, (char *)value);
		if ( ret )
			pMib->ACPCNum++;

		return ret;
	}
	if (id == MIB_ACPC_DEL) {
		ret = delete_linkchain(&ACPCChain, (char *)value);
		if ( ret )
			pMib->ACPCNum--;

		return ret;
	}
	if (id == MIB_ACPC_DELALL) {
		delete_all_linkchain(&ACPCChain);
		pMib->ACPCNum = 0;
		return 1;
	}
///////////////////////////////////////////////////////////////////////////////////////////
	if (id == MIB_IPFILTER_ADD) {
		ret = add_linkchain(&ipFilterChain, (char *)value);
		if ( ret )
			pMib->ipFilterNum++;

		return ret;
	}
	if (id == MIB_IPFILTER_DEL) {
		ret = delete_linkchain(&ipFilterChain, (char *)value);
		if ( ret )
			pMib->ipFilterNum--;

		return ret;
	}
	if (id == MIB_IPFILTER_DELALL) {
		delete_all_linkchain(&ipFilterChain);
		pMib->ipFilterNum = 0;
		return 1;
	}

	if (id == MIB_PORTFILTER_ADD) {
		ret = add_linkchain(&portFilterChain, (char *)value);
		if ( ret )
			pMib->portFilterNum++;

		return ret;
	}
	if (id == MIB_PORTFILTER_DEL) {
		ret = delete_linkchain(&portFilterChain, (char *)value);
		if ( ret )
			pMib->portFilterNum--;

		return ret;
	}
	if (id == MIB_PORTFILTER_DELALL) {
		delete_all_linkchain(&portFilterChain);
		pMib->portFilterNum = 0;
		return 1;
	}

	if (id == MIB_MACFILTER_ADD) {
		ret = add_linkchain(&macFilterChain, (char *)value);
		if ( ret )
			pMib->macFilterNum++;

		return ret;
	}
	if (id == MIB_MACFILTER_DEL) {
		ret = delete_linkchain(&macFilterChain, (char *)value);
		if ( ret )
			pMib->macFilterNum--;

		return ret;
	}
	if (id == MIB_MACFILTER_DELALL) {
		delete_all_linkchain(&macFilterChain);
		pMib->macFilterNum = 0;
		return 1;
	}
	if (id == MIB_TRIGGERPORT_ADD) {
		ret = add_linkchain(&triggerPortChain, (char *)value);
		if ( ret )
			pMib->triggerPortNum++;

		return ret;
	}
	if (id == MIB_TRIGGERPORT_DEL) {
		ret = delete_linkchain(&triggerPortChain, (char *)value);
		if ( ret )
			pMib->triggerPortNum--;

		return ret;
	}
	if (id == MIB_TRIGGERPORT_DELALL) {
		delete_all_linkchain(&triggerPortChain);
		pMib->triggerPortNum = 0;
		return 1;
	}
	if (id == MIB_WAN1_QOS_ADD) {	//WAN1
		ret = add_linkchain(&wan1QosChain, (char *)value);
		if ( ret )
			(pMib->QosNum[0])++;
		return ret;
	}
	if (id == MIB_WAN1_QOS_EDIT) {
		ret = edit_linkchain(&wan1QosChain, (char *)value);
		return ret;
	}
	if (id == MIB_WAN1_QOS_UP) {
		ret = moveUp_linkchain(&wan1QosChain, (char *)value);
		return ret;
	}
	if (id == MIB_WAN1_QOS_DOWN) {
		ret = moveDown_linkchain(&wan1QosChain, (char *)value);
		return ret;
	}
	if (id == MIB_WAN1_QOS_DEL) {
		ret = delete_linkchain(&wan1QosChain, (char *)value);
		if ( ret )
			pMib->QosNum[0]--;
		return ret;
	}
	if (id == MIB_WAN1_QOS_DELALL) {
		delete_all_linkchain(&wan1QosChain);
		pMib->QosNum[0] = 0;
		return 1;
	}

#if  defined(_MPPPOE_)
	if (id == MIB_WAN2_QOS_ADD) {	//WAN2
		ret = add_linkchain(&wan2QosChain, (char *)value);
		if ( ret )
			(pMib->QosNum[1])++;
		return ret;
	}
	if (id == MIB_WAN2_QOS_EDIT) {
		ret = edit_linkchain(&wan2QosChain, (char *)value);
		return ret;
	}
	if (id == MIB_WAN2_QOS_UP) {
		ret = moveUp_linkchain(&wan2QosChain, (char *)value);
		return ret;
	}
	if (id == MIB_WAN2_QOS_DOWN) {
		ret = moveDown_linkchain(&wan2QosChain, (char *)value);
		return ret;
	}
	if (id == MIB_WAN2_QOS_DEL) {
		ret = delete_linkchain(&wan2QosChain, (char *)value);
		if ( ret )
			pMib->QosNum[1]--;
		return ret;
	}
	if (id == MIB_WAN2_QOS_DELALL) {
		delete_all_linkchain(&wan2QosChain);
		pMib->QosNum[1] = 0;
		return 1;
	}
#endif
///////////////////////////////////////////////////////////////////////////////////////
#else
                                //Radius User
	if (id == MIB_RSER_USR_ADD) {
		ret = add_linkchain(&rserUsrChain, (char *)value);
		if ( ret )
			pMib->rserUsrNum++;

		return ret;
	}
	if (id == MIB_RSER_USR_DEL) {
		ret = delete_linkchain(&rserUsrChain, (char *)value);
		if ( ret )
			pMib->rserUsrNum--;

		return ret;
	}
	if (id == MIB_RSER_USR_DELALL) {
		delete_all_linkchain(&rserUsrChain);
		pMib->rserUsrNum = 0;
		return 1;
	}
                                //Radius Client
	if (id == MIB_RSER_CLT_ADD) {
		ret = add_linkchain(&rserCltChain, (char *)value);
		if ( ret )
			pMib->rserCltNum++;

		return ret;
	}
	if (id == MIB_RSER_CLT_DEL) {
		ret = delete_linkchain(&rserCltChain, (char *)value);
		if ( ret )
			pMib->rserCltNum--;

		return ret;
	}
	if (id == MIB_RSER_CLT_DELALL) {
		delete_all_linkchain(&rserCltChain);
		pMib->rserCltNum = 0;
		return 1;
	}
///////////////////////////////////////////////////////////////////////////////////////////
#endif

	for (i=0; mib_table[i].id; i++) {
		if ( mib_table[i].id == id )
			break;
	}
	if ( mib_table[i].id != 0 ) {
		pMibTbl = (void *)pMib;
		pTbl = mib_table;
	}
	else {
		// if not found, search hw setting table
		for (i=0; hwmib_table[i].id; i++) {
			if ( hwmib_table[i].id == id )
				break;
		}
		if (hwmib_table[i].id == 0) // not found
			return 0;
		pMibTbl = (void *)pHwSetting;
		pTbl = hwmib_table;
	}

	switch (pTbl[i].type) {
	case BYTE_T:
//		*((unsigned char *)(((long)pMibTbl) + pTbl[i].offset)) = (unsigned char)(*((int *)value));
		ch = (unsigned char)(*((int *)value));
		memcpy( ((char *)pMibTbl) + pTbl[i].offset, &ch, 1);
		break;

	case WORD_T:
//		*((unsigned short *)(((long)pMibTbl) + pTbl[i].offset)) = (unsigned short)(*((int *)value));
		wd = (unsigned short)(*((int *)value));
		memcpy( ((char *)pMibTbl) + pTbl[i].offset, &wd, 2);
		break;

	case STRING_T:
		if ( strlen(value)+1 > pTbl[i].size )
			return 0;
		strcpy((char *)(((long)pMibTbl) + pTbl[i].offset), (char *)value);
		break;

	case BYTE5_T:
		memcpy((unsigned char *)(((long)pMibTbl) + pTbl[i].offset), (unsigned char *)value, 5);
		break;

	case BYTE6_T:
		memcpy((unsigned char *)(((long)pMibTbl) + pTbl[i].offset), (unsigned char *)value, 6);
		break;

	case BYTE8_T:
		memcpy((unsigned char *)(((long)pMibTbl) + pTbl[i].offset), (unsigned char *)value, 8);
		break;

	case BYTE13_T:
		memcpy((unsigned char *)(((long)pMibTbl) + pTbl[i].offset), (unsigned char *)value, 13);
		break;

	case BYTE16_T:
		memcpy((unsigned char *)(((long)pMibTbl) + pTbl[i].offset), (unsigned char *)value, 16);
		break;

	case DWORD_T:
		dwd = *(unsigned long *)value;
		memcpy( ((char *)pMibTbl) + pTbl[i].offset, &dwd, 4);
		break;

	case INDEX_T:
		dwd = *(unsigned long *)value;
		memcpy( ((char *)pMibTbl) + pTbl[i].offset, &dwd, 4);
		break;

	case INDEXDOS_T://////////////////////////////////////////////////////////////
		dwd = *(unsigned long *)value;
		memcpy( ((char *)pMibTbl) + pTbl[i].offset, &dwd, 4);
		break;
#if 0
#ifdef WLAN_WPA
	case BYTE32_T:
		memcpy((unsigned char *)(((long)pMibTbl) + pTbl[i].offset), (unsigned char *)value, 32);
		break;
#endif
#endif

	case BYTE_ARRAY_T:

		tmp = (unsigned char*) value;
		max_chan_num = (id == MIB_HW_TX_POWER_CCK)? MAX_CCK_CHAN_NUM: MAX_OFDM_CHAN_NUM ;
		if(tmp[2] == 0xff){ // set one channel value
			memcpy((unsigned char *)(((long)pMibTbl) + pTbl[i].offset + (long)tmp[0] -1), (unsigned char *)(tmp+1), 1);
		}
		else
			memcpy((unsigned char *)(((long)pMibTbl) + pTbl[i].offset), (unsigned char *)value, max_chan_num);
		break;



//		memcpy((unsigned char *)(((long)pMibTbl) + pTbl[i].offset), (unsigned char *)value, pTbl[i].size);
//		break;

	case IA_T:
		memcpy((unsigned char *)(((long)pMibTbl) + pTbl[i].offset), (unsigned char *)value,  4);
		break;

	case WLAC_ARRAY_T:
// Static DHCP Lease --------------
	case SDHCP_ARRAY_T:
//---------------------------------

#ifdef _EZVIEW_
	case EZVIEW_ARRAY_T:
#endif

#ifdef _STARCRAFT_
	case STARCRAFT_ARRAY_T:
#endif

#ifdef _QoS_MODE_SELECT_
	case WAN_QOSCON_ARRAY_T:
#endif

#ifdef _VLAN_CONFIG_SUPPORTED_
	case VLANCONFIG_ARRAY_T:		
#endif

#ifdef HOME_GATEWAY
	case VSER_ARRAY_T://///////////////////////////////////////////////////////////////
	case DMZ_ARRAY_T://///////////////////////////////////////////////////////////////
	case URLB_ARRAY_T://///////////////////////////////////////////////////////////////
	case ACPC_ARRAY_T://///////////////////////////////////////////////////////////////
	case SROUT_ARRAY_T://///////////////////////////////////////////////////////////////
	case PORTFW_ARRAY_T:
	case IPFILTER_ARRAY_T:
	case PORTFILTER_ARRAY_T:
	case MACFILTER_ARRAY_T:
	case TRIGGERPORT_ARRAY_T:
	case WAN1_QOS_ARRAY_T:
#if  defined(_MPPPOE_)
	case WAN2_QOS_ARRAY_T:
#endif
#else
	case RSER_CLT_ARRAY_T:	//radius client
	case RSER_USR_ARRAY_T:	//radius user
#endif
		return 0;
	}
	return 1;
}


////////////////////////////////////////////////////////////////////////////////
/*int apmib_setDef(int id, void *value)		//removed by Erwin
{
	int ret;
	APMIB_Tp saveMib=pMib;

	pMib = pMibDef;
	ret = apmib_set(id, value);
	pMib = saveMib;
	return ret;
}
*/


////////////////////////////////////////////////////////////////////////////////
/* Update current used MIB into flash in current setting area
 */
int apmib_update(CONFIG_DATA_T type)
{
	int i, len;
	unsigned char checksum;
	unsigned char *data;

	if (type & HW_SETTING) {
		data = (char *)pHwSetting;
		checksum = CHECKSUM(data, hsHeader.len-1);
		data[hsHeader.len-1] = checksum;

		if ( flash_write((char *)data, HW_SETTING_OFFSET+sizeof(hsHeader), hsHeader.len)==0 ) {
			printf("write hs MIB failed!\n");
			return 0;
		}
	}

	if ((type & CURRENT_SETTING) || (type & DEFAULT_SETTING)) {
		memset( pMib->acAddrArray, '\0', MAX_WLAN_AC_NUM*sizeof(MACFILTER_T) );
		for (i=0; i<pMib->acNum; i++) {
			get_linkchain(&wlanMacChain, (void *)&pMib->acAddrArray[i], i+1);
		}
#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_ )
		memset( pMib->acAddrArray2, '\0', MAX_WLAN_AC_NUM*sizeof(MACFILTER_T) );
		for (i=0; i<pMib->acNum2; i++) {
			get_linkchain(&wlanMacChain2, (void *)&pMib->acAddrArray2[i], i+1);
		}
		memset( pMib->acAddrArray3, '\0', MAX_WLAN_AC_NUM*sizeof(MACFILTER_T) );
		for (i=0; i<pMib->acNum3; i++) {
			get_linkchain(&wlanMacChain3, (void *)&pMib->acAddrArray3[i], i+1);
		}
		memset( pMib->acAddrArray4, '\0', MAX_WLAN_AC_NUM*sizeof(MACFILTER_T) );
		for (i=0; i<pMib->acNum4; i++) {
			get_linkchain(&wlanMacChain4, (void *)&pMib->acAddrArray4[i], i+1);
		}
		memset( pMib->acAddrArray5, '\0', MAX_WLAN_AC_NUM*sizeof(MACFILTER_T) );
		for (i=0; i<pMib->acNum5; i++) {
			get_linkchain(&wlanMacChain5, (void *)&pMib->acAddrArray5[i], i+1);
		}
#endif

                memset( pMib->WDSList, '\0', MAX_WDS_NUM*sizeof(WDSList_T) );
                for (i=0; i<pMib->WDSNum; i++) {
                        get_linkchain(&wdsListChain, (void *)&pMib->WDSList[i], i+1);
                }

//vance 2008.11.20
#if defined(_WIRELESS_SCHEDULE_)
		memset( pMib->WlanSchArray, '\0', MAX_WLAN_SCH_NUM*sizeof(SCHEDULE_T) );
		for (i=0; i<pMib->WlanSchNum; i++) {
			get_linkchain(&wlanSchChain, (void *)&pMib->WlanSchArray[i], i+1);
		}
#endif

//Felix 21011222
#if defined(_PROFILE_LIST_)
		memset( pMib->ProfileListArray, '\0', MAX_PROFILE_LIST_NUM*sizeof(PROFILELIST_T) );
		for (i=0; i<pMib->ProfileListNum; i++) {
			get_linkchain(&profilelistChain, (void *)&pMib->ProfileListArray[i], i+1);
		}
#endif

// Static DHCP Lease -------------------------------------------------------------
		memset( pMib->SDHCPArray, '\0', MAX_STATIC_DHCP_NUM*sizeof(SDHCP_T) );
		for (i=0; i<pMib->SDHCPNum; i++) {
			get_linkchain(&staticDhcpChain, (void *)&pMib->SDHCPArray[i], i+1);
		}
// -------------------------------------------------------------------------------

#ifdef _EZVIEW_
		memset( pMib->EZVIEWArray, '\0', MAX_EZVIEW_NUM*sizeof(EZVIEW_T) );
		for (i=0; i<pMib->EZVIEWNum; i++) {
			get_linkchain(&EZVIEWChain, (void *)&pMib->EZVIEWArray[i], i+1);
		}
#endif

#ifdef _STARCRAFT_
		memset( pMib->STARCRAFTArray, '\0', MAX_STARCRAFT_NUM*sizeof(STARCRAFT_T) );
		for (i=0; i<pMib->STARCRAFTNum; i++) {
			get_linkchain(&STARCRAFTChain, (void *)&pMib->STARCRAFTArray[i], i+1);
		}
#endif

#ifdef _QoS_MODE_SELECT_
		memset( pMib->QosCONArray, '\0', MAX_QOSCON_NUM*sizeof(QOSCON_T) );
		for (i=0; i<pMib->QosCONNum; i++) {
			get_linkchain(&wanQosCONChain, (void *)&pMib->QosCONArray[i], i+1);
		}
#endif

#ifdef _VLAN_CONFIG_SUPPORTED_
		memset(pMib->VlanConfigArray, '\0', MAX_IFACE_VLAN_CONFIG*sizeof(VLAN_CONFIG_T));
		for (i=0; i<pMib->VlanConfigNum; i++) {
			get_linkchain(&vlanConfigChain, (void *)&pMib->VlanConfigArray[i], i+1);
		}
#endif	

#ifdef HOME_GATEWAY
		memset( pMib->portFwArray, '\0', MAX_FILTER_NUM*sizeof(PORTFW_T) );
		for (i=0; i<pMib->portFwNum; i++) {
			get_linkchain(&portFwChain, (void *)&pMib->portFwArray[i], i+1);
		}
////////////////////////////////////////////////////////////////////////////////////////////
		memset( pMib->DMZArray, '\0', MAX_FILTER_NUM*sizeof(DMZ_T) );
		for (i=0; i<pMib->DMZNum; i++) {
			get_linkchain(&dmzChain, (void *)&pMib->DMZArray[i], i+1);
		}
		memset( pMib->URLBArray, '\0', MAX_FILTER_NUM*sizeof(URLB_T) );
		for (i=0; i<pMib->URLBNum; i++) {
			get_linkchain(&URLBChain, (void *)&pMib->URLBArray[i], i+1);
		}
		memset( pMib->ACPCArray, '\0', MAX_FILTER_NUM*sizeof(ACPC_T) );
		for (i=0; i<pMib->ACPCNum; i++) {
			get_linkchain(&ACPCChain, (void *)&pMib->ACPCArray[i], i+1);
		}
		memset( pMib->sroutArray, '\0', MAX_FILTER_NUM*sizeof(SROUT_T) );
		for (i=0; i<pMib->sroutNum; i++) {
			get_linkchain(&sroutChain, (void *)&pMib->sroutArray[i], i+1);
		}
//		memset( pMib->ipFilterArray, '\0', MAX_FILTER_NUM*sizeof(IPFILTER_T) );
		memset( pMib->ipFilterArray, '\0', sizeof(IPFILTER_T) );
		for (i=0; i<pMib->ipFilterNum; i++) {
			get_linkchain(&ipFilterChain, (void *)&pMib->ipFilterArray[i], i+1);
		}
		memset( pMib->vserArray, '\0', MAX_FILTER_NUM*sizeof(PORTFW_T) );
		for (i=0; i<pMib->vserNum; i++) {
			get_linkchain(&virtSvChain, (void *)&pMib->vserArray[i], i+1);
		}
//		memset( pMib->portFilterArray, '\0', MAX_FILTER_NUM*sizeof(PORTFILTER_T) );
		memset( pMib->portFilterArray, '\0', sizeof(PORTFILTER_T) );
		for (i=0; i<pMib->portFilterNum; i++) {
			get_linkchain(&portFilterChain, (void *)&pMib->portFilterArray[i], i+1);
		}
		memset( pMib->macFilterArray, '\0', MAX_FILTER_NUM*sizeof(MACFILTER_T) );
		for (i=0; i<pMib->macFilterNum; i++) {
			get_linkchain(&macFilterChain, (void *)&pMib->macFilterArray[i], i+1);
		}
		memset( pMib->triggerPortArray, '\0', MAX_FILTER_NUM*sizeof(TRIGGERPORT_T) );
		for (i=0; i<pMib->triggerPortNum; i++) {
			get_linkchain(&triggerPortChain, (void *)&pMib->triggerPortArray[i], i+1);
		}
		memset( pMib->QosArray[0], '\0', MAX_QOS_NUM*sizeof(QOS_T) );
		for (i=0; i<pMib->QosNum[0]; i++) {
			get_linkchain(&wan1QosChain, (void *)&pMib->QosArray[0][i], i+1);
		}
#if  defined(_MPPPOE_)
		memset( pMib->QosArray[1], '\0', MAX_QOS_NUM*sizeof(QOS_T) );
		for (i=0; i<pMib->QosNum[1]; i++) {
			get_linkchain(&wan2QosChain, (void *)&pMib->QosArray[1][i], i+1);
		}
#endif
////////////////////////////////////////////////////////////////////////////////////
#else
		memset( pMib->rserCltArray, '\0', 16*sizeof(RSER_CLT_T) );
		for (i=0; i<pMib->rserCltNum; i++) {
			get_linkchain(&rserCltChain, (void *)&pMib->rserCltArray[i], i+1);
		}
		memset( pMib->rserUsrArray, '\0', 96*sizeof(RSER_USR_T) );
		for (i=0; i<pMib->rserUsrNum; i++) {
			get_linkchain(&rserUsrChain, (void *)&pMib->rserUsrArray[i], i+1);
		}
////////////////////////////////////////////////////////////////////////////////////
#endif
/*		if (type & CURRENT_SETTING) {		//removed by Erwin
			data = (unsigned char *)pMib;
			checksum = CHECKSUM(data, csHeader.len-1);
			*(data + csHeader.len - 1) = checksum;
			i = CURRENT_SETTING_OFFSET + sizeof(csHeader);
			len = csHeader.len;
		}
		else {
			data = (unsigned char *)pMibDef;
*/			data = (unsigned char *)pMib;
			checksum = CHECKSUM(data, dsHeader.len-1);
			*(data + dsHeader.len - 1) = checksum;
			i = DEFAULT_SETTING_OFFSET + sizeof(dsHeader);
			len = dsHeader.len;
//		}								//removed by Erwin

		if ( flash_write((char *)data, i, len)==0 ) {
			printf("Write flash current-setting failed!1\n");
			return 0;
		}
	}

#ifdef __TARGET_BOARD__		//Erwin 2004.01.30
//	system("/bin/flash all > /var/flash.inc");  // move this to init.sh
#endif
	return 1;
}


////////////////////////////////////////////////////////////////////////////////
/* Update default setting MIB into current setting area
 */
/*int apmib_updateDef(void)		//removed by Erwin
{
	unsigned char *data, checksum;
	PARAM_HEADER_T header;
	int i;

	memcpy(header.signature, CURRENT_SETTING_HEADER_TAG, TAG_LEN);
	memcpy(&header.signature[TAG_LEN], &dsHeader.signature[TAG_LEN], SIGNATURE_LEN-TAG_LEN);

	header.len = dsHeader.len;
	data = (unsigned char *)pMibDef;
	checksum = CHECKSUM(data, header.len-1);
	*(data + header.len - 1) = checksum;

	i = CURRENT_SETTING_OFFSET;
	if ( flash_write((char *)&header, i, sizeof(header))==0 ) {
		printf("Write flash current-setting header failed!\n");
		return 0;
	}
	i += sizeof(header);

	if ( flash_write((char *)data, i, header.len)==0 ) {
		printf("Write flash current-setting failed!\n");
		return 0;
	}

	return 1;
}
*/

////////////////////////////////////////////////////////////////////////////////
/* Update MIB into flash current setting area
 */
int apmib_updateFlash(CONFIG_DATA_T type, char *data, int len, int force, int ver)
{
	unsigned char checksum, checksum1, *ptr=NULL;
	int i, offset=0, curLen, curVer;
	unsigned char *pMibData, *pHdr, tmpBuf[20];

	if ( type == HW_SETTING ) {
		curLen = hsHeader.len - 1;
		pMibData = (unsigned char *)pHwSetting;
		pHdr = (unsigned char *)&hsHeader;
		i = HW_SETTING_OFFSET;
	}
	else if ( type == DEFAULT_SETTING ) {
		curLen = dsHeader.len - 1;
//		pMibData = (unsigned char *)pMibDef;	//removed by Erwin
		pMibData = (unsigned char *)pMib;
		pHdr = (unsigned char *)&dsHeader;
		i = DEFAULT_SETTING_OFFSET;
	}
/*	else  {						//removed by Erwin
		curLen = csHeader.len - 1;
		pMibData = (unsigned char *)pMib;
		pHdr = (unsigned char *)&csHeader;
		i = CURRENT_SETTING_OFFSET;
	}
*/
	if (force==2) { // replace by input mib
		((PARAM_HEADER_Tp)pHdr)->len = len + 1;
		sprintf(tmpBuf, "%02d", ver);
		memcpy(&pHdr[TAG_LEN], tmpBuf, SIGNATURE_LEN-TAG_LEN);
		checksum = CHECKSUM(data, len);
		pMibData = data;
		curLen = len;
	}
	else if (force==1) { // update mib but keep not used mib
		sscanf(&((PARAM_HEADER_Tp)pHdr)->signature[TAG_LEN], "%02d", &curVer);
		if ( curVer < ver ) {
			sprintf(tmpBuf, "%02d", ver);
			memcpy(&((PARAM_HEADER_Tp)pHdr)->signature[TAG_LEN],
					tmpBuf, SIGNATURE_LEN-TAG_LEN);
		}
		checksum = CHECKSUM(data, len);
		if (curLen > len) {
			((PARAM_HEADER_Tp)pHdr)->len = curLen + 1;
			ptr = pMibData + len;
			offset = curLen - len;
			checksum1 = CHECKSUM(ptr, offset);
			checksum +=  checksum1;
		}
		else
			((PARAM_HEADER_Tp)pHdr)->len = len + 1;

		curLen = len;
		pMibData = data;
	}
	else { // keep old mib, only update new added portion
		sscanf(&((PARAM_HEADER_Tp)pHdr)->signature[TAG_LEN], "%02d", &curVer);
		if ( curVer < ver ) {
			sprintf(tmpBuf, "%02d", ver);
			memcpy(&((PARAM_HEADER_Tp)pHdr)->signature[TAG_LEN],
					tmpBuf, SIGNATURE_LEN-TAG_LEN);
		}
		if ( len > curLen ) {
			((PARAM_HEADER_Tp)pHdr)->len = len + 1;
			offset = len - curLen;
			checksum = CHECKSUM(pMibData, curLen);
			ptr = data + curLen;
			checksum1 = CHECKSUM(ptr, offset);
			checksum +=  checksum1;
		}
		else
			checksum = CHECKSUM(pMibData, curLen);
	}

	if ( flash_write((char *)pHdr, i, sizeof(PARAM_HEADER_T))==0 ) {
		printf("Write flash current-setting header failed!\n");
		return 0;
	}
	i += sizeof(PARAM_HEADER_T);

	if ( flash_write(pMibData, i, curLen)==0 ) {
		printf("Write flash current-setting failed!\n");
		return 0;
	}
	i += curLen;

	if (offset > 0) {
		if ( flash_write((char *)ptr, i, offset)==0 ) {
			printf("Write flash current-setting failed!\n");
			return 0;
		}
		i += offset;
	}

	if ( flash_write((char *)&checksum, i, sizeof(checksum))==0 ) {
		printf("Write flash current-setting checksum failed!\n");
		return 0;
	}

	return 1;
}

/////////////////////////////////////////////////////////////////////////////////
static int flash_read(char *buf, int offset, int len)
{
	int fh;
	int ok=1;
	fh = open(FLASH_DEVICE_NAME, O_RDWR);
	if ( fh == -1 ){
		return 0;
	}
	lseek(fh, offset, SEEK_SET);

	if ( read(fh, buf, len) != len){
		ok = 0;
	}
	close(fh);

	return ok;
}


////////////////////////////////////////////////////////////////////////////////
static int flash_write(char *buf, int offset, int len)
{
	int fh;
	int ok=1;

	fh = open(FLASH_DEVICE_NAME, O_RDWR);

	if ( fh == -1 )
		return 0;

	lseek(fh, offset, SEEK_SET);

	if ( write(fh, buf, len) != len)
		ok = 0;

	close(fh);
	sync();

	return ok;
}


///////////////////////////////////////////////////////////////////////////////
static int init_linkchain(LINKCHAIN_Tp pLinkChain, int size, int num)
{
	FILTER_Tp entry;
	int offset=sizeof(FILTER_Tp)*2;
	char *pBuf;
	int i;
	int remainder;
	remainder=size%4;
	if(remainder==0)
		pLinkChain->appendNumber=0;
	else{
		pLinkChain->appendNumber=4-remainder;
		size = (size/4+1)*4;
	}
	pBuf = calloc(num, size+offset);
//printf("size offset = %d %d\n", size, offset);

	if ( pBuf == NULL )
		return 0;

//printf("bbbb 3-3\n");
	pLinkChain->buf = pBuf;
	pLinkChain->pUsedList = NULL;
	pLinkChain->pFreeList = NULL;
	entry = (FILTER_Tp)pBuf;

//printf("bbbb 3-30 %p %p\n", pLinkChain->pFreeList, entry);
	ADD_LINK_LIST(pLinkChain->pFreeList, entry);
//printf("bbbb 3-31 %p %p\n", pLinkChain->pFreeList, entry);
	for (i=1; i<num; i++) {
		entry = (FILTER_Tp)&pBuf[i*(size+offset)];
//printf("bbbb 3-32 %p %p\n", pLinkChain->pFreeList, entry);
		ADD_LINK_LIST(pLinkChain->pFreeList, entry);
	}

//printf("bbbb 3-4\n");
	pLinkChain->size = size;
	pLinkChain->num = num;
	pLinkChain->usedNum = 0;
	return 1;
}

///////////////////////////////////////////////////////////////////////////////
static int add_linkchain(LINKCHAIN_Tp pLinkChain, char *val)
{
	FILTER_Tp entry;

	// get a free entry
	entry = pLinkChain->pFreeList;
	if (entry == NULL)
		return 0;

	if (entry->next==pLinkChain->pFreeList)
		pLinkChain->pFreeList = NULL;
	else
		pLinkChain->pFreeList = entry->next;

	REMOVE_LINK_LIST(entry);

	// copy content
	memcpy(entry->val,val, pLinkChain->size);

	// add to used list
	if (pLinkChain->pUsedList == NULL) {
		ADD_LINK_LIST(pLinkChain->pUsedList, entry);
	}
	else {
		ADD_LINK_LIST(pLinkChain->pUsedList->prev, entry);
	}
	pLinkChain->usedNum++;
	return 1;
}

///////////////////////////////////////////////////////////////////////////////
static int edit_linkchain(LINKCHAIN_Tp pLinkChain, char *val)	//Erwin add
{
	FILTER_Tp curEntry=pLinkChain->pUsedList;

	while (curEntry != NULL) {
		if ( !memcmp(curEntry->val, (unsigned char *)(((EDIT_Tp)val)->selEntry), pLinkChain->compareLen) ) {
			memcpy(curEntry->val, ((EDIT_Tp)val)->editEntry, pLinkChain->size);

			return 1;
		}
		if ( curEntry->next == pLinkChain->pUsedList )
			return 0;
		curEntry = curEntry->next;
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////
static int moveUp_linkchain(LINKCHAIN_Tp pLinkChain, char *val)	//Erwin add
{
	FILTER_Tp curEntry=pLinkChain->pUsedList->next;
	while (curEntry != NULL) {
		if ( !memcmp(curEntry->val, (unsigned char *)val, pLinkChain->compareLen) ) {
			if (curEntry==pLinkChain->pUsedList->next)
				pLinkChain->pUsedList=curEntry;

			curEntry->prev->prev->next = curEntry;
			curEntry->next->prev = curEntry->prev;

			curEntry->prev->next = curEntry->next;
			curEntry->next = curEntry->prev;
			curEntry->prev = curEntry->next->prev;
			curEntry->next->prev = curEntry;
			return 1;
		}
		if ( curEntry->next == pLinkChain->pUsedList )
			return 0;
		curEntry = curEntry->next;
	}
	return 0;
}
///////////////////////////////////////////////////////////////////////////////
static int moveDown_linkchain(LINKCHAIN_Tp pLinkChain, char *val)	//Erwin add
{
	FILTER_Tp curEntry=pLinkChain->pUsedList;
	while (curEntry != NULL) {
		if ( !memcmp(curEntry->val, (unsigned char *)val, pLinkChain->compareLen) ) {
			if (curEntry==pLinkChain->pUsedList)
				pLinkChain->pUsedList=curEntry->next;

			curEntry->prev->next = curEntry->next;
			curEntry->next->next->prev = curEntry;

			curEntry->next->prev = curEntry->prev;
			curEntry->prev = curEntry->next;
			curEntry->next = curEntry->prev->next;
			curEntry->prev->next = curEntry;
			return 1;
		}
		if ( curEntry->next == pLinkChain->pUsedList->prev )
			return 0;
		curEntry = curEntry->next;
	}
}
///////////////////////////////////////////////////////////////////////////////
static int delete_linkchain(LINKCHAIN_Tp pLinkChain, char *val)
{
	FILTER_Tp curEntry=pLinkChain->pUsedList;

	while (curEntry != NULL) {

		if ( !memcmp(curEntry->val,(unsigned char *)val,pLinkChain->compareLen) ) {

				if (curEntry == pLinkChain->pUsedList) {
					if ( pLinkChain->pUsedList->next != pLinkChain->pUsedList )
						pLinkChain->pUsedList = pLinkChain->pUsedList->next;
					else
						pLinkChain->pUsedList = NULL;
				}
				REMOVE_LINK_LIST(curEntry);
				ADD_LINK_LIST(pLinkChain->pFreeList, curEntry);
				pLinkChain->usedNum--;
				return 1;
		}
		if ( curEntry->next == pLinkChain->pUsedList )
			return 0;
		curEntry = curEntry->next;
	}

	return 0;
}

///////////////////////////////////////////////////////////////////////////////
static void delete_all_linkchain(LINKCHAIN_Tp pLinkChain)
{
	FILTER_Tp curEntry;

	if (pLinkChain->pUsedList==NULL)
		return;

	// search for matched mac address
	while (pLinkChain->pUsedList) {
		curEntry = pLinkChain->pUsedList;
		if (pLinkChain->pUsedList->next != pLinkChain->pUsedList)
			pLinkChain->pUsedList = pLinkChain->pUsedList->next;
		else
			pLinkChain->pUsedList = NULL;

		REMOVE_LINK_LIST(curEntry);
		ADD_LINK_LIST(pLinkChain->pFreeList, curEntry);
		pLinkChain->usedNum--;
	}
}

///////////////////////////////////////////////////////////////////////////////
static int get_linkchain(LINKCHAIN_Tp pLinkChain, char *val, int index)
{
	FILTER_Tp curEntry=pLinkChain->pUsedList;

	if ( curEntry == NULL || index > pLinkChain->usedNum)
 		return 0;

	while (--index > 0)
        	curEntry = curEntry->next;
	
	memcpy( (unsigned char *)val, curEntry->val, (pLinkChain->size-pLinkChain->appendNumber));

	return 1;
}
int update_linkchain(int fmt, void *Entry_old, void *Entry_new, int type_size)
{
	LINKCHAIN_Tp pLinkChain=NULL;
	FILTER_Tp curEntry;
	void *entry;
	int i; 
	int entry_cmp;
#ifdef HOME_GATEWAY
	 	if(fmt==PORTFW_ARRAY_T){
			pLinkChain = &portFwChain;	 
		}else if(fmt == IPFILTER_ARRAY_T){
			pLinkChain = &ipFilterChain;
		}else if(fmt == PORTFILTER_ARRAY_T){
			pLinkChain = &portFilterChain;
		}else if(fmt == MACFILTER_ARRAY_T){
			pLinkChain = &macFilterChain;
		}else	if(fmt==TRIGGERPORT_ARRAY_T){
				pLinkChain = &triggerPortChain;
		}
#ifdef ROUTE_SUPPORT				
		if(fmt==STATICROUTE_ARRAY_T){
		 	pLinkChain = &staticRouteChain;
		}
#endif			
#else
		//if(fmt==DHCPRSVDIP_ARRY_T){
		//	pLinkChain = &dhcpRsvdIpChain;	
		//}		
#endif			

#ifdef _VLAN_CONFIG_SUPPORTED_
		 if(fmt==VLANCONFIG_ARRAY_T){
		 	pLinkChain = &vlanConfigChain;	
		}
#endif	
	curEntry = pLinkChain->pUsedList;
	for(i=0;i<pLinkChain->usedNum;i++){
		entry = curEntry->val;
		entry_cmp=memcmp(entry, Entry_old, type_size );
		if(entry_cmp ==0){
		//	fprintf(stderr,"find the entry to update!\n");
			memcpy(entry, Entry_new, type_size);
			break;
		}
		curEntry = curEntry->next;
	}
	return 1;
}
