/*
 *      Operation routines for FLASH MIB access
 *
 *      Authors: David Hsu	<davidhsu@realtek.com.tw>
 *
 *      $Id: flash.c,v 1.2 2005/03/15 14:24:06 rex Exp $
 *
 */


/* System include files */
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <ctype.h>

#define noPARSE_TXT_FILE

/* Local include files */
#include "apmib.h"
#include "mibtbl.h"
#include "utility.h"
#include "../../../define/C_DEFINE.h"
/* Constand definitions */
#define DEC_FORMAT	("%d")
#define LONGDEC_FORMAT  ("%lu")
#define BYTE5_FORMAT	("%02x%02x%02x%02x%02x")
#define BYTE6_FORMAT	("%02x%02x%02x%02x%02x%02x")
#define BYTE8_FORMAT	("%02x%02x%02x%02x%02x%02x%02x%02x")
#define BYTE13_FORMAT	("%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x")
#define BYTE16_FORMAT	("%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x")

//#define STR_FORMAT	("%s")
//
#define STR_FORMAT	("'%s'")
#define HEX_FORMAT	("%02x")

#ifdef HOME_GATEWAY
#if defined(_Customer_) | defined(_MPPPOE_)
#define PORTFW_FORMAT	("%s,%d,%d,%d,%d")
#else
#define PORTFW_FORMAT	("%s,%d,%d,%d")
#endif
#define PORTFILTER_FORMAT ("%d,%d,%d")
#define IPFILTER_FORMAT	("%s,%d")
//#define TRIGGERPORT_FORMAT ("%d,%d,%d,%d,%d,%d")
#define DMZ_FORMAT	("%s,%s")
#define SROUT_FORMAT	("%s,%s,%s,%d,%d")
#define URLB_FORMAT	("%s")
#define ACPC_FORMAT	 ("%s,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%lu,%d")//////

#define TRIGGERPORT_FORMAT ("%d-%d:%d:%d,%d,%d,%d,%d,%d,%d,%d,%d,%d:%d")
#endif
#define MACFILTER_FORMAT ("%02x%02x%02x%02x%02x%02x")
#define MACFILTER_COLON_FORMAT ("%02x:%02x:%02x:%02x:%02x:%02x")
#define QOS_FORMAT ("%lu,%d,%lu,%d,%d,%s,%s,%s,%s,%d,%s,%s")
#define VLANCONFIG_FORMAT ("%s,%d,%d,%d,%d,%d,%d")
#ifdef WLAN_WDS
#define WDS_FORMAT      MACFILTER_FORMAT
#endif

#define SPACE	(' ')
#define EOL	('\n')

#ifdef _EZVIEW_
#define EZVIEW_FORMAT ("%s,%02x%02x%02x%02x%02x%02x,%d.%d.%d.%d,%d")
#endif

#ifdef _STARCRAFT_
#define STARCRAFT_FORMAT ("%s")
#endif
static int wlan_va_idx = 0; //EDX Felix
static int wlan_idx=0;
static int config_area;
static int initWlan(char *ifname);
//extern int wlan_idx;
static int vwlan_idxx;
/* Macro definition */
static int _is_hex(char c)
{
    return (((c >= '0') && (c <= '9')) ||
            ((c >= 'A') && (c <= 'F')) ||
            ((c >= 'a') && (c <= 'f')));
}

static int string_to_hex(char *string, unsigned char *key, int len)
{
	char tmpBuf[4];
	int idx, ii=0;
	for (idx=0; idx<len; idx+=2) {
		tmpBuf[0] = string[idx];
		tmpBuf[1] = string[idx+1];
		tmpBuf[2] = 0;
		if ( !_is_hex(tmpBuf[0]) || !_is_hex(tmpBuf[1]))
			return 0;

		key[ii++] = (unsigned char) strtol(tmpBuf, (char**)NULL, 16);
	}
	return 1;
}

static void convert_lower(char *str)
{	int i;
	int len = strlen(str);
	for (i=0; i<len; i++)
		str[i] = tolower(str[i]);
}

static int APMIB_GET(int id, void *val)
{
/*	if (config_area == 2)		//removed by Erwin
		return apmib_getDef(id, val);
	else
*/		return apmib_get(id, val);
}

static int APMIB_SET(int id, void *val)
{
/*	if (config_area == 2)		//removed by Erwin
		return apmib_setDef(id, val);
	else
*/		return apmib_set(id, val);
}

/* Local declarations routines */
static int flash_read(char *buf, int offset, int len);
static int writeDefault(void);
static int searchMIB(char *token);
//static void getMIB(char *name, int id, TYPE_T type, int num, int array_separate);
static void getMIB(char *name, int id, TYPE_T type, int num, int array_separate, char **val);
static void setMIB(char *name, int id, TYPE_T type, int len, int valnum, char **val);
#if defined(_SPECIAL_CHAR_FILTER_IN_SCRIPT_)
static int dump_filter_enable=0;
#endif

static void dumpAll(void);
static void showHelp(void);
static void showAllMibName(void);
static void showSetACHelp(void);
static void showSetwdsHelp(void);

#ifdef _WIRELESS_SCHEDULE_//vance 2008.11.17
static void showSetSchHelp(void);
#endif
static int read_flash_webpage(char *prefix, char *file);

#ifdef WLAN_WDS
static void showSetWdsHelp(void);
#endif

#ifdef _EZVIEW_
static void showSetEZviewHelp(void);
#endif

#ifdef _STARCRAFT_
static void showSetSTARCRAFTHelp(void);
#endif

#ifdef _QoS_MODE_SELECT_
static void showSetConHelp(void);
#endif

#ifdef HOME_GATEWAY
static void showSetVlanConfigHelp(void);
static void showSetPortFwHelp(void);
static void showSetPortFilterHelp(void);
static void showSetIpFilterHelp(void);
static void showSetMacFilterHelp(void);
static void showSetTriggerPortHelp(void);

#endif
// Static DHCP Lease
static void showSetStaticDhcpHelp(void);

//EDX
#ifdef WLAN_WPA
//static void generateWpaConf(char *outputFile);
static void generateWpaConf(char *outputFile, int isWds);
#endif

//#define WIFI_SIMPLE_CONFIG
#ifdef WIFI_SIMPLE_CONFIG
static int updateWscConf(char *in, char *out, int genpin);
#endif
//EDX
static void generateWdsList(void);
static void generateAclList(void);
#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_)
static void generateAclList2(void);
static void generateAclList3(void);
static void generateAclList4(void);
static void generateAclList5(void);
#endif
static void generatewdsList(void);

//vance 2008.11.20
#if defined(_WIRELESS_SCHEDULE_)
static void generateSchList(char *sel);
#endif

#ifdef PARSE_TXT_FILE
static int parseTxtConfig(char *filename, APMIB_Tp pConfig);
static int getToken(char *line, char *value);
static int set_mib(APMIB_Tp pMib, int id, void *value);
static void getVal2(char *value, char **p1, char **p2);
#ifdef HOME_GATEWAY
static void getVal3(char *value, char **p1, char **p2, char **p3);
static void getVal4(char *value, char **p1, char **p2, char **p3, char **p4);
static void getVal5(char *value, char **p1, char **p2, char **p3, char **p4, char **p5);
#endif



static int acNum;




static int SDHCPNum;
//------------------
#ifdef _EZVIEW_
static int EZVIEWNum;
#endif

#ifdef _STARCRAFT_
static int STARCRAFTNum;
#endif

#ifdef WLAN_WDS
static int wdsNum;
#endif

#ifdef _QoS_MODE_SELECT_
static int QosCONNum;
#endif

#ifdef HOME_GATEWAY
static int macFilterNum, portFilterNum, ipFilterNum, portFwNum, triggerPortNum, vserNum, DMZNum, ACPCNum, sroutNum, wan1QosNum;
#else
static int reserUsrNum, reserCltNun;
#endif
#ifdef _VLAN_CONFIG_SUPPORTED_
static int vlanConfigNum=MAX_IFACE_VLAN_CONFIG;	
#endif

/////////////////////////////////////////////////////////////////////////////////////////
static char __inline__ *getVal(char *value, char **p)
{
	int len=0;

	while (*value == ' ' ) value++;

	*p = value;

	while (*value && *value!=',') {
		value++;
		len++;
	}

	if ( !len ) {
		*p = NULL;
		return NULL;
	}

	if ( *value == 0)
		return NULL;

	*value = 0;
	value++;

	return value;
}
#endif  // PARSE_TXT_FILE

// Static DHCP Lease
static void showSetStaticDhcpHelp(void)
{
	printf("flash set SDHCP_TBL cmd\n");
	printf("cmd:\n");
	printf("      add mac-addr ip-addr -- append a static dhcp lease address.\n");
	printf("      del mac-addr -- delete a static dhcp lease address.\n");
	printf("      delall -- delete all static dhcp lease address.\n");
}

//////////////////////////////////////////////////////////////////////
static int resetHWDefault()
{
	PARAM_HEADER_T header;
	HW_SETTING_T mib;
	HW_SETTING_Tp pMib=&mib;
	unsigned char checksum;
	int fh,fh2, status;
	pMib = malloc (sizeof(APMIB_T) + 1);
	if( sizeof(HW_SETTING_T) != (HW_SETTING_LEN-sizeof(header)-sizeof(checksum)))
		printf("size of HW_SETTNG = %d, should be %d\n", sizeof(HW_SETTING_T), (HW_SETTING_LEN-sizeof(header)-sizeof(checksum)));


	fh = open(DEFAULT_SETTING_FILE, O_RDONLY);

	if ( fh == -1 ) {
		printf("open %s file failed!\n", DEFAULT_SETTING_FILE);
		return -1;
	}

	lseek(fh, HW_SETTING_OFFSET_IN_FILE, SEEK_SET);

	if (read(fh, &header, sizeof(header)) != sizeof(header)) {
		printf("read default setting header failed!\n");
		return -1;
	}

	if (read(fh, pMib, header.len) != header.len) {
		printf("read HW setting failed!, header.len = 0x%x\n", header.len);
		return -1;
	}

	close(fh);
		//printf("21Invalid hw setting signature or version number [sig=%01x%01x len=%d]!\n", header.signature[0],
		//	header.signature[1], header.len);


//	dump_mem(pMib, 0x1ffa);
	checksum = CHECKSUM((void *)pMib, header.len-1);
	status = CHECKSUM_OK((void *)pMib, header.len);
	if ( !status) {
		printf("default setting Checksum error!, checksum = %x\n", checksum);
		return -1;
	}


#ifdef __TARGET_BOARD__
	fh = open(FLASH_DEVICE_NAME, O_RDWR);
#else
	fh = open(FLASH_DEVICE_NAME, O_RDWR);
#endif

	if ( fh == -1 ) {
		printf("create file failed!\n");
		return -1;
	}

	lseek(fh, HW_SETTING_OFFSET, SEEK_SET);


	if ( write(fh, (const void *)&header, sizeof(header)) != sizeof(header)) {
        printf("write HW header failed!\n");
        return -1;
    }

	lseek(fh, HW_SETTING_OFFSET + sizeof(header), SEEK_SET);

	if ( write(fh, (const void *)pMib, header.len) !=  header.len) {
        printf("write hw MIB failed!\n");
        return -1;
    }


	sync();
	return 0;
}

static int resetDFDefault()
{
	PARAM_HEADER_T header;
//	APMIB_T mib;
//	APMIB_Tp pMib=&mib;
	APMIB_Tp pMib;
	unsigned char checksum;
	int fh, status;
	int ver;
		#ifdef _VLAN_CONFIG_SUPPORTED_
	int vlan_entry=0;
#endif

 //system("flash set WPS_CONFIG_TYPE 0");//reset wps = unConfigured
	pMib = malloc (sizeof(APMIB_T) + 1);

	if ( sscanf(&header.signature[2], "%02d", &ver) != 1)
		ver = -1;

	if( sizeof(HW_SETTING_T) != (HW_SETTING_LEN-sizeof(header)-sizeof(checksum)))
		printf("size of HW_SETTNG = %d, should be %d\n", sizeof(HW_SETTING_T), (HW_SETTING_LEN-sizeof(header)-sizeof(checksum)));


	if( sizeof(APMIB_T) != (DF_SETTING_LEN-sizeof(header)-sizeof(checksum)))
	{
		printf("size of APMIB = %d, should be %d. set_app_defined need + %d.\n", sizeof(APMIB_T), (DF_SETTING_LEN-sizeof(header)-sizeof(checksum)),sizeof(APMIB_T)-(DF_SETTING_LEN-sizeof(header)-sizeof(checksum)));
		printf("The size of APMIB_T->Reserved should be = %d\n", sizeof(pMib->Reserved) + (DF_SETTING_LEN-sizeof(header)-sizeof(checksum)) -  sizeof(APMIB_T));
	}


	fh = open(DEFAULT_SETTING_FILE, O_RDONLY);

	if ( fh == -1 ) {
		printf("open %s file failed!\n", DEFAULT_SETTING_FILE);
		return -1;
	}

	lseek(fh, DF_SETTING_OFFSET_IN_FILE, SEEK_SET);

	if (read(fh, &header, sizeof(header)) != sizeof(header)) {
		printf("read default setting header failed!\n");
		return -1;
	}
    //	printf("DFDEFAULT [sig=%01x%01x len=%d talbeSize=%d]!\n", header.signature[0],
			//header.signature[1], header.len,sizeof(APMIB_T));
			int readbyte = read(fh, pMib, header.len);
	if ( readbyte != header.len) {
		printf("read default setting failed!, header.len = 0x%x read=0x%x\n", header.len,readbyte);
		return -1;
	}

	
	close(fh);
		//printf("DFDEFAULT [sig=%01x%01x len=%d]!\n", header.signature[0],
		//	header.signature[1], header.len);
//	dump_mem(pMib, 0x1ffa);

	checksum = CHECKSUM((void *)pMib, header.len-1);
	status = CHECKSUM_OK((void *)pMib, header.len);

	if ( !status) {
		printf("%s:default setting Checksum error!, checksum = %x\n",__func__,checksum);
		return -1;
	}


	fh = open(FLASH_DEVICE_NAME, O_RDWR);

	if ( fh == -1 ) {
		printf("create file failed!\n");
		return -1;
	}

	lseek(fh, DEFAULT_SETTING_OFFSET, SEEK_SET);

	
	if ( write(fh, (const void *)&header, sizeof(header)) != sizeof(header)) {
        printf("write DS header failed!\n");
        return -1;
    }

	lseek(fh, DEFAULT_SETTING_OFFSET + sizeof(header), SEEK_SET);


	if ( write(fh, (const void *)pMib, header.len) !=  header.len) {
        printf("write DS MIB failed!\n");
        return -1;
    }

	close(fh);
	sync();

	return 0;
	/*PARAM_HEADER_T header;
	APMIB_T mib;
//	APMIB_Tp pMib=&mib;
	APMIB_Tp pMib;
	char *data;
	unsigned char checksum;
	int fh, status;
	int ver, offset;
		unsigned char *buff;
	#ifdef _VLAN_CONFIG_SUPPORTED_
	int vlan_entry=0;
#endif
	buff=calloc(1, 0x6000);
	if ( buff == NULL ) {
		printf("Allocate buffer failed!\n");
		return -1;
	}
	
	
 //system("flash set WPS_CONFIG_TYPE 0");//reset wps = unConfigured
	pMib = malloc (sizeof(APMIB_T) + 1);

	if ( sscanf(&header.signature[2], "%02d", &ver) != 1)
		ver = -1;

	if( sizeof(HW_SETTING_T) != (HW_SETTING_LEN-sizeof(header)-sizeof(checksum)))
		printf("size of HW_SETTNG = %d, should be %d\n", sizeof(HW_SETTING_T), (HW_SETTING_LEN-sizeof(header)-sizeof(checksum)));


	if( sizeof(APMIB_T) != (DF_SETTING_LEN-sizeof(header)-sizeof(checksum)))
	{
		printf("size of APMIB = %d, should be %d. set_app_defined need + %d.\n", sizeof(APMIB_T), (DF_SETTING_LEN-sizeof(header)-sizeof(checksum)),sizeof(APMIB_T)-(DF_SETTING_LEN-sizeof(header)-sizeof(checksum)));
		printf("The size of APMIB_T->Reserved should be = %d\n", sizeof(pMib->Reserved) + (DF_SETTING_LEN-sizeof(header)-sizeof(checksum)) -  sizeof(APMIB_T));
	}


	fh = open(DEFAULT_SETTING_FILE, O_RDONLY);

	if ( fh == -1 ) {
		printf("open %s file failed!\n", DEFAULT_SETTING_FILE);
		return -1;
	}

	lseek(fh, DF_SETTING_OFFSET_IN_FILE, SEEK_SET);

	if (read(fh, &header, sizeof(header)) != sizeof(header)) {
		printf("read default setting header failed!\n");
		return -1;
	}
    //	printf("DFDEFAULT [sig=%01x%01x len=%d talbeSize=%d]!\n", header.signature[0],
			//header.signature[1], header.len,sizeof(APMIB_T));
			int readbyte = read(fh, pMib, header.len);
	if ( readbyte != header.len) {
		printf("read default setting failed!, header.len = 0x%x read=0x%x\n", header.len,readbyte);
		return -1;
	}
	close(fh);
		//printf("DFDEFAULT [sig=%01x%01x len=%d]!\n", header.signature[0],
		//	header.signature[1], header.len);
//	dump_mem(pMib, 0x1ffa);

	checksum = CHECKSUM((void *)pMib, header.len-1);
	status = CHECKSUM_OK((void *)pMib, header.len);

	if ( !status) {
		printf("%s:default setting Checksum error!, checksum = %x\n",__func__,checksum);
		return -1;
	}


	fh = open(FLASH_DEVICE_NAME, O_RDWR);

	if ( fh == -1 ) {
		printf("create file failed!\n");
		return -1;
	}

	lseek(fh, DEFAULT_SETTING_OFFSET, SEEK_SET);

	if ( write(fh, (const void *)&header, sizeof(header)) != sizeof(header)) {
        printf("write DS header failed!\n");
        return -1;
    }

	lseek(fh, DEFAULT_SETTING_OFFSET + sizeof(header), SEEK_SET);


	if ( write(fh, (const void *)pMib, header.len) !=  header.len) {
        printf("write DS MIB failed!\n");
        return -1;
    }


	//close(fh);
	//sync();
	
	
	memset(pMib, '\0', sizeof(APMIB_T));
	
*/	
#if defined _VLAN_CONFIG_SUPPORTED_
for(vlan_entry=0;vlan_entry<MAX_IFACE_VLAN_CONFIG;vlan_entry++){
	pMib->VlanConfigArray[vlan_entry].enabled=0;
	pMib->VlanConfigArray[vlan_entry].vlanId = 1;
	pMib->VlanConfigArray[vlan_entry].cfi = 1;
}	
pMib->VlanConfigEnabled=0;

pMib->VlanConfigNum=MAX_IFACE_VLAN_CONFIG;
sprintf(pMib->VlanConfigArray[0].netIface, "%s", "eth0");
sprintf(pMib->VlanConfigArray[1].netIface, "%s", "eth2");
sprintf(pMib->VlanConfigArray[2].netIface, "%s", "eth3");
sprintf(pMib->VlanConfigArray[3].netIface, "%s", "eth4");
sprintf(pMib->VlanConfigArray[4].netIface, "%s", "wlan0");
sprintf(pMib->VlanConfigArray[5].netIface, "%s", "wlan0-va0");
sprintf(pMib->VlanConfigArray[6].netIface, "%s", "wlan0-va1");
sprintf(pMib->VlanConfigArray[7].netIface, "%s", "wlan0-va2");
sprintf(pMib->VlanConfigArray[8].netIface, "%s", "wlan0-va3");
sprintf(pMib->VlanConfigArray[9].netIface, "%s", "eth1");

pMib->VlanConfigArray[0].vlanId = 3022;
pMib->VlanConfigArray[0].priority = 7;
pMib->VlanConfigArray[1].vlanId = 3030;
pMib->VlanConfigArray[1].priority = 0;
pMib->VlanConfigArray[2].vlanId = 500;
pMib->VlanConfigArray[2].priority = 3;
#endif
	close(fh);
	sync();

	return 0;
/*
data = (char *)pMib;

fh = open(FLASH_DEVICE_NAME, O_RDWR);

	// write current setting
	sprintf(header.signature, "%s%02d", CURRENT_SETTING_HEADER_TAG, CURRENT_SETTING_VER);
	header.len = sizeof(APMIB_T) + sizeof(checksum);
	lseek(fh, CURRENT_SETTING_OFFSET, SEEK_SET);
	if ( write(fh, (const void *)&header, sizeof(header))!=sizeof(header) ) {
		printf("write cs header failed!\n");
		return -1;
	}
	if ( write(fh, (const void *)pMib, sizeof(mib))!=sizeof(mib) ) {
		printf("write cs MIB failed!\n");
		return -1;
	}
	if ( write(fh, (const void *)&checksum, sizeof(checksum))!=sizeof(checksum) ) {
		printf("write cs checksum failed!\n");
		return -1;
	}

	close(fh);
	sync();


	offset = DEFAULT_SETTING_OFFSET;
	if ( flash_read((char *)&header, offset, sizeof(header)) == 0) {
		printf("read ds header failed!\n");
		return -1;
	}
	offset += sizeof(header);
	if ( flash_read(buff, offset, header.len) == 0) {
		printf("read ds MIB failed!\n");
		return -1;
	}
	status = CHECKSUM_OK(buff, header.len);
	if ( !status) {
		printf("ds Checksum error!\n");
		return -1;
	}

	offset = CURRENT_SETTING_OFFSET;
	if ( flash_read((char *)&header, offset, sizeof(header)) == 0) {
		printf("read cs header failed!\n");
		return -1;
	}
	offset += sizeof(header);
	if ( flash_read(buff, offset, header.len) == 0) {
		printf("read cs MIB failed!\n");
		return -1;
	}
	status = CHECKSUM_OK(buff, header.len);

	if ( !status) {
		printf("cs Checksum error!\n");
		return -1;
	}


	free(buff);
	
	return 0;*/
}

/*static int resetDefault()		//removed by Erwin
{
	PARAM_HEADER_T header;
	APMIB_T mib;
	APMIB_Tp pMib=&mib;
	unsigned char checksum;
	int fh, status;
#ifndef __TARGET_BOARD__
	char	tmpBuf[80];
#endif
	unsigned int tmpLen;

	if( sizeof(HW_SETTING_T) != (HW_SETTING_LEN-sizeof(header)-sizeof(checksum)))
		printf("size of HW_SETTNG = %d, should be %d\n", sizeof(HW_SETTING_T), (HW_SETTING_LEN-sizeof(header)-sizeof(checksum)));

	if( sizeof(APMIB_T) != (DF_SETTING_LEN-sizeof(header)-sizeof(checksum)))
	{
		printf("size of APMIB = %d, should be %d\n", sizeof(APMIB_T), (DF_SETTING_LEN-sizeof(header)-sizeof(checksum)));
		printf("The size of APMIB_T->Reserved should be = %d\n", sizeof(pMib->Reserved) + (DF_SETTING_LEN-sizeof(header)-sizeof(checksum)) -  sizeof(APMIB_T));
	}

	fh = open(DEFAULT_SETTING_FILE, O_RDONLY);

	if ( fh == -1 ) {
		printf("open %s file failed!\n", DEFAULT_SETTING_FILE);
		return -1;
	}

	lseek(fh, HW_SETTING_LEN, SEEK_SET);

	if (read(fh, &header, sizeof(header)) != sizeof(header)) {
		printf("read default setting header failed!\n");
		return -1;
	}

	tmpLen = header.len;

	if (read(fh, pMib, header.len) != header.len) {
		printf("read default setting failed!, header.len = 0x%x\n", header.len);
		return -1;
	}
	close(fh);

//	dump_mem(pMib, 0x1ffa);

	checksum = CHECKSUM((void *)pMib, header.len-1);
	status = CHECKSUM_OK((void *)pMib, header.len);
	if ( !status) {
		printf("default setting Checksum error!, checksum = %x\n", checksum);
		return -1;
	}

#ifndef __TARGET_BOARD__
	sprintf(tmpBuf, "cp -f %s %s", DEFAULT_SETTING_FILE, FLASH_DEVICE_NAME);

	system(tmpBuf);
#endif

#ifdef __TARGET_BOARD__
	fh = open(FLASH_DEVICE_NAME, O_RDWR);
#endif

#ifdef __TARGET_BOARD__
	fh = open(FLASH_DEVICE_NAME, O_RDWR);
#endif
	if ( fh == -1 ) {
		printf("create file failed!\n");
		return -1;
	}

    sprintf(header.signature, "%s%02d", CURRENT_SETTING_HEADER_TAG, CURRENT_SETTING_VER); // This will corruptheader.len


	header.len = tmpLen;

	lseek(fh, CURRENT_SETTING_OFFSET, SEEK_SET);

	if ( write(fh, (const void *)&header, sizeof(header)) != sizeof(header)) {
        printf("write cs header failed!\n");
        return -1;
    }

	lseek(fh, CURRENT_SETTING_OFFSET + sizeof(header), SEEK_SET);

	if ( write(fh, (const void *)pMib, header.len) !=  header.len) {
        printf("write cs MIB failed!\n");
        return -1;
    }

	close(fh);
	sync();
	return 0;
}
*/


//////////////////////////////////////////////////////////////////////
int main(int argc, char** argv)
{
	int argNum=1, action=0, idx, num, valNum=0;
	char mib[100]={0}, valueArray[170][100], *value[170], *ptr;
#ifdef PARSE_TXT_FILE
	char filename[100]={0};
	APMIB_T apmib;
#endif
	if ( argc > 1 ) {
#ifdef PARSE_TXT_FILE
		if ( !strcmp(argv[argNum], "-f") ) {
			if (++argNum < argc)
				sscanf(argv[argNum++], "%s", filename);
		}
#endif
		if ( !strcmp(argv[argNum], "get") ) {
			action = 1;
			if (++argNum < argc)
				sscanf(argv[argNum], "%s", mib);

				while (++argNum < argc) {	//Erwin
					sscanf(argv[argNum], "%s", valueArray[valNum]);
					value[valNum] = valueArray[valNum];
					valNum++;
				}
				value[valNum]= NULL;
		}
		else if ( !strcmp(argv[argNum], "set") ) {
			action = 2;
			if (++argNum < argc) {
				sscanf(argv[argNum], "%s", mib);

				while (++argNum < argc) {

					if( !strcmp(mib,"SSID")
						|| !strcmp(mib,"REPEATER_SSID")
						|| !strcmp(mib,"SSID_1")
						|| !strcmp(mib,"SSID_2")
						|| !strcmp(mib,"SSID_3")
						|| !strcmp(mib,"SSID_4")
						|| !strcmp(mib,"SSID_5")
						|| !strcmp(mib,"SSID_6")
						|| !strcmp(mib,"SSID_7")
						|| !strcmp(mib,"STADRV_SSID")
						|| !strcmp(mib,"WLAN_WPA_PSK")
						|| !strcmp(mib,"DEVICE_NAME")
					)
					{
						memcpy(valueArray[valNum],argv[argNum],strlen(argv[argNum]));
					}
					else{
						sscanf(argv[argNum], "%s", valueArray[valNum]);
					}
					value[valNum] = valueArray[valNum];
					valNum++;
				}
				value[valNum]= NULL;	//Erwin

			}
		}
		else if ( !strcmp(argv[argNum], "all") ) {
			action = 3;
		}
		else if ( !strcmp(argv[argNum], "default") ) {
			return writeDefault();
		}
		else if ( !strcmp(argv[argNum], "reset") ) {
//			return resetDefault();		//removed by Erwin
			return writeDefault();
		}
#if defined(_SPECIAL_CHAR_FILTER_IN_SCRIPT_)
		else if ( !strcmp(argv[argNum], "all2") ) {
			action = 4;
		}
#endif
#if 0
		else if ( !strcmp(argv[argNum], "extr") ) {
			if (++argNum < argc) {
				char prefix[20], file[20]={0};
				int ret;
				sscanf(argv[argNum], "%s", prefix);
				if (++argNum < argc)
					sscanf(argv[argNum], "%s", file);
				ret  = read_flash_webpage(prefix, file);
				if (ret == 0) // success
					unlink(file);
				return ret;
			}
			printf("Usage: %s web prefix\n", argv[0]);
			return -1;
		}
#endif
		else if ( !strcmp(argv[argNum], "test-hwconf") ) {
			if ((ptr=apmib_hwconf()) == NULL)
				return -1;
			free(ptr);
			return 0;
		}
		/*else if ( !strcmp(argv[argNum], "set_mib") ) {					//Walter add
			if ((argNum+1) < argc) {
				if (memcmp(argv[++argNum], "wlan", 4)) {
					printf("Miss wlan_interface argument!\n");
					return -1;
				}
#ifdef MSSID
				if (strlen(argv[argNum]) >= 9 && argv[argNum][5] == '-' &&
						argv[argNum][6] == 'v' && argv[argNum][7] == 'a') {
					idx = atoi(&argv[argNum][8]);
					if (idx >= NUM_VWLAN_INTERFACE) {
						printf("invalid virtual wlan interface index number!\n");
						return 0;
					}
					vwlan_idxx = idx+1;
				}
				if (strlen(argv[argNum]) >= 9 && argv[argNum][5] == '-' &&
						!memcmp(&argv[argNum][6], "vxd", 3)) {
					vwlan_idxx = NUM_VWLAN_INTERFACE;
				}

#endif
				wlan_idx = atoi(&argv[argNum][4]);
				if (wlan_idx >= NUM_WLAN_INTERFACE) {
					printf("invalid wlan interface index number!\n");
					goto error_return;
				}
#ifdef MSSID
				int ret;
				ret = initWlan(argv[argNum]);
				vwlan_idxx = 0;
				return ret;
#else
				return initWlan(argv[argNum]);
#endif
			}
			else
				printf("Miss arguments [wlan_interface]!\n");
			return -1;
		}*/
		else if ( !strcmp(argv[argNum], "test-dsconf") ) {
			if ((ptr=apmib_dsconf()) == NULL)
				return -1;
			free(ptr);
			return 0;
		}		else if ( !strcmp(argv[argNum], "hw-default") ) {
				resetHWDefault();
			return 0;
		}


/*		else if ( !strcmp(argv[argNum], "test-csconf") ) {		//removed by Erwin
			if ((ptr=apmib_csconf()) == NULL)
				return -1;
			free(ptr);
			return 0;
		}
*/


        else if ( !strcmp(argv[argNum], "wdsList") ) {  // Rexhua
			generateWdsList();
        	return 0;
        }
        else if ( !strcmp(argv[argNum], "aclList") ) {  // Rexhua
			generateAclList();
        	return 0;
        }
#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_)

        else if ( !strcmp(argv[argNum], "aclList2") ) {  // Kyle for Multiple SSID2
        	generateAclList2();
        	return 0;
        }
        else if ( !strcmp(argv[argNum], "aclList3") ) {  // Kyle for Multiple SSID3
        	generateAclList3();
        	return 0;
        }
        else if ( !strcmp(argv[argNum], "aclList4") ) {  // Kyle for Multiple SSID4
        	generateAclList4();
        	return 0;
        }
		else if ( !strcmp(argv[argNum], "aclList5") ) {  // Kyle for Multiple SSID5
			generateAclList5();
			return 0;
		}

#endif

        else if ( !strcmp(argv[argNum], "WdsList") ) {  // Kyle for Multiple SSID2
                generatewdsList();
                return 0;
        }

//vance 2008.11.20
#if defined(_WIRELESS_SCHEDULE_)
        else if ( !strcmp(argv[argNum], "schList") ) {
          	generateSchList(argv[2]);
          	return 0;
          }
#endif

//EDX Jeff
		else if ( !strcmp(argv[argNum], "wpa") ) {
			int isWds = 0;
			int wlan_idx = 0; //EDX Jeff
			if ((argNum+2) < argc) {
				if (memcmp(argv[++argNum], "wlan", 4)) {
					printf("Miss wlan_interface argument!\n");
					return 0;
				}
#ifdef _MSSID_
				if (strlen(argv[argNum]) >= 9 && argv[argNum][5] == '-' &&
						argv[argNum][6] == 'v' && argv[argNum][7] == 'a') {
					idx = atoi(&argv[argNum][8]);
					if (idx >= NUM_VWLAN_INTERFACE) {
						printf("invalid virtual wlan interface index number!\n");
						return 0;
					}
					vwlan_idxx = idx+1;
					wlan_va_idx = idx+1;
				}
#ifdef UNIVERSAL_REPEATER
				if (strlen(argv[argNum]) >= 9 && argv[argNum][5] == '-' &&
						!memcmp(&argv[argNum][6], "vxd", 3)) {
					vwlan_idxx = NUM_VWLAN_INTERFACE;
				}
#endif
#endif
				wlan_idx = atoi(&argv[argNum][4]);
				if (wlan_idx >= NUM_WLAN_INTERFACE) {
					printf("invalid wlan interface index number!\n");
					goto normal_return;
				}

#ifdef CONFIG_RTK_MESH
				if ((argNum+2) < argc)
				{
					if( !strcmp(argv[argNum+2], "wds"))
						isWds = 1;
					else if(!strcmp(argv[argNum+2], "msh"))
						isWds = 7;
				}
#else
				if (((argc-1) > (argNum+1)) && !strcmp(argv[argNum+2], "wds"))
					isWds = 1;
#endif // CONFIG_RTK_MESH

				generateWpaConf(argv[argNum+1], isWds);
				goto normal_return;
			}
			else {
				printf("Miss arguments [wlan_interface config_filename]!\n");
				return 0;
			}
		}
#ifdef WIFI_SIMPLE_CONFIG
		else if ( !strcmp(argv[argNum], "upd-wsc-conf") ) {
			return updateWscConf(argv[argNum+1], argv[argNum+2], 0);
		}

		else if ( !strcmp(argv[argNum], "gen-pin") ) {
			return updateWscConf(0, 0, 1);
		}
#endif // WIFI_SIMPLE_CONFIG
//EDX Jeff
	}
	if ( action == 0) {
		showHelp();
		return -1;
	}
	if ( (action==1 && !mib[0]) ||
	     (action==2 && !mib[0]) ) {
		showAllMibName();
		return -1;
	}
#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_ )
	if ( action==2 && (!strcmp(mib, "WLAN_MACAC_ADDR") || !strcmp(mib, "DEF_WLAN_MACAC_ADDR") || !strcmp(mib, "WLAN_MACAC2_ADDR") || !strcmp(mib, "DEF_WLAN_MACAC2_ADDR") || !strcmp(mib, "WLAN_MACAC3_ADDR") || !strcmp(mib, "DEF_WLAN_MACAC3_ADDR") || !strcmp(mib, "WLAN_MACAC4_ADDR") || !strcmp(mib, "DEF_WLAN_MACAC4_ADDR") || !strcmp(mib, "WLAN_MACAC5_ADDR") || !strcmp(mib, "DEF_WLAN_MACAC5_ADDR"))) {
#else

	if ( action==2 && (!strcmp(mib, "WLAN_MACAC_ADDR") || !strcmp(mib, "DEF_WLAN_MACAC_ADDR"))) {
#endif

		if ( strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall") ) {
			showSetACHelp();
			return -1;
		}
		if ( (!strcmp(value[0], "del") && !value[1]) || (!strcmp(value[0], "add") && !value[1]) ) {
			showSetACHelp();
			return -1;
		}
	}
        if ( action==2 && (!strcmp(mib, "WDS_LIST") )) {
                if ( strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall") ) {
                        showSetwdsHelp();
                        return -1;
                }
                if ( (!strcmp(value[0], "del") && !value[1]) || (!strcmp(value[0], "add") && !value[1]) ) {
                        showSetwdsHelp();
                        return -1;
                }
        }


// wireless schedule --------------------------------Vance 2008.11.17
#if defined(_WIRELESS_SCHEDULE_)
	if ( action==2 && (!strcmp(mib, "WLAN_SCH"))) {
		if ( strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall") ) {
			showSetSchHelp();
			return -1;
		}
		if ( (!strcmp(value[0], "del") && !value[1]) || (!strcmp(value[0], "add") && !value[1]) ) {
			showSetSchHelp();
			return -1;
		}
	}
#endif
//-----------------------------------------------------------

// Felix 21011222
#if defined(_PROFILE_LIST_)
	if ( action==2 && (!strcmp(mib, "PROFILE_LIST"))) {
		if ( strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall") ) {
			return -1;
		}
		if ( (!strcmp(value[0], "del") && !value[1]) || (!strcmp(value[0], "add") && !value[1]) ) {
			return -1;
		}
	}
#endif
//-----------------------------------------------------------

	// Static DHCP Lease --------------------------------
	if ( action==2 && (!strcmp(mib, "SDHCP_TBL") || !strcmp(mib, "DEF_SDHCP_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetStaticDhcpHelp();
			return -1;
		}
	}
//---------------------------------------------------
#ifdef WLAN_WDS
        if ( action==2 && (!strcmp(mib, "WLAN_WDS") || !strcmp(mib, "DEF_WLAN_WDS"))) {
                if ( strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall") ) {
                        showSetWdsHelp();
                        return -1;
                }
                if ( (!strcmp(value[0], "del") && !value[1]) || (!strcmp(value[0], "add") && !value[1]) ) {
                        showSetWdsHelp();
                        return -1;
                }
        }
#endif

#ifdef _EZVIEW_
	if ( action==2 && (!strcmp(mib, "EZVIEW_TBL") || !strcmp(mib, "DEF_EZVIEW_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetEZviewHelp();
			return -1;
		}
	}
#endif

#ifdef _STARCRAFT_
	if ( action==2 && (!strcmp(mib, "STARCRAFT_TBL") || !strcmp(mib, "DEF_STARCRAFT_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetSTARCRAFTHelp();
			return -1;
		}
	}
#endif

#ifdef _QoS_MODE_SELECT_
	if ( action==2 && (!strcmp(mib, "WAN_QOSCON_TBL") || !strcmp(mib, "DEF_WAN_QOSCON_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetConHelp();
			return -1;
		}
	}
#endif

#ifdef _VLAN_CONFIG_SUPPORTED_

	if ( action==2 && (!strcmp(mib, "VLANCONFIG_TBL") || !strcmp(mib, "DEF_VLANCONFIG_TBL")) ) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall") && strcmp(value[0], "add1"))) {
			showSetVlanConfigHelp();
			goto error_return;
		}
	}
#endif	


#ifdef HOME_GATEWAY
////////////////////////////////////////////////////////////////////////////////////////////////
	if ( action==2 && (!strcmp(mib, "VSER_TBL") || !strcmp(mib, "DEF_VSER_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetPortFwHelp();
			return -1;
		}
	}
	if ( action==2 && (!strcmp(mib, "DMZ_TBL") || !strcmp(mib, "DEF_DMZ_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "edit") && strcmp(value[0], "delall"))) {
			showSetPortFwHelp();
			return -1;
		}
	}
	if ( action==2 && (!strcmp(mib, "URLB_TBL") || !strcmp(mib, "DEF_URLB_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetPortFwHelp();
			return -1;
		}
	}
	if ( action==2 && (!strcmp(mib, "SROUT_TBL") || !strcmp(mib, "DEF_SROUT_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetPortFwHelp();
			return -1;
		}
	}
	if ( action==2 && (!strcmp(mib, "ACPC_TBL") || !strcmp(mib, "DEF_ACPC_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetPortFwHelp();
			return -1;
		}
	}
////////////////////////////////////////////////////////////////////////////////////////////////
	if ( action==2 && (!strcmp(mib, "PORTFW_TBL") || !strcmp(mib, "DEF_PORTFW_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetPortFwHelp();
			return -1;
		}
	}
	if ( action==2 && (!strcmp(mib, "PORTFILTER_TBL") || !strcmp(mib, "DEF_PORTFILTER_TBL")) ) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetPortFilterHelp();
			return -1;
		}
	}
	if ( action==2 && (!strcmp(mib, "IPFILTER_TBL") || !strcmp(mib, "DEF_IPFILTER_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetIpFilterHelp();
			return -1;
		}
	}
	if ( action==2 && (!strcmp(mib, "MACFILTER_TBL") || !strcmp(mib, "DEF_MACFILTER_TBL")) ) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetMacFilterHelp();
			return -1;
		}
	}
	if ( action==2 && (!strcmp(mib, "TRIGGERPORT_TBL") || !strcmp(mib, "DEF_TRIGGERPORT_TBL")) ) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetTriggerPortHelp();
			return -1;
		}
	}
	if ( action==2 && (!strcmp(mib, "WAN1_QOS_TBL") || !strcmp(mib, "DEF_WAN1_QOS_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetPortFwHelp();
			return -1;
		}
	}
#if defined(_Customer_) | defined(_MPPPOE_)
	if ( action==2 && (!strcmp(mib, "WAN2_QOS_TBL") || !strcmp(mib, "DEF_WAN2_QOS_TBL"))) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
			showSetPortFwHelp();
			return -1;
		}
	}
#endif
//////////////////////////////////////////////////////////////////////////////////////
#else
	if ( action==2 && (!strcmp(mib, "RSER_USR_TBL") || !strcmp(mib, "DEF_RSER_USR_TBL")) ) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
//			showSetTriggerPortHelp();
			return -1;
		}
	}
	if ( action==2 && (!strcmp(mib, "RSER_CLT_TBL") || !strcmp(mib, "DEF_RSER_CLT_TBL")) ) {
		if (!valNum || (strcmp(value[0], "add") && strcmp(value[0], "del") && strcmp(value[0], "delall"))) {
//			showSetTriggerPortHelp();
			return -1;
		}
	}
#endif

	switch (action) {
	case 1: // get

#ifdef PARSE_TXT_FILE
		if ( filename[0] ) {
			if ( parseTxtConfig(filename, &apmib) < 0) {
				printf("Parse text file error!\n");
				return -1;
			}

			if ( !apmib_init(&apmib)) {
				printf("Initialize AP MIB failed!\n");
				return -1;
			}
		}
		else

#endif
		if ( !apmib_init()) {
			printf("Initialize AP MIB failed!\n");
			return -1;
		}

		idx = searchMIB(mib);
		if ( idx == -1 ) {
			showHelp();
			showAllMibName();
			return -1;
		}

		if (!strcmp(mib, "WLAN_MACAC_ADDR"))
			APMIB_GET(MIB_WLAN_AC_NUM, (void *)&num);
// Static DHCP Lease
		else if (!strcmp(mib, "SDHCP_TBL"))
			APMIB_GET(MIB_STATIC_DHCP_NUM, (void *)&num);
//-------------------
#ifdef WLAN_WDS
           else if (!strcmp(mib, "WLAN_WDS"))
                        APMIB_GET(MIB_WLAN_WDS_NUM, (void *)&num);
#endif
		else if (!strcmp(mib, "WDS_LIST"))
                        APMIB_GET(MIB_WDS_NUM, (void *)&num);

#ifdef _EZVIEW_
		else if (!strcmp(mib, "EZVIEW_TBL"))
			APMIB_GET(MIB_EZVIEW_NUM, (void *)&num);
#endif

#ifdef _STARCRAFT_
		else if (!strcmp(mib, "STARCRAFT_TBL"))
			APMIB_GET(MIB_STARCRAFT_NUM, (void *)&num);
#endif

#ifdef _QoS_MODE_SELECT_
		else if (!strcmp(mib, "WAN_QOSCON_TBL"))
			APMIB_GET(MIB_WANCON_QOS_NUM, (void *)&num);
#endif

#ifdef HOME_GATEWAY
///////////////////////////////////////////////////////////////////////////////////////////////
		else if (!strcmp(mib, "VSER_TBL"))
			APMIB_GET(MIB_VSER_NUM, (void *)&num);
		else if (!strcmp(mib, "DMZ_TBL"))
			APMIB_GET(MIB_DMZ_NUM, (void *)&num);
		else if (!strcmp(mib, "SROUT_TBL"))
			APMIB_GET(MIB_SROUT_NUM, (void *)&num);
		else if (!strcmp(mib, "URLB_TBL"))
			APMIB_GET(MIB_URLB_NUM, (void *)&num);
		else if (!strcmp(mib, "ACPC_TBL"))
			APMIB_GET(MIB_ACPC_NUM, (void *)&num);
///////////////////////////////////////////////////////////////////////////////////////////////
		else if (!strcmp(mib, "PORTFW_TBL"))
			APMIB_GET(MIB_PORTFW_NUM, (void *)&num);
		else if (!strcmp(mib, "PORTFILTER_TBL"))
			APMIB_GET(MIB_PORTFILTER_NUM, (void *)&num);
		else if (!strcmp(mib, "IPFILTER_TBL"))
			APMIB_GET(MIB_IPFILTER_NUM, (void *)&num);
		else if (!strcmp(mib, "MACFILTER_TBL"))
			APMIB_GET(MIB_MACFILTER_NUM, (void *)&num);
		else if (!strcmp(mib, "TRIGGERPORT_TBL"))
			APMIB_GET(MIB_TRIGGERPORT_NUM, (void *)&num);
		else if (!strcmp(mib, "WAN1_QOS_TBL"))
			APMIB_GET(MIB_WAN1_QOS_NUM, (void *)&num);
#if defined(_Customer_) | defined(_MPPPOE_)
		else if (!strcmp(mib, "WAN2_QOS_TBL"))
			APMIB_GET(MIB_WAN2_QOS_NUM, (void *)&num);
#endif
///////////////////////////////////////////////////////////////////////////////////////////////
#else
		else if (!strcmp(mib, "RSER_CLT_TBL"))
			APMIB_GET(MIB_RSER_CLT_NUM, (void *)&num);
		else if (!strcmp(mib, "RSER_USR_TBL"))
			APMIB_GET(MIB_RSER_USR_NUM, (void *)&num);
#endif
//vance 2008.11.20
#ifdef _WIRELESS_SCHEDULE_
		else if (!strcmp(mib, "WLAN_SCH"))
					APMIB_GET(MIB_WLAN_SCH_NUM, (void *)&num);
#endif
//Felix 21011222
#ifdef _PROFILE_LIST_
		else if (!strcmp(mib, "PROFILE_LIST"))
					APMIB_GET(MIB_PROFILE_LIST_NUM, (void *)&num);
#endif

#ifdef _VLAN_CONFIG_SUPPORTED_
		else if (!strcmp(mib, "VLANCONFIG_TBL"))
		{
			APMIB_GET(MIB_VLANCONFIG_NUM, (void *)&num);
		}
#endif
		else
			num = 1;

		if (config_area == 1)
			getMIB(mib, hwmib_table[idx].id, hwmib_table[idx].type, num, 1, value);
		else
			getMIB(mib, mib_table[idx].id, mib_table[idx].type, num, 1, value);
		break;

	case 2: // set
		if ( !apmib_init()) {
			printf("Initialize AP MIB failed!\n");
			return -1;
		}
		idx = searchMIB(mib);
		if ( idx == -1 ) {
			showHelp();
			showAllMibName();
			return -1;
		}
		if ( valNum < 1) {
			showHelp();
			return -1;
		}
		if (config_area == 1)
			setMIB(mib, hwmib_table[idx].id, hwmib_table[idx].type, hwmib_table[idx].size, valNum, value);
		else
			setMIB(mib, mib_table[idx].id, mib_table[idx].type, mib_table[idx].size, valNum, value);
		break;

	case 3: // all
		dumpAll();
		break;
#if defined(_SPECIAL_CHAR_FILTER_IN_SCRIPT_)
	case 4: //all 2
		dump_filter_enable=1;
		dumpAll();
		dump_filter_enable=0;
		break;
#endif
	}
normal_return:
	return 0;
error_return:
	vwlan_idxx = 0;
	
	return -1;
}

//////////////////////////////////////////////////////////////////////////////////
static int writeDefault(void)
{
#ifdef __TARGET_BOARD__
	void *ptr;
	if ((ptr=apmib_hwconf()) == NULL)
	{

		resetHWDefault();
	}
	else
		free(ptr);

	if ((ptr=apmib_dsconf()) == NULL)
	{

		resetDFDefault();
	}
	else
		free(ptr);

#endif

	resetDFDefault();
#if defined(_ENVLAN_)
#ifdef _VLAN_CONFIG_SUPPORTED_
	#ifdef __TARGET_BOARD__		//because Realtek VLAN format is diff other , so use "flash set VLANTABLE add1" to defalut VLAN setting
	system("flash set VLANCONFIG_TBL add1 setVlandef");
	#else
	system("./flash-gw set VLANCONFIG_TBL add1 setVlandef");
	#endif
#endif
#endif
//	resetDefault();		//removed by Erwin
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
static int flash_read(char *buf, int offset, int len)
{
	int fh;
	int ok=1;

#ifdef __TARGET_BOARD__
	fh = open(FLASH_DEVICE_NAME, O_RDWR);
#else
	fh = open(FLASH_DEVICE_NAME, O_RDONLY);
#endif
	if ( fh == -1 ){
		return 0;
	}
	lseek(fh, offset, SEEK_SET);
	if ( read(fh, buf, len) != len){
		ok = 0;
	}
	close(fh);

	return ok;
}

///////////////////////////////////////////////////////////////////////////////
static int searchMIB(char *token)
{
	int idx = 0;
	char tmpBuf[100];

	if (!memcmp(token, "HW_", 3))
		config_area = 1;
	else
		config_area = 2;
/*	else if (!memcmp(token, "DEF_", 4)) {	//removed by
		config_area = 2;
		strcpy(tmpBuf, &token[4]);
		strcpy(token, tmpBuf);
	}
	else
		config_area = 3;
*/
	if ( config_area == 1 ) {
		while (hwmib_table[idx].id) {
			if ( !strcmp(hwmib_table[idx].name, token)) {
				return idx;
			}
			idx++;
		}
		return -1;
	}
	else {
		while (mib_table[idx].id) {
			if ( !strcmp(mib_table[idx].name, token)) {
				return idx;
			}
			idx++;
		}
		return -1;
	}
}
///////////////////////////////////////////////////////////////////////////////
#ifdef HOME_GATEWAY
void stuffVal(unsigned char *array_val, char *buf, char *tmp1, char *scIpTmp, char *dtIpTmp, char *scPtTmp, char *dtPtTmp)
{
	QOS_Tp pEntry=(QOS_Tp)array_val;
	int j;
	char trafType[5][5] = {"None","SMTP","HTTP","POP3","FTP"};

	memset(scPtTmp, '\0', sizeof(scPtTmp));
	memset(dtPtTmp, '\0', sizeof(dtPtTmp));
	//sprintf(tmp1, "%s", inet_ntoa(*((struct in_addr*)pEntry->sourStIp)));
	//sprintf(scIpTmp, "%s-%s", tmp1, inet_ntoa(*((struct in_addr*)pEntry->sourEdIp)));
	//sprintf(tmp1, "%s", inet_ntoa(*((struct in_addr*)pEntry->destStIp)));
	//sprintf(dtIpTmp, "%s-%s", tmp1, inet_ntoa(*((struct in_addr*)pEntry->destEdIp)));
	sprintf(tmp1, "%s", fixed_inet_ntoa(pEntry->sourStIp));
	sprintf(scIpTmp, "%s-%s", tmp1, fixed_inet_ntoa(pEntry->sourEdIp));
	sprintf(tmp1, "%s", fixed_inet_ntoa(pEntry->destStIp));
	sprintf(dtIpTmp, "%s-%s", tmp1, fixed_inet_ntoa(pEntry->destEdIp));
	strcpy(tmp1,"");

#ifndef _QOS_GUARANTEE_ //vance 2009.04.07
	for(j=0; j<=7; j++) {
		sprintf(buf, "%02x:%02x:%02x:%02x:%02x:%02x", pEntry->sourMac[j][0], pEntry->sourMac[j][1], pEntry->sourMac[j][2], pEntry->sourMac[j][3], pEntry->sourMac[j][4], pEntry->sourMac[j][5]);
		strcat(tmp1, buf);
		if (j != 7)
			strcat(tmp1, "-");
	}
#endif
	for (j=0; j<=4; j++) {
		sprintf(buf, "%d:%d", pEntry->sourPort[j][0], pEntry->sourPort[j][1]);
		strcat(scPtTmp, buf);
		sprintf(buf, "%d:%d", pEntry->destPort[j][0], pEntry->destPort[j][1]);
		strcat(dtPtTmp, buf);
		if (j != 4) {
			strcat(dtPtTmp, "-");
			strcat(scPtTmp, "-");
		}
	}
	sprintf(buf, QOS_FORMAT,
		pEntry->bwidthVal, pEntry->bwidthType, pEntry->downVal, pEntry->downType, pEntry->sourType,
		scIpTmp, tmp1, dtIpTmp, trafType[pEntry->trafType], pEntry->protl, scPtTmp, dtPtTmp);
}
#endif

///////////////////////////////////////////////////////////////////////////////
static void getMIB(char *name, int id, TYPE_T type, int num, int array_separate, char **val)
//static void getMIB(char *name, int id, TYPE_T type, int num, int array_separate)	//Erwin
{
	unsigned char array_val[2048];
	unsigned long longVal;
	struct in_addr ia_val;
	void *value;
	char tmpBuf[1024]={0}, *format=NULL, *buf, tmp1[400];
	char scIpTmp[40], dtIpTmp[40], scPtTmp[60], dtPtTmp[60];
	int int_val, i;
	int index=1, tbl=0;
	char mibName[100]={0};
	unsigned long indexVal;
	int j;
//vance 2008.11.20
#ifdef _WIRELESS_SCHEDULE_
	char WEEKArray[7][10]= {"Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"};
#endif
	if (num ==0)
		goto ret;

//	if (config_area == 2)	//removed by Erwin
//		strcpy(mibName, "DEF_");
	strcat(mibName, name);

getval:
	buf = &tmpBuf[strlen(tmpBuf)];
	switch (type) {
	case BYTE_T:
		value = (void *)&int_val;
		format = DEC_FORMAT;
		break;
	case WORD_T:
		value = (void *)&int_val;
		format = DEC_FORMAT;
		break;
	case IA_T:
		value = (void *)&ia_val;
		format = STR_FORMAT;
		break;
	case BYTE5_T:
		value = (void *)array_val;
		format = BYTE5_FORMAT;
		break;
	case BYTE6_T:
		value = (void *)array_val;
		format = BYTE6_FORMAT;
		break;
	case BYTE8_T:
		value = (void *)array_val;
		format = BYTE8_FORMAT;
		break;
	case BYTE13_T:
		value = (void *)array_val;
		format = BYTE13_FORMAT;
		break;
	case BYTE16_T:
		value = (void *)array_val;
		format = BYTE16_FORMAT;
		break;
	case STRING_T:
		value = (void *)array_val;
		format = STR_FORMAT;
		break;

	case BYTE_ARRAY_T:
		value = (void *)array_val;
		break;

	case DWORD_T:
		value = (void *)&longVal;
		format = LONGDEC_FORMAT;
		break;

#ifdef HOME_GATEWAY
	case INDEX_T://////////////////////////////////////////////////////////////
		value = (void *)&longVal;
		format = LONGDEC_FORMAT;
		break;

	case INDEXDOS_T://////////////////////////////////////////////////////////////
		value = (void *)&longVal;
		format = LONGDEC_FORMAT;
		break;


#endif
#ifdef WLAN_WPA
#if 0
	case BYTE32_T:
		value = (void *)array_val;
		format = HEX_FORMAT;
		break;
#endif
#endif

	case WLAC_ARRAY_T:
#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_)
	case WLAC_ARRAY2_T:
	case WLAC_ARRAY3_T:
	case WLAC_ARRAY4_T:
	case WLAC_ARRAY5_T:
#endif
	case WDSLIST_T:
//vance 2008.11.17
#if defined(_WIRELESS_SCHEDULE_)
	case WLAN_SCH_ARRAY_T:
#endif

//Felix 21011222
#if defined(_PROFILE_LIST_)
	case PROFILE_LIST_ARRAY_T:
#endif

// Static DHCP Lease
	case SDHCP_ARRAY_T:
//-------------------
#ifdef WLAN_WDS
        case WDS_ARRAY_T:
#endif

#ifdef _EZVIEW_
	case EZVIEW_ARRAY_T:
#endif

#ifdef _STARCRAFT_
	case STARCRAFT_ARRAY_T:
#endif

#ifdef _QoS_MODE_SELECT_
	case WAN_QOSCON_ARRAY_T:
#endif

#ifdef HOME_GATEWAY
	case VSER_ARRAY_T:///////////////////////////////////////////////////////////
	case DMZ_ARRAY_T:///////////////////////////////////////////////////////////
	case URLB_ARRAY_T:///////////////////////////////////////////////////////////
	case ACPC_ARRAY_T:///////////////////////////////////////////////////////////
	case SROUT_ARRAY_T:///////////////////////////////////////////////////////////
	case PORTFW_ARRAY_T:
	case IPFILTER_ARRAY_T:
	case PORTFILTER_ARRAY_T:
	case MACFILTER_ARRAY_T:
	case TRIGGERPORT_ARRAY_T:
	case WAN1_QOS_ARRAY_T:
#if defined(_Customer_) | defined(_MPPPOE_)
	case WAN2_QOS_ARRAY_T:
#endif
#else
	case RSER_CLT_ARRAY_T:
	case RSER_USR_ARRAY_T:
#endif
#ifdef _VLAN_CONFIG_SUPPORTED_
	case VLANCONFIG_ARRAY_T:
#endif	
		tbl = 1;
		value = (void *)array_val;
		array_val[0] = index++;
		break;
	default: printf("invalid mib1!\n"); return;
	}
       //joseph
	if ( !APMIB_GET(id, value)) {
		printf("%s:Get MIB failed on \"%s\" id:%d !\n",__func__,name,id);
		return;
	}

	if ( type == IA_T )
		value = inet_ntoa(ia_val);

	if (type == BYTE_T || type == WORD_T)
	{
		sprintf(buf, format, int_val);

	}
	else if ( type == IA_T || type == STRING_T )
		sprintf(buf, format, value);
	else if (type == BYTE5_T) {
		sprintf(buf, format, array_val[0],array_val[1],array_val[2],
			array_val[3],array_val[4],array_val[5]);
		convert_lower(buf);
	}
	else if (type == BYTE6_T ) {
		sprintf(buf, format, array_val[0],array_val[1],array_val[2],
			array_val[3],array_val[4],array_val[5],array_val[6]);
		convert_lower(buf);
	}
	else if (type == BYTE8_T ) {
		sprintf(buf, format, array_val[0],array_val[1],array_val[2],
			array_val[3],array_val[4],array_val[5],array_val[6],
			array_val[7]);
		convert_lower(buf);
	}
	else if (type == BYTE13_T) {
		sprintf(buf, format, array_val[0],array_val[1],array_val[2],
			array_val[3],array_val[4],array_val[5],array_val[6],
			array_val[7],array_val[8],array_val[9],array_val[10],
			array_val[11],array_val[12]);
		convert_lower(buf);
	}
	else if (type == BYTE16_T) {
		sprintf(buf, format, array_val[0],array_val[1],array_val[2],
			array_val[3],array_val[4],array_val[5],array_val[6],
			array_val[7],array_val[8],array_val[9],array_val[10],
			array_val[11],array_val[12],array_val[13],array_val[14],array_val[15]);
		convert_lower(buf);
	}
	else if (type == DWORD_T){
		sprintf(buf, format, longVal);}
#ifdef HOME_GATEWAY
	else if (type == INDEX_T)
		{ // Lance 2003.08.27
			int first=1;
			indexVal=0x0001;
			#if defined(_General_) || defined(_Getnet_)
				int ALGArray_Number=6;
			#else		
				int ALGArray_Number=12;
			#endif
			
			#ifdef _ALGRTSP_
				ALGArray_Number +=1;
			#endif

			#ifdef _PPTP_PASSSTHROUGH_
				ALGArray_Number +=1;
			#endif

			#ifdef _ALGSIP_
				ALGArray_Number +=1;
			#endif
/*
	#ifdef _ALGRTSP_
		#ifdef _PPTP_PASSSTHROUGH_
			for( j=0; j<14 ;j++ )
		#else
			for( j=0; j<13 ;j++ )
		#endif

	#else
		#ifdef _PPTP_PASSSTHROUGH_
			for( j=0; j<13 ;j++ )
		#else
			for( j=0; j<12 ;j++ )
		#endif

	#endif
*/
			for( j=0; j<ALGArray_Number; j++ )
			{
				if (longVal & indexVal)
				{
					if (first == 0)
						strcat(buf,",");
					strcat(buf,ALGArray[j]);
					first = 0;
				}
				indexVal *=2;
			}
		}
	else if (type == INDEXDOS_T)
		{ // Lance 2003.08.27
			int first=1;
			indexVal=0x0001;
			for( j=0; j<=6 ;j++ )
			{
				if (longVal & indexVal)
				{
					if (first == 0)
						strcat(buf,",");
					strcat(buf,DOSArray[j]);
					first = 0;
				}
				indexVal *=2;
			}
		}


#endif
#ifdef WLAN_WPA
#if 0
	else if (type == BYTE32_T) {
		buf[0] = '\0';
		for (i=0; i<32; i++) {
			sprintf(tmp1, format, array_val[i]);
			strcat(buf, tmp1);
		}
		convert_lower(buf);
	}
#endif
#endif

#ifdef _EZVIEW_
	else if (type == EZVIEW_ARRAY_T) {
		EZVIEW_Tp pEntry=(EZVIEW_Tp)array_val;
		sprintf(buf, "%s,%05d:%05d,%05d:%05d,%05d:%05d,%05d:%05d,%05d:%05d,%03d,%02x%02x%02x%02x%02x%02x,%02d,%02d,%d,%d,%d,%d",
			pEntry->DeviceName,
			pEntry->pf_from[0],pEntry->pf_to[0],
			pEntry->pf_from[1],pEntry->pf_to[1],
			pEntry->pf_from[2],pEntry->pf_to[2],
			pEntry->pf_from[3],pEntry->pf_to[3],
			pEntry->pf_from[4],pEntry->pf_to[4],
			pEntry->ipAddr,
//			fixed_inet_ntoa(pEntry->ipAddr),
			pEntry->macAddr[0],
			pEntry->macAddr[1],
			pEntry->macAddr[2],
			pEntry->macAddr[3],
			pEntry->macAddr[4],
			pEntry->macAddr[5],
			pEntry->BlockStartTime,
			pEntry->BlockStopTime,
			(pEntry->QoS%128)/8,
			pEntry->QoS%8,
			pEntry->TypeAndBlock/4,
			pEntry->TypeAndBlock%2
		);
	}

#endif

#ifdef _STARCRAFT_
	else if (type == STARCRAFT_ARRAY_T) {
		STARCRAFT_Tp pEntry=(STARCRAFT_Tp)array_val;
		sprintf(buf, "%d.%d.%d.%d",
			pEntry->ip[0],
			pEntry->ip[1],
			pEntry->ip[2],
			pEntry->ip[3]
		);
	}

#endif

#ifdef _QoS_MODE_SELECT_
	else if (type == WAN_QOSCON_ARRAY_T) {
		QOSCON_Tp pEntry=(QOSCON_Tp)array_val;
//		sprintf(buf, "%s,%lu,%lu",pEntry->connectName,pEntry->ptpbw,pEntry->consize);
		sprintf(buf, "%s,%lu,%lu,%s",pEntry->connectName,pEntry->ptpbw,pEntry->consize,pEntry->connectIP);
	}
#endif
#ifdef _VLAN_CONFIG_SUPPORTED_
	else if (type == VLANCONFIG_ARRAY_T) 
	{
		OPMODE_T opmode=-1;
		int isLan=1;
		apmib_get( MIB_OP_MODE, (void *)&opmode);
		
		VLAN_CONFIG_Tp pEntry=(VLAN_CONFIG_Tp)array_val;

		if(strncmp(pEntry->netIface,"eth1",strlen("eth1")) == 0)
		{			
			if(opmode == WISP_MODE || opmode == BRIDGE_MODE)
				isLan=1;
			else
				isLan=0;
		}
		else if(strncmp("wlan0",pEntry->netIface, strlen(pEntry->netIface)) == 0)
		{						
			if(opmode == WISP_MODE)
				isLan=0;
			else
				isLan=1;
		}
		else
		{						
			isLan=1;
		}

		sprintf(buf, VLANCONFIG_FORMAT, 			
			pEntry->netIface,pEntry->enabled,pEntry->tagged,pEntry->priority,pEntry->cfi, pEntry->vlanId,isLan);
	}		
#endif	

#ifdef HOME_GATEWAY
	else if (type == WAN1_QOS_ARRAY_T) {		//Qos of WAN1
		stuffVal(array_val, buf, tmp1, scIpTmp, dtIpTmp, scPtTmp, dtPtTmp);
	}
#if defined(_Customer_) | defined(_MPPPOE_)
	else if (type == WAN2_QOS_ARRAY_T) {		//Qos of WAN2
		stuffVal(array_val, buf, tmp1, scIpTmp, dtIpTmp, scPtTmp, dtPtTmp);
	}
#endif
////////////////////////////////////////////////////////////////////////////////////////////
	else if (type == VSER_ARRAY_T) {
		PORTFW_Tp pEntry=(PORTFW_Tp)array_val;
		sprintf(buf, PORTFW_FORMAT, fixed_inet_ntoa(pEntry->ipAddr),
			 pEntry->fromPort, pEntry->protoType, pEntry->toPort);
		if ( strlen(pEntry->comment) ) {
			sprintf(tmp1, ",\"%s\"", pEntry->comment);
			strcat(buf, tmp1);
		}
	}
	else if (type == SROUT_ARRAY_T) {
		SROUT_Tp pEntry=(SROUT_Tp)array_val;
		sprintf(buf, "%s,", fixed_inet_ntoa(pEntry->ipAddr));
		strcat(buf, fixed_inet_ntoa(pEntry->maskAddr));
		sprintf(tmp1, ",%s,%d,%d", fixed_inet_ntoa(pEntry->gatewayAddr), pEntry->hopCount, pEntry->interface);
		strcat(buf, tmp1);
	}
	else if (type == DMZ_ARRAY_T) {
		DMZ_Tp pEntry=(DMZ_Tp)array_val;
		sprintf(buf, "%d,%d,%s,", pEntry->type, pEntry->session, fixed_inet_ntoa(pEntry->pipAddr));
		sprintf(tmp1, "%s", fixed_inet_ntoa(pEntry->cipAddr));
		strcat(buf, tmp1);
	}
	else if (type == URLB_ARRAY_T) {
		URLB_Tp pEntry=(URLB_Tp)array_val;
			sprintf(buf, "%s", pEntry->URLBAddr);
	}
	else if (type == ACPC_ARRAY_T)
	{
		static int index=0;
		int first=1;
		char tmp_buf[20]={0};

		ACPC_Tp pEntry=(ACPC_Tp)array_val;
		indexVal=0x0001;
		sprintf(IpAddr, "'|%s:", fixed_inet_ntoa(pEntry->sipAddr));
		strcat(IpAddr, fixed_inet_ntoa(pEntry->eipAddr));
		strcat(IpAddr,"|");
		index++;
		for (j=0; j<=16; j++) {
			if (pEntry->serindex & indexVal ) {
				if( first != 1 ) {
					sprintf(tmp_buf,"\nACPC_TBL%d_%d=",index,j);
					strcat(buf,tmp_buf);
				}
				else
					first = 0;
				strcat(buf,IpAddr);
				strcat(buf,Protocol[j]);
				strcat(buf,"|'");
			}
				indexVal *= 2;
		}
		memset(tmp1,0,sizeof(tmp1));
		if ((pEntry->Port[0][0])!=0)
		{
			if( first != 1 )
			{
				sprintf(tmp_buf,"\nACPC_TBL%d_%d=",index,j);
				strcat(buf,tmp_buf);
			}
			strcat(buf, IpAddr);

			for (j=0;j<5;j++)
			{
				if ((pEntry->Port[j][0])!=0)
				{
					if (pEntry->Port[j][0]!=pEntry->Port[j][1])
						sprintf(tmp1, "%d:%d", pEntry->Port[j][0], pEntry->Port[j][1]);
					else
						sprintf(tmp1, "%d", pEntry->Port[j][0]);
					if(j != 0)
						strcat(buf, ",");
					strcat(buf, tmp1);
				}
			}
			if ((pEntry->Port[0][0])!=0)
			{
				if (pEntry->protoType==1) strcat(buf, "|TCP|'");
				if (pEntry->protoType==2) strcat(buf, "|UDP|'");
				if (pEntry->protoType==3) strcat(buf, "|ALL|'");
			}
		}
	}
////////////////////////////////////////////////////////////////////////////////////////////
	else if (type == PORTFW_ARRAY_T) {
		PORTFW_Tp pEntry=(PORTFW_Tp)array_val;
		sprintf(buf, PORTFW_FORMAT, fixed_inet_ntoa(pEntry->ipAddr),
			 pEntry->fromPort, pEntry->toPort, pEntry->protoType);
		if ( strlen(pEntry->comment) ) {
			sprintf(tmp1, ",\"%s\"", pEntry->comment);
			strcat(buf, tmp1);
		}
	}

	else if (type == PORTFILTER_ARRAY_T) {
		PORTFILTER_Tp pEntry=(PORTFILTER_Tp)array_val;
		sprintf(buf, PORTFILTER_FORMAT,
			 pEntry->fromPort, pEntry->toPort, pEntry->protoType);
		if ( strlen(pEntry->comment) ) {
			sprintf(tmp1, ",\"%s\"", pEntry->comment);
			strcat(buf, tmp1);
		}
	}
	else if (type == IPFILTER_ARRAY_T) {
		IPFILTER_Tp pEntry=(IPFILTER_Tp)array_val;
		sprintf(buf, IPFILTER_FORMAT, fixed_inet_ntoa(pEntry->ipAddr),
			 pEntry->protoType);
		if ( strlen(pEntry->comment) ) {
			sprintf(tmp1, ",\"%s\"", pEntry->comment);
			strcat(buf, tmp1);
		}
	}
	else if (type == MACFILTER_ARRAY_T) {
		MACFILTER_Tp pEntry=(MACFILTER_Tp)array_val;
		sprintf(buf, MACFILTER_COLON_FORMAT, pEntry->macAddr[0],pEntry->macAddr[1],pEntry->macAddr[2],
			 pEntry->macAddr[3],pEntry->macAddr[4],pEntry->macAddr[5]);
		if ( strlen(pEntry->comment) ) {
			sprintf(tmp1, ",\"%s\"", pEntry->comment);
			strcat(buf, tmp1);
		}
	}

	else if (type == TRIGGERPORT_ARRAY_T) {
		TRIGGERPORT_Tp pEntry=(TRIGGERPORT_Tp)array_val;
/*		sprintf(buf, TRIGGERPORT_FORMAT,
			 pEntry->tri_fromPort, pEntry->tri_toPort, pEntry->tri_protoType,
			 pEntry->inc_fromPort[0][0],pEntry->inc_fromPort[0][1],pEntry->inc_fromPort[1][0],
			 pEntry->inc_fromPort[1][1],pEntry->inc_fromPort[2][0],pEntry->inc_fromPort[2][1],
			 pEntry->inc_fromPort[3][0],pEntry->inc_fromPort[3][1],pEntry->inc_fromPort[4][0],
			 pEntry->inc_fromPort[4][1], pEntry->inc_protoType);
		if ( strlen(pEntry->comment) ) {
			sprintf(tmp1, ", %s", pEntry->comment);
			strcat(buf, tmp1);
*/
//------------------------------------------------------------------------------------------
		/*kyle
		sprintf(buf, "%d-%d:",pEntry->tri_fromPort, pEntry->tri_toPort);
		if (pEntry->tri_protoType==1) strcat(buf, "TCP:");
		if (pEntry->tri_protoType==2) strcat(buf, "UDP:");
		if (pEntry->tri_protoType==3) strcat(buf, "ALL:");
		kyle*/


		//sprintf(buf, "%s@", inet_ntoa(*((struct in_addr*)buffer)));
		//struct in_addr ia_value;

		//memcpy( (unsigned char *)&ia_value,pEntry->ip, 4);

		sprintf(buf, "%s@",fixed_inet_ntoa(pEntry->ip));

		memset(tmp1,0,sizeof(tmp1));
			for (j=0 ; j<5 ; j++)
		{
			if ( pEntry->tcp_port[j][0] != 0)
			{
				if ( pEntry->tcp_port[j][0] == pEntry->tcp_port[j][1] )
					sprintf(tmp1,  "%d", pEntry->tcp_port[j][0]);
				else
					sprintf(tmp1,  "%d:%d", pEntry->tcp_port[j][0], pEntry->tcp_port[j][1]);
				if(j != 0)
					strcat(buf, ",");
				strcat(buf, tmp1);
			}
		}
		strcat(buf, "@TCP@");

		for (j=0 ; j<5 ; j++)
		{
			if ( pEntry->udp_port[j][0] != 0)
			{
				if ( pEntry->udp_port[j][0] == pEntry->udp_port[j][1] )
					sprintf(tmp1,  "%d", pEntry->udp_port[j][0]);
				else
					sprintf(tmp1,  "%d:%d", pEntry->udp_port[j][0], pEntry->udp_port[j][1]);
				if(j != 0)
					strcat(buf, ",");
				strcat(buf, tmp1);
			}
		}
		strcat(buf, "@UDP");
		/*kyle
		if (pEntry->inc_protoType==1) strcat(buf, ":TCP");
		if (pEntry->inc_protoType==2) strcat(buf, ":UDP");
		if (pEntry->inc_protoType==3) strcat(buf, ":ALL");
		kyle*/
	}
#else
///////////////////////////////////////////////////////////////////////////////////////
	else if (type == RSER_CLT_ARRAY_T) {
		RSER_CLT_Tp pEntry=(RSER_CLT_Tp)array_val;
		sprintf(buf, "'%s,%s'", fixed_inet_ntoa(pEntry->ipAddr),
			 pEntry->password);
	}
	else if (type == RSER_USR_ARRAY_T) {
		RSER_USR_Tp pEntry=(RSER_USR_Tp)array_val;
		sprintf(buf, "'%s,%s'", pEntry->username, pEntry->password);
	}
#endif
#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_ )
	else if ((type == WLAC_ARRAY_T)||(type == WLAC_ARRAY2_T)||(type == WLAC_ARRAY3_T)||(type == WLAC_ARRAY4_T)||(type == WLAC_ARRAY5_T)){
#else
	else if (type == WLAC_ARRAY_T) {
#endif

		MACFILTER_Tp pEntry=(MACFILTER_Tp)array_val;
		sprintf(buf, MACFILTER_FORMAT, pEntry->macAddr[0],pEntry->macAddr[1],pEntry->macAddr[2],
			 pEntry->macAddr[3],pEntry->macAddr[4],pEntry->macAddr[5]);
		if ( strlen(pEntry->comment) ) {
			sprintf(tmp1, ",\"%s\"", pEntry->comment);
			strcat(buf, tmp1);
		}
	}
        else if (type == WDSLIST_T) {
                MACFILTER_Tp pEntry=(MACFILTER_Tp)array_val;
                sprintf(buf, MACFILTER_FORMAT, pEntry->macAddr[0],pEntry->macAddr[1],pEntry->macAddr[2],
                         pEntry->macAddr[3],pEntry->macAddr[4],pEntry->macAddr[5]);
                if ( strlen(pEntry->comment) ) {
                        sprintf(tmp1, ",\"%s\"", pEntry->comment);
                        strcat(buf, tmp1);
                }
        }

#if defined(_WIRELESS_SCHEDULE_) //vance 2008.11.17
	else if (type == WLAN_SCH_ARRAY_T){
		//printf("run getmib\n");
		SCHEDULE_Tp pEntry=(SCHEDULE_Tp)array_val;
		//sprintf(buf, "%d %d %d %d", pEntry->minute,pEntry->hour,pEntry->day,pEntry->month);

		if(pEntry->minute == 100)
			strcat(buf,"\"*");
		else
			sprintf(buf, "\"%dm", pEntry->minute);

		if(pEntry->hour == 100)
			strcat(buf," *");
		else
			sprintf(buf, "%s %dh",buf, pEntry->hour);

		if(pEntry->day == 100)
			strcat(buf," *");
		else
			sprintf(buf, "%s %dd",buf, pEntry->day);

		if(pEntry->month == 100)
			strcat(buf," *");
		else
			sprintf(buf, "%s %dmo",buf, pEntry->month);


		//use binary type to decide that weekday be selected
		int first=1;
		indexVal=0x0001;
		for( j=0; j<=6 ;j++ )
		{
			if (pEntry->week & indexVal)
			{
				if (first == 0)
					strcat(buf,",");
				else
					strcat(buf," ");
				strcat(buf,WEEKArray[j]);
				//sprintf(buf,"%s%d",buf,j);
				first = 0;
			}
			indexVal *=2;
		}

		sprintf(tmp1, " %u\"", pEntry->command);
		strcat(buf, tmp1);
		//printf("run getmib next\n");
		if ( strlen(pEntry->comment) ) {
			sprintf(tmp1, ",\"%s\"", pEntry->comment);
			strcat(buf, tmp1);
		}
	}
#endif

#if defined(_PROFILE_LIST_) //Felix 21011222
	else if (type == PROFILE_LIST_ARRAY_T){
		//printf("run getmib\n");
		PROFILELIST_Tp pEntry=(PROFILELIST_Tp)array_val;
		//sprintf(buf, "%d %c %c %d %c %c %c", pEntry->profile_num,pEntry->profile_name,pEntry->ssid,pEntry->channel,pEntry->auth,pEntry->encrypt,pEntry->networktype);

		sprintf(buf, "\"%d", pEntry->profile_num);
		sprintf(buf, "%s %s",buf, pEntry->profile_name);
		sprintf(buf, "%s %s",buf, pEntry->ssid);
		sprintf(buf, "%s %d",buf, pEntry->channel);
		sprintf(buf, "%s %d",buf, pEntry->auth);
		sprintf(buf, "%s %d",buf, pEntry->encrypt);
		sprintf(buf, "%s %d",buf, pEntry->networktype);
		sprintf(buf, "%s %d",buf, pEntry->RTS_enable);
		sprintf(buf, "%s %d",buf, pEntry->Fragment_enable);
		sprintf(buf, "%s %d",buf, pEntry->RTS_value);
		sprintf(buf, "%s %d",buf, pEntry->Fragment_value);
		sprintf(buf, "%s %d",buf, pEntry->preamble_type);
		sprintf(buf, "%s %d",buf, pEntry->cam_psm);
		sprintf(buf, "%s %d",buf, pEntry->wep_length);
		sprintf(buf, "%s %d",buf, pEntry->wep_format);
		sprintf(buf, "%s %d",buf, pEntry->wepkey_index);
		if(pEntry->wepkey1 == '*')
			strcat(buf," *");
		else
			sprintf(buf, "%s %s",buf, pEntry->wepkey1);
		if(pEntry->wepkey2 == '*')
			strcat(buf," *");
		else
			sprintf(buf, "%s %s",buf, pEntry->wepkey2);

		if(pEntry->wepkey3 == '*')
			strcat(buf," *");
		else
			sprintf(buf, "%s %s",buf, pEntry->wepkey3);

		if(pEntry->wepkey4 == '*')
			strcat(buf," *");
		else
			sprintf(buf, "%s %s",buf, pEntry->wepkey4);

		if(pEntry->wpapsk == '*')
			strcat(buf," *\"");
		else
			sprintf(buf, "%s %s\"",buf, pEntry->wpapsk);
		//strcat(buf, tmp1);
	}
#endif

// Static DHCP Lease
	else if (type == SDHCP_ARRAY_T) {
		SDHCP_Tp pEntry=(SDHCP_Tp)array_val;
		sprintf(buf, MACFILTER_COLON_FORMAT, pEntry->macAddr[0],pEntry->macAddr[1],pEntry->macAddr[2],
			 pEntry->macAddr[3],pEntry->macAddr[4],pEntry->macAddr[5]);
		strcat(buf, ",");
		strcat(buf, fixed_inet_ntoa(pEntry->ipAddr));

//vance 2008.11.03
#ifdef _PC_DATABASE_
		if ( strlen(pEntry->hostName) ) {
			sprintf(tmp1, ",\"%s\"", pEntry->hostName);
			strcat(buf, tmp1);
		}
		sprintf(tmp1, ",%d", pEntry->groupIdx);
		strcat(buf, tmp1);
#endif
		//if ( strlen(pEntry->comment) ) {
		//	sprintf(tmp1, ",\"%s\"", pEntry->comment);
		//	strcat(buf, tmp1);
		//}
	}
//--------------------
#ifdef WLAN_WDS
        else if (type == WDS_ARRAY_T) {
                WDS_Tp pEntry=(WDS_Tp)array_val;
                sprintf(buf, WDS_FORMAT, pEntry->macAddr[0],pEntry->macAddr[1],pEntry->macAddr[2],
			pEntry->macAddr[3],pEntry->macAddr[4],pEntry->macAddr[5]);
                if ( strlen(pEntry->comment) ) {
                        sprintf(tmp1, ",\"%s\"", pEntry->comment);
                        strcat(buf, tmp1);
                }
        }
#endif


	else if(type == BYTE_ARRAY_T ) {
		int max_chan_num = 0, chan;
#if 0
		max_chan_num = (id == MIB_HW_TX_POWER_CCK)? MAX_CCK_CHAN_NUM: MAX_OFDM_CHAN_NUM ;
		if(val == NULL || val[0]== NULL){
			for(i=0 ;i< max_chan_num ;i++){
				sprintf(tmp1, "%02x", array_val[i]);
				strcat(buf, tmp1);
			}
			convert_lower(buf);
		}
		else{
			if(atoi(val[0]) > max_chan_num || atoi(val[0])<=0){
				printf("invalid channel number\n");
				return;
			}
			sprintf(buf, "%d", *(((unsigned char *)value)+ atoi(val[0])-1) );
		}
#endif
                if((id >= MIB_HW_TX_POWER_CCK_A &&  id <=MIB_HW_TX_POWER_DIFF_OFDM))
                	max_chan_num = MAX_2G_CHANNEL_NUM_MIB;
                else if((id >= MIB_HW_TX_POWER_5G_HT40_1S_A &&  id <=MIB_HW_TX_POWER_5G_DIFF_OFDM))
                	max_chan_num = MAX_5G_CHANNEL_NUM_MIB;
                if(val == NULL || val[0] == NULL){
                	for(i=0;i< max_chan_num;i++){
                        	sprintf(tmp1, "%02x", array_val[i]);
                        	strcat(buf, tmp1);
                        }
                        convert_lower(buf);
                }
                else{
                	chan = atoi(val[0]);
                	if(chan < 1 || chan > max_chan_num){
                		printf("invalid channel number\n");
                		return;
                	}
                        sprintf(buf, "%d", *(((unsigned char *)value)+chan-1) );
                }
	}


/*	else if ( type == BYTE_ARRAY_T ) {
		if ( id == MIB_HW_TX_POWER_CCK || id == MIB_HW_TX_POWER_OFDM ) {
			buf[0] = '\0';
			for (i=0; i<MAX_CHAN_NUM; i++) {
				sprintf(tmp1, "%d", *(((unsigned char *)value)+ i) );
				if ( i+1 != MAX_CHAN_NUM)
					strcat(tmp1, ",");
				strcat(buf, tmp1);
			}
		}
	}
*/
//Kyle Add to filter special character for shell script.
#if defined(_SPECIAL_CHAR_FILTER_IN_SCRIPT_)


	if( dump_filter_enable == 1 &&
	(!strncmp(mibName,"SSID",strlen(mibName)) ||
	!strncmp(mibName,"SSID_1",strlen(mibName)) ||
	!strncmp(mibName,"SSID_2",strlen(mibName)) ||
	!strncmp(mibName,"SSID_3",strlen(mibName)) ||
	!strncmp(mibName,"SSID_4",strlen(mibName)) ||
	!strncmp(mibName,"SSID_5",strlen(mibName)) ||
	!strncmp(mibName,"SSID_6",strlen(mibName)) ||
	!strncmp(mibName,"SSID_7",strlen(mibName)) ||
	!strncmp(mibName,"REPEATER_SSID",strlen(mibName)) ||
	!strncmp(mibName,"STADRV_SSID",strlen(mibName)) ||
	!strncmp(mibName,"WLAN_WPA_PSK",strlen(mibName)) ||
	!strncmp(mibName,"DEVICE_NAME",strlen(mibName))
	)){
		CharFilter(tmpBuf);
		//char tmpBuf2[1024]={0};
		//printf("DEBUG ORG SSID: %s\n",tmpBuf);
		//strncpy(tmpBuf2,tmpBuf+1,strlen(tmpBuf)-2);
		//StrReplace(tmpBuf2,"\"","\\\"");
		//StrReplace(tmpBuf2,"$","\\$");
		//sprintf(tmpBuf,"\"%s\"",tmpBuf2);
		//printf("DEBUG NEW SSID: %s\n",tmpBuf);
	}
#endif
	if (--num > 0) {
		if (!array_separate)
			strcat(tmpBuf, " ");
		else
		{
			if (tbl)
				printf("%s%d=%s\n", mibName, index-1, tmpBuf);
			else
				printf("%s=%s\n", mibName, tmpBuf);
			tmpBuf[0] = '\0';
		}
		goto getval;
	}
ret:
	if (tbl)
	{
		printf("%s%d=%s\n", mibName, index-1, tmpBuf);
	}
	else
		printf("%s=%s\n", mibName, tmpBuf);
}

///////////////////////////////////////////////////////////////////////////////
static void setMIB(char *name, int id, TYPE_T type, int len, int valNum, char **val)
{
	unsigned char key[170];
	struct in_addr ia_val;
	void *value;
	int int_val, i;
	unsigned long longVal;
	MACFILTER_T wlAc;
	WDSList_T wdsLIST;

#ifdef _WIRELESS_SCHEDULE_
	SCHEDULE_T wsch;
#endif

#ifdef _PROFILE_LIST_
	PROFILELIST_T prolist;
#endif

    int max_chan_num, tx_power_cnt=0;
// Static DHCP Lease
	SDHCP_T SDHCP;
//-------------------

#ifdef WLAN_WDS
        WDS_T wds;
#endif

#ifdef _EZVIEW_
	EZVIEW_T ezview;
#endif

#ifdef _STARCRAFT_
	STARCRAFT_T starcraft;
#endif

#ifdef _QoS_MODE_SELECT_
	QOSCON_T QOSCON;
#endif

#ifdef HOME_GATEWAY
	PORTFW_T portFw;
	PORTFILTER_T portFilter;
	IPFILTER_T ipFilter;
	DMZ_T DMZ;//////////////////////////////////////////////////////////
	URLB_T URLB;//////////////////////////////////////////////////////////
	ACPC_T ACPC;//////////////////////////////////////////////////////////
	SROUT_T SROUT;//////////////////////////////////////////////////////////
	MACFILTER_T macFilter;
	TRIGGERPORT_T triggerPort;
	EDIT_T Edit;
#else
	RSER_USR_T rserUsr;
	RSER_CLT_T rserClt;
#endif
#ifdef _VLAN_CONFIG_SUPPORTED_
	VLAN_CONFIG_T vlanConfig_entry;
#endif	
	int entryNum;

	switch (type) {
	
	case BYTE_T:
	case WORD_T:
		int_val = atoi(val[0]);
		value = (void *)&int_val;
		break;

	case IA_T:
		if ( !inet_aton(val[0], &ia_val) ) {
			printf("invalid internet address!\n");
			return;
		}
		value = (void *)&ia_val;
		break;

	case BYTE5_T:
		if ( strlen(val[0])!=10 || !string_to_hex(val[0], key, 10)) {
			printf("invalid value!\n");
			return;
		}
		value = (void *)key;
		break;

	case BYTE6_T:
		if ( strlen(val[0])!=12 || !string_to_hex(val[0], key, 12)) {
			printf("invalid value!\n");
			return;
		}
		value = (void *)key;
		break;

	case BYTE8_T:
		if ( strlen(val[0])!=16 || !string_to_hex(val[0], key, 16)) {
			printf("invalid value!\n");
			return;
		}
		value = (void *)key;
		break;

	case BYTE_ARRAY_T:
#if 0
		if ( (id!=MIB_HW_TX_POWER_CCK) && (id!=MIB_HW_TX_POWER_OFDM)) {
			printf("invalid mib!\n");
			return;
		}
		max_chan_num = (id==MIB_HW_TX_POWER_CCK)? MAX_CCK_CHAN_NUM : MAX_OFDM_CHAN_NUM ;
		for (i=0; i<max_chan_num ; i++) {
			if(val[i] == NULL) break;
			if ( !sscanf(val[i], "%d", &int_val) ) {
				printf("invalid value!\n");
				return;
			}
			key[i] = (unsigned char)int_val;
			tx_power_cnt ++;
		}
		if(tx_power_cnt != 1 && tx_power_cnt !=2 && tx_power_cnt != max_chan_num){
			printf("invalid parameter number\n");
			return;
		}
		if(tx_power_cnt == 1){
			for(i=1 ; i < max_chan_num; i++) {
				key[i] = key[0];
			}
		}
		else if(tx_power_cnt ==2){
			//key[0] is channel number to set
			//key[1] is tx power value
			if(key[0] < 1 || key[0] > max_chan_num){
				printf("invalid channel number\n");
				return;
			}
			key[2] = 0xff ;
		}
#endif
                if(!(id >= MIB_HW_TX_POWER_CCK_A &&  id <=MIB_HW_TX_POWER_DIFF_OFDM) &&
                        !(id >= MIB_HW_TX_POWER_5G_HT40_1S_A &&  id <=MIB_HW_TX_POWER_5G_DIFF_OFDM)){
                                printf("invalid mib!\n");
                                return;
                        }
                if((id >= MIB_HW_TX_POWER_CCK_A &&  id <=MIB_HW_TX_POWER_DIFF_OFDM))
                        max_chan_num = MAX_2G_CHANNEL_NUM_MIB;
                else if((id >= MIB_HW_TX_POWER_5G_HT40_1S_A &&  id <=MIB_HW_TX_POWER_5G_DIFF_OFDM))
                        max_chan_num = MAX_5G_CHANNEL_NUM_MIB;

                        for (i=0; i<max_chan_num ; i++) {
                                if(val[i] == NULL) break;
                                if ( !sscanf(val[i], "%d", &int_val) ) {
                                        printf("invalid value!\n");
                                        return;
                                }
                                key[i+1] = (unsigned char)int_val;
                                tx_power_cnt ++;
                        }
                        if(tx_power_cnt != 1 && tx_power_cnt !=2 && tx_power_cnt != max_chan_num){
                                unsigned char key_tmp[200];
                                memcpy(key_tmp, key+1, tx_power_cnt);
                                APMIB_GET(id, key+1);
                                memcpy(key+1, key_tmp, tx_power_cnt);
                        }
                        if(tx_power_cnt == 1){
                                for(i=1 ; i <= max_chan_num; i++) {
                                        key[i] = key[1];
                                }
                        }
                else if(tx_power_cnt ==2){
                        //key[1] is channel number to set
                        //key[2] is tx power value
                        //key[3] is tx power key for check set mode
                        if(key[1] < 1 || key[1] > max_chan_num){
                                if((key[1]<1) || ((id >= MIB_HW_TX_POWER_CCK_A &&  id <=MIB_HW_TX_POWER_DIFF_OFDM)) ||
                                         ((id >= MIB_HW_TX_POWER_5G_HT40_1S_A &&  id <=MIB_HW_TX_POWER_5G_DIFF_OFDM) && (key[1]>216))){
                                        printf("invalid channel number\n");
                                        return;
                                }
                        }
                        key[3] = 0xff ;
                }
                key[0] = tx_power_cnt;
		value = (void *)key;
		break;

/*	case BYTE_ARRAY_T:
		if ( (id!=MIB_HW_TX_POWER_CCK) && (id!=MIB_HW_TX_POWER_OFDM))  {
			printf("invalid mib!\n");
			return;
		}
		for (i=0; i<MAX_CHAN_NUM; i++) {
			if ( !sscanf(val[i], "%d", &int_val) ) {
				printf("invalid value!\n");
				return;
			}
			key[i] = (unsigned char)int_val;
		}
		value = (void *)key;
		break;
*/
	case DWORD_T:
		longVal = strtoul(val[0], NULL, 0);
		value = (void *)&longVal;
		break;

#ifdef HOME_GATEWAY
	case INDEX_T:////////////////////////////////////////////////////////////////
		longVal = strtoul(val[0], NULL, 0);
		value = (void *)&longVal;
		break;
	case INDEXDOS_T://////////////////////////////////////////////////////////////
		longVal = strtoul(val[0], NULL, 0);
		value = (void *)&longVal;
		break;


#endif
#ifdef WLAN_WPA
#if 0
	case BYTE32_T:
		if ( strlen(val[0])!=64 || !string_to_hex(val[0], key, 64)) {
			printf("invalid value!\n");
			return;
		}
		value = (void *)key;
		break;
#endif
#endif

#ifdef _EZVIEW_
	case EZVIEW_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_EZVIEW_ADD;
			if ( valNum < 7 ) {
				printf("input argument is not enough!\n");
				return;
			}
			strcpy( ezview.DeviceName , val[1]);
			ezview.ipAddr=atoi(val[2]);
/*			if ( !fixed_inet_aton(val[2],ezview.ipAddr)) {
				printf("invalid ip address!\n");
				return;
			}
*/			if ( strlen(val[3])!=12 || !string_to_hex(val[3], ezview.macAddr, 12)) {
				printf("invalid value!\n");
				return;
			}
			ezview.BlockStartTime=atoi(val[4]);
			ezview.BlockStopTime =atoi(val[5]);
			ezview.QoS           =atoi(val[6]);
			ezview.TypeAndBlock  =atoi(val[7]);
		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_EZVIEW_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_EZVIEW_NUM, (void *)&entryNum)) {
				printf("Get EZVIEW entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&ezview) = (char)int_val;
			if ( !APMIB_GET(MIB_EZVIEW, (void *)&ezview)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_EZVIEW_DELALL;

		value = (void *)&ezview;
		break;
#endif

#ifdef _STARCRAFT_
	case STARCRAFT_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_STARCRAFT_ADD;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			fixed_inet_aton(val[1], starcraft.ip);
			/* not support */
		}
		else if ( !strcmp(val[0], "del")) {
			/* not support */
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_STARCRAFT_DELALL;

		value = (void *)&starcraft;
		break;
#endif


#ifdef HOME_GATEWAY
//////////////////////////////////////////////////////////////////////////////////////////////
	case VSER_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_VSER_ADD;
			if ( valNum < 5 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( !fixed_inet_aton(val[1],portFw.ipAddr)) {
				printf("invalid internet address!\n");
				return;
			}
			portFw.fromPort = atoi(val[2]);
			portFw.protoType = atoi(val[3]);
			portFw.toPort = atoi(val[4]);
			if ( valNum > 5)
				strcpy(portFw.comment, val[5]);
			else
				portFw.comment[0] = '\0';

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_VSER_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_VSER_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&portFw) = (char)int_val;
			if ( !APMIB_GET(MIB_VSER, (void *)&portFw)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_VSER_DELALL;

		value = (void *)&portFw;
		break;


	case SROUT_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_SROUT_ADD;
			if ( valNum < 5 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( !fixed_inet_aton(val[1],SROUT.ipAddr)) {
				printf("invalid destination ip address!\n");
				return;
			}
			if ( !fixed_inet_aton(val[2], SROUT.maskAddr)) {
				printf("invalid mask address!\n");
				return;
			}
			if ( !fixed_inet_aton(val[3], SROUT.gatewayAddr)) {
				printf("invalid gateway address!\n");
				return;
			}
			SROUT.hopCount = atoi(val[4]);
			SROUT.interface = atoi(val[5]);

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_SROUT_DEL;
			if ( valNum < 5 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_SROUT_NUM, (void *)&entryNum)) {
				printf("Get static routing entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&SROUT) = (char)int_val;
			if ( !APMIB_GET(MIB_SROUT, (void *)&SROUT)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_SROUT_DELALL;

		value = (void *)&SROUT;
		break;


	case DMZ_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_DMZ_ADD;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( !fixed_inet_aton(val[1],DMZ.pipAddr)) {
				printf("invalid internet address!\n");
				return;
			}
			if ( !fixed_inet_aton(val[2],DMZ.cipAddr)) {
				printf("invalid internet address!\n");
				return;
			}

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_DMZ_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_DMZ_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&DMZ) = (char)int_val;
			if ( !APMIB_GET(MIB_DMZ, (void *)&DMZ)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "edit")) {
			DMZ_T selDMZ;
			id = MIB_DMZ_EDIT;
			if ( valNum < 5 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_DMZ_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&selDMZ) = (char)int_val;
			if ( !APMIB_GET(MIB_DMZ, (void *)&selDMZ)) {
				printf("Get table entry error!");
				return;
			}
			Edit.selEntry = (char *)&selDMZ;
			//*****************************************
			Edit.editEntry = (char *)&DMZ;
			DMZ.type = atoi(val[2]);
			DMZ.session = atoi(val[3]);
			if ( !fixed_inet_aton(val[4], DMZ.pipAddr)) {
				printf("invalid internet address!\n");
				return;
			}
			if ( !fixed_inet_aton(val[5],DMZ.cipAddr)) {
				printf("invalid internet address!\n");
				return;
			}

			value = (void *)&Edit;
			break;
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_DMZ_DELALL;

		value = (void *)&DMZ;
		break;

	case URLB_ARRAY_T:
	
		if ( !strcmp(val[0], "add")) {
			id = MIB_URLB_ADD;
			if ( valNum < 1 ) {
				printf("input argument is not enough!\n");
				return;
			}
				strcpy(URLB.URLBAddr, val[1]);

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_URLB_DEL;
			if ( valNum < 1 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_URLB_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&URLB) = (char)int_val;
			if ( !APMIB_GET(MIB_URLB, (void *)&URLB)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_URLB_DELALL;
		value = (void *)&URLB;
		break;
//---------------------------------------------------------------------------------
	case ACPC_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_ACPC_ADD;
			if ( valNum < 15 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( !fixed_inet_aton(val[1], ACPC.sipAddr)) {
				printf("invalid internet address!\n");
				return;
			}
			if ( !fixed_inet_aton(val[2], ACPC.eipAddr)) {
				printf("invalid internet address!\n");
				return;
			}
			ACPC.Port[0][0] = atoi(val[3]);		ACPC.Port[0][1] = atoi(val[4]);
			ACPC.Port[1][0] = atoi(val[5]);		ACPC.Port[1][1] = atoi(val[6]);
			ACPC.Port[2][0] = atoi(val[7]);		ACPC.Port[2][1] = atoi(val[8]);
			ACPC.Port[3][0] = atoi(val[9]);		ACPC.Port[3][1] = atoi(val[10]);
			ACPC.Port[4][0] = atoi(val[11]);	ACPC.Port[4][1] = atoi(val[12]);
			ACPC.serindex = atol(val[13]);
			ACPC.protoType = atoi(val[14]);
			if ( valNum > 15)
				strcpy(ACPC.comment, val[15]);
			else
				ACPC.comment[0] = '\0';

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_IPFILTER_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_ACPC_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&ACPC) = (char)int_val;
			if ( !APMIB_GET(MIB_ACPC, (void *)&ACPC)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_ACPC_DELALL;

		value = (void *)&ACPC;
		break;

///////////////////////////////////////////////////////////////////////////////////////////////
	case PORTFW_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_PORTFW_ADD;
			if ( valNum < 5 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( !fixed_inet_aton(val[1], portFw.ipAddr)) {
				printf("invalid internet address!\n");
				return;
			}
			portFw.fromPort = atoi(val[2]);
			portFw.toPort = atoi(val[3]);
			portFw.protoType = atoi(val[4]);
			if ( valNum > 5)
				strcpy(portFw.comment, val[5]);
			else
				portFw.comment[0] = '\0';

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_PORTFW_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_PORTFW_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&portFw) = (char)int_val;
			if ( !APMIB_GET(MIB_PORTFW, (void *)&portFw)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_PORTFW_DELALL;

		value = (void *)&portFw;
		break;

	case PORTFILTER_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_PORTFILTER_ADD;
			if ( valNum < 4 ) {
				printf("input argument is not enough!\n");
				return;
			}
			portFilter.fromPort = atoi(val[1]);
			portFilter.toPort = atoi(val[2]);
			portFilter.protoType = atoi(val[3]);
			if ( valNum > 4)
				strcpy(portFilter.comment, val[4]);
			else
				portFilter.comment[0] = '\0';

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_PORTFILTER_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_PORTFILTER_NUM, (void *)&entryNum)) {
				printf("Get port filter entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&portFilter) = (char)int_val;
			if ( !APMIB_GET(MIB_PORTFILTER, (void *)&portFilter)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_PORTFILTER_DELALL;

		value = (void *)&portFilter;
		break;

	case IPFILTER_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_IPFILTER_ADD;
			if ( valNum < 3 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( !fixed_inet_aton(val[1],ipFilter.ipAddr)) {
				printf("invalid internet address!\n");
				return;
			}
			ipFilter.protoType = atoi(val[2]);
			if ( valNum > 3)
				strcpy(ipFilter.comment, val[3]);
			else
				ipFilter.comment[0] = '\0';

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_IPFILTER_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_IPFILTER_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&ipFilter) = (char)int_val;
			if ( !APMIB_GET(MIB_IPFILTER, (void *)&ipFilter)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_IPFILTER_DELALL;

		value = (void *)&ipFilter;
		break;

	case MACFILTER_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_MACFILTER_ADD;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( strlen(val[1])!=12 || !string_to_hex(val[1], wlAc.macAddr, 12)) {
				printf("invalid value!\n");
				return;
			}

			if ( valNum > 2)
				strcpy(macFilter.comment, val[2]);
			else
				macFilter.comment[0] = '\0';

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_MACFILTER_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_MACFILTER_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&macFilter) = (char)int_val;
			if ( !APMIB_GET(MIB_MACFILTER, (void *)&macFilter)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_MACFILTER_DELALL;
		value = (void *)&macFilter;
		break;
		
#ifdef _VLAN_CONFIG_SUPPORTED_
		case VLANCONFIG_ARRAY_T:

		if ( !strcmp(val[0], "add")) {

			id = MIB_VLANCONFIG_ADD;
			if ( valNum < 6 ) {
				printf("input argument is not enough!\n");
				return;
			}
			vlanConfig_entry.enabled=(unsigned char)atoi(val[1]);
			vlanConfig_entry.enabled=(unsigned char)atoi(val[1]);
			sprintf(vlanConfig_entry.netIface, "%s", val[2]);
			vlanConfig_entry.tagged = (unsigned char)atoi(val[3]);
			vlanConfig_entry.priority = (unsigned char)atoi(val[4]);
			vlanConfig_entry.cfi = (unsigned char)atoi(val[5]);
			vlanConfig_entry.vlanId = (unsigned short)atoi(val[6]);
		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_VLANCONFIG_DEL;
			if ( valNum < 8 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]); //index of entry
			if ( !APMIB_GET(MIB_VLANCONFIG_NUM, (void *)&entryNum)) {
				printf("Get VLAN config entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&vlanConfig_entry) = (char)int_val;
			if ( !APMIB_GET(MIB_VLANCONFIG, (void *)&vlanConfig_entry)) {
				printf("Get table entry error!");
				return;
			}
			vlanConfig_entry.enabled=0;

		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_VLANCONFIG_DELALL;
		
		
		else if ( !strcmp(val[0], "add1")) {
		id = MIB_VLANCONFIG_ADD1;
		strcpy(vlanConfig_entry.netIface, val[1]);
		
	//printf("test\n");
		//vlanConfig_entry=(unsigned char)atoi(val[1]);
		//VlanConfigArray[0].netIface
		}
		
		value = (void *)&vlanConfig_entry;
		break;
#endif
	case TRIGGERPORT_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_TRIGGERPORT_ADD;
			if ( valNum < 15 ) {
				printf("input argument is not enough!\n");
				return;
			}

			if ( !fixed_inet_aton(val[1], triggerPort.ip)) {
				printf("invalid internet address!\n");
				return;
			}
			triggerPort.tcp_port[0][0] = atoi(val[2]);
			triggerPort.tcp_port[0][1] = atoi(val[3]);
			triggerPort.tcp_port[1][0] = atoi(val[4]);
			triggerPort.tcp_port[1][1] = atoi(val[5]);
			triggerPort.tcp_port[2][0] = atoi(val[6]);
			triggerPort.tcp_port[2][1] = atoi(val[7]);
			triggerPort.tcp_port[3][0] = atoi(val[8]);
			triggerPort.tcp_port[3][1] = atoi(val[9]);
			triggerPort.tcp_port[4][0] = atoi(val[10]);
			triggerPort.tcp_port[4][1] = atoi(val[11]);

			triggerPort.udp_port[0][0] = atoi(val[12]);
			triggerPort.udp_port[0][1] = atoi(val[13]);
			triggerPort.udp_port[1][0] = atoi(val[14]);
			triggerPort.udp_port[1][1] = atoi(val[15]);
			triggerPort.udp_port[2][0] = atoi(val[16]);
			triggerPort.udp_port[2][1] = atoi(val[17]);
			triggerPort.udp_port[3][0] = atoi(val[18]);
			triggerPort.udp_port[3][1] = atoi(val[29]);
			triggerPort.udp_port[4][0] = atoi(val[20]);
			triggerPort.udp_port[4][1] = atoi(val[21]);
			/*kyle
			triggerPort.tri_fromPort = atoi(val[1]);
			triggerPort.tri_toPort = atoi(val[2]);
			triggerPort.tri_protoType = atoi(val[3]);
			triggerPort.inc_fromPort[0][0] = atoi(val[4]);
			triggerPort.inc_fromPort[0][1] = atoi(val[5]);
			triggerPort.inc_fromPort[1][0] = atoi(val[6]);
			triggerPort.inc_fromPort[1][1] = atoi(val[7]);
			triggerPort.inc_fromPort[2][0] = atoi(val[8]);
			triggerPort.inc_fromPort[2][1] = atoi(val[9]);
			triggerPort.inc_fromPort[3][0] = atoi(val[10]);
			triggerPort.inc_fromPort[3][1] = atoi(val[11]);
			triggerPort.inc_fromPort[4][0] = atoi(val[12]);
			triggerPort.inc_fromPort[4][1] = atoi(val[13]);
			triggerPort.inc_toPort = atoi(val[14]);
			triggerPort.inc_protoType = atoi(val[15]);
			kyle*/
////////////////////////////////////////////////////////////////////////////////////////////
/*          if ( strlen(val[4])!=10 || !string_to_dec(val[4], triggerPort.inc_fromPort, 10)) {
                printf("invalid value!\n");
                return;
            }
            triggerPort.inc_fromPort=strtoul(val[4],NULL,0);
*/
////////////////////////////////////////////////////////////////////////////////////////////

			if ( valNum > 21)
				strcpy(triggerPort.comment, val[22]);
			else
				triggerPort.comment[0] = '\0';

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_TRIGGERPORT_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_TRIGGERPORT_NUM, (void *)&entryNum)) {
				printf("Get trigger-port entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&triggerPort) = (char)int_val;
			if ( !APMIB_GET(MIB_TRIGGERPORT, (void *)&triggerPort)) {
				printf("Get trigger-port table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_TRIGGERPORT_DELALL;

		value = (void *)&triggerPort;
		break;
/////////////////////////////////////////////////////////////////////////////////////////////
#else
	case RSER_USR_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_RSER_USR_ADD;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
				strcpy(rserUsr.username, val[1]);
				strcpy(rserUsr.password, val[2]);

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_RSER_USR_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_RSER_USR_NUM, (void *)&entryNum)) {
				printf("Get Radius Server User entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&rserUsr) = (char)int_val;
			if ( !APMIB_GET(MIB_RSER_USR, (void *)&rserUsr)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_RSER_USR_DELALL;

		value = (void *)&rserUsr;
		break;

	case RSER_CLT_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_RSER_CLT_ADD;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( !fixed_inet_aton(val[1], rserClt.ipAddr)) {
				printf("invalid internet address!\n");
				return;
			}
				strcpy(rserClt.password, val[2]);

		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_RSER_CLT_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_RSER_CLT_NUM, (void *)&entryNum)) {
				printf("Get Radius Server Client entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&rserClt) = (char)int_val;
			if ( !APMIB_GET(MIB_RSER_CLT, (void *)&rserClt)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_RSER_CLT_DELALL;

		value = (void *)&rserClt;
		break;

#endif

	case WLAC_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_WLAN_AC_ADDR_ADD;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( strlen(val[1])!=12 || !string_to_hex(val[1], wlAc.macAddr, 12)) {
				printf("invalid value!\n");
				return;
			}

			if ( valNum > 2)
				strcpy(wlAc.comment, val[2]);
			else
				wlAc.comment[0] = '\0';
		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_WLAN_AC_ADDR_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_WLAN_AC_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&wlAc) = (char)int_val;
			if ( !APMIB_GET(MIB_WLAN_AC_ADDR, (void *)&wlAc)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_WLAN_AC_ADDR_DELALL;
		value = (void *)&wlAc;
		break;


#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_ )
	case WLAC_ARRAY2_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_WLAN_AC2_ADDR_ADD;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( strlen(val[1])!=12 || !string_to_hex(val[1], wlAc.macAddr, 12)) {
				printf("invalid value!\n");
				return;
			}

			if ( valNum > 2)
				strcpy(wlAc.comment, val[2]);
			else
				wlAc.comment[0] = '\0';
		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_WLAN_AC2_ADDR_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_WLAN_AC2_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&wlAc) = (char)int_val;
			if ( !APMIB_GET(MIB_WLAN_AC2_ADDR, (void *)&wlAc)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_WLAN_AC2_ADDR_DELALL;
		value = (void *)&wlAc;
		break;
	case WLAC_ARRAY3_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_WLAN_AC3_ADDR_ADD;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( strlen(val[1])!=12 || !string_to_hex(val[1], wlAc.macAddr, 12)) {
				printf("invalid value!\n");
				return;
			}

			if ( valNum > 2)
				strcpy(wlAc.comment, val[2]);
			else
				wlAc.comment[0] = '\0';
		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_WLAN_AC3_ADDR_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_WLAN_AC3_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&wlAc) = (char)int_val;
			if ( !APMIB_GET(MIB_WLAN_AC3_ADDR, (void *)&wlAc)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_WLAN_AC3_ADDR_DELALL;
		value = (void *)&wlAc;
		break;
	case WLAC_ARRAY4_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_WLAN_AC4_ADDR_ADD;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( strlen(val[1])!=12 || !string_to_hex(val[1], wlAc.macAddr, 12)) {
				printf("invalid value!\n");
				return;
			}

			if ( valNum > 2)
				strcpy(wlAc.comment, val[2]);
			else
				wlAc.comment[0] = '\0';
		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_WLAN_AC4_ADDR_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_WLAN_AC4_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&wlAc) = (char)int_val;
			if ( !APMIB_GET(MIB_WLAN_AC4_ADDR, (void *)&wlAc)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_WLAN_AC4_ADDR_DELALL;
		value = (void *)&wlAc;
		break;
	case WLAC_ARRAY5_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_WLAN_AC5_ADDR_ADD;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( strlen(val[1])!=12 || !string_to_hex(val[1], wlAc.macAddr, 12)) {
				printf("invalid value!\n");
				return;
			}

			if ( valNum > 2)
				strcpy(wlAc.comment, val[2]);
			else
				wlAc.comment[0] = '\0';
		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_WLAN_AC5_ADDR_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_WLAN_AC5_NUM, (void *)&entryNum)) {
				printf("Get port forwarding entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&wlAc) = (char)int_val;
			if ( !APMIB_GET(MIB_WLAN_AC5_ADDR, (void *)&wlAc)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_WLAN_AC5_ADDR_DELALL;
		value = (void *)&wlAc;
		break;

#endif

        case WDSLIST_T:
                if ( !strcmp(val[0], "add")) {
                        id = MIB_WDS_LIST_ADD;
                        if ( valNum < 2 ) {
                                printf("input argument is not enough!\n");
                                return;
                        }
                        if ( strlen(val[1])!=12 || !string_to_hex(val[1], wdsLIST.macAddr, 12)) {
                                printf("invalid value!\n");
                                return;
                        }

                        if ( valNum > 2)
                                strcpy(wdsLIST.comment, val[2]);
                        else
                                wdsLIST.comment[0] = '\0';
                }
                else if ( !strcmp(val[0], "del")) {
                        id = MIB_WDS_LIST_DEL;
                        if ( valNum < 2 ) {
                                printf("input argument is not enough!\n");
                                return;
                        }
                        int_val = atoi(val[1]);
                        if ( !APMIB_GET(MIB_WDS_NUM, (void *)&entryNum)) {
                                printf("Get port forwarding entry number error!");
                                return;
                        }
                        if ( int_val > entryNum ) {
                                printf("Element number is too large!\n");
                                return;
                        }
                        *((char *)&wdsLIST) = (char)int_val;
                        if ( !APMIB_GET(MIB_WDS_LIST, (void *)&wdsLIST)) {
                                printf("Get table entry error!");
                                return;
                        }
                }
                else if ( !strcmp(val[0], "delall"))
                        id = MIB_WDS_LIST_DELALL;
                value = (void *)&wdsLIST;
                break;


#if defined(_WIRELESS_SCHEDULE_) //vance 2008.11.17
	case WLAN_SCH_ARRAY_T:
			if ( !strcmp(val[0], "add")) {
				id = MIB_WLAN_SCH_ADD;

				if ( valNum < 7 ) {
					printf("input argument is not enough!\n");
					return;
				}
				/*if ( strlen(val[1])!=12 || !string_to_hex(val[1], wlAc.macAddr, 12)) {
					printf("invalid value!\n");
					return;
				}*/
				//needful parameters
				if ( valNum >= 7 )
				{
					//if( !strcmp(val[1], "*"))
					wsch.minute=atoi(val[1]);
					wsch.hour=atoi(val[2]);
					wsch.day=atoi(val[3]);
					wsch.month=atoi(val[4]);
					wsch.week=atoi(val[5]);
					wsch.command=atoi(val[6]);
				}

				if ( valNum > 7)
					strcpy(wsch.comment, val[7]);
				else
					wsch.comment[0] = '\0';

			}
			else if ( !strcmp(val[0], "del")) {
				id = MIB_WLAN_SCH_DEL;
				if ( valNum < 2 ) {
					printf("input argument is not enough!\n");
					return;
				}
				int_val = atoi(val[1]);
				if ( !APMIB_GET(MIB_WLAN_SCH_NUM, (void *)&entryNum)) {
					printf("Get port forwarding entry number error!");
					return;
				}
				if ( int_val > entryNum ) {
					printf("Element number is too large!\n");
					return;
				}
				*((char *)&wsch) = (char)int_val;
				if ( !APMIB_GET(MIB_WLAN_SCH, (void *)&wsch)) {
					printf("Get table entry error!");
					return;
				}
			}
			else if ( !strcmp(val[0], "delall"))
				id = MIB_WLAN_SCH_DELALL;
			value = (void *)&wsch;
			break;
#endif

#if defined(_PROFILE_LIST_) //Felix 21011222
	case PROFILE_LIST_ARRAY_T:
			if ( !strcmp(val[0], "add")) {
				id = MIB_PROFILE_LIST_ADD;

				if ( valNum < 8 ) {
					printf("input argument is not enough!\n");
					return;
				}
				if ( valNum >= 22 )
				{
					prolist.profile_num	=	atoi(val[1]);
					strcpy(prolist.profile_name,val[2]);
					strcpy(prolist.ssid,val[3]);
					prolist.channel		=	atoi(val[4]);
					prolist.auth		=	atoi(val[5]);
					prolist.encrypt		=	atoi(val[6]);
					prolist.networktype	=	atoi(val[7]);
					prolist.RTS_enable	=	atoi(val[8]);
					prolist.Fragment_enable	=       atoi(val[9]);
					prolist.RTS_value	=     	atoi(val[10]);
					prolist.Fragment_value	=	atoi(val[11]);
					prolist.preamble_type	=	atoi(val[12]);
					prolist.cam_psm		=	atoi(val[13]);
					prolist.wep_length	=       atoi(val[14]);
					prolist.wep_format      =       atoi(val[15]);
					prolist.wepkey_index	=       atoi(val[16]);
					strcpy(prolist.wepkey1,val[17]);
					strcpy(prolist.wepkey2,val[18]);
					strcpy(prolist.wepkey3,val[19]);
					strcpy(prolist.wepkey4,val[20]);
					strcpy(prolist.wpapsk,val[21]);	
				}

			}
			else if ( !strcmp(val[0], "del")) {
				id = MIB_PROFILE_LIST_DEL;
				if ( valNum < 2 ) {
					printf("input argument is not enough!\n");
					return;
				}
				int_val = atoi(val[1]);
				if ( !APMIB_GET(MIB_PROFILE_LIST_NUM, (void *)&entryNum)) {
					printf("Get profile list entry number error!");
					return;
				}
				if ( int_val > entryNum ) {
					printf("Element number is too large!\n");
					return;
				}
				*((char *)&prolist) = (char)int_val;
				if ( !APMIB_GET(MIB_PROFILE_LIST, (void *)&prolist)) {
					printf("Get table entry error!");
					return;
				}
			}
			else if ( !strcmp(val[0], "delall"))
				id = MIB_PROFILE_LIST_DELALL;
			value = (void *)&prolist;
			break;
#endif

// Static DHCP Lease--------------------------------------
	case SDHCP_ARRAY_T:
		if ( !strcmp(val[0], "add")) {
			id = MIB_STATIC_DHCP_ADD;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			if ( strlen(val[1])!=12 || !string_to_hex(val[1], SDHCP.macAddr, 12)) {
				printf("invalid value!\n");
				return;
			}

			if ( !fixed_inet_aton(val[2],SDHCP.ipAddr)) {
				printf("invalid internet address!\n");
				return;
			}
//vance 2008.11.03
#ifdef _PC_DATABASE_
			if ( valNum > 2)
				strcpy(SDHCP.hostName, val[3]);
			else
				SDHCP.hostName[0] = '\0';

			if ( valNum > 3)
				SDHCP.groupIdx = atoi(val[4]);
			else
				SDHCP.groupIdx = 0;
#endif
			//if ( valNum > 2)
				//strcpy(wlAc.comment, val[2]);
			//else
				//wlAc.comment[0] = '\0';
		}
		else if ( !strcmp(val[0], "del")) {
			id = MIB_STATIC_DHCP_DEL;
			if ( valNum < 2 ) {
				printf("input argument is not enough!\n");
				return;
			}
			int_val = atoi(val[1]);
			if ( !APMIB_GET(MIB_STATIC_DHCP_NUM, (void *)&entryNum)) {
				printf("Get Static DHCP lease entry number error!");
				return;
			}
			if ( int_val > entryNum ) {
				printf("Element number is too large!\n");
				return;
			}
			*((char *)&SDHCP) = (char)int_val;
			if ( !APMIB_GET(MIB_STATIC_DHCP, (void *)&SDHCP)) {
				printf("Get table entry error!");
				return;
			}
		}
		else if ( !strcmp(val[0], "delall"))
			id = MIB_STATIC_DHCP_DELALL;
		value = (void *)&SDHCP;
		break;
//--------------------------------------------------------
#ifdef WLAN_WDS
        case WDS_ARRAY_T:
                if ( !strcmp(val[0], "add")) {
                        id = MIB_WLAN_WDS_ADD;
                        if ( valNum < 2 ) {
                                printf("input argument is not enough!\n");
                                return;
                        }
                        if ( strlen(val[1])!=12 || !string_to_hex(val[1], wds.macAddr, 12)) {
                                printf("invalid value!\n");
                                return;
                        }
                        if ( valNum > 2)
                                strcpy(wds.comment, val[2]);
                        else
                                wds.comment[0] = '\0';
	        }
                else if ( !strcmp(val[0], "del")) {
                        id = MIB_WLAN_WDS_DEL;
                        if ( valNum < 2 ) {
                                printf("input argument is not enough!\n");
                                return;
                        }
                        int_val = atoi(val[1]);
                        if ( !APMIB_GET(MIB_WLAN_WDS_NUM, (void *)&entryNum)) {
                                printf("Get wds number error!");
                                return;
                        }
                        if ( int_val > entryNum ) {
                                printf("Element number is too large!\n");
                                return;
                        }
                        *((char *)&wds) = (char)int_val;
                        if ( !APMIB_GET(MIB_WLAN_WDS, (void *)&wds)) {
                                printf("Get table entry error!");
                                return;
                        }
                }
                else if ( !strcmp(val[0], "delall"))
                        id = MIB_WLAN_WDS_DELALL;
                value = (void *)&wds;
                break;
#endif // WLAN_WDS

	case BYTE13_T:
		if ( strlen(val[0])!=26 || !string_to_hex(val[0], key, 26)) {
			printf("invalid value!\n");
			return;
		}
		value = (void *)key;
		break;
	case BYTE16_T:
		if ( strlen(val[0])!=32 || !string_to_hex(val[0], key, 32)) {
			printf("invalid value!\n");
			return;
		}
		value = (void *)key;
		break;

	case STRING_T:
		if ( strlen(val[0]) > len) {
			printf("string value too long!\n");
			return;
		}
		value = (void *)val[0];
		break;
	default: printf("invalid mib2!\n"); return;
	}
	if ( !APMIB_SET(id, value))
		printf("set MIB failed!\n");
	if (config_area == 1)
		apmib_update(HW_SETTING);
	else if (config_area == 2)
		apmib_update(DEFAULT_SETTING);
	else
		apmib_update(CURRENT_SETTING);
}


////////////////////////////////////////////////////////////////////////////////
static void dumpAll(void)
{
	int idx=0, num;
	mib_table_entry_T *pTbl;

	if ( !apmib_init()) {
		printf("Initialize AP MIB failed!\n");
		return;
	}

	 config_area=0;

next_tbl:
//	if (++config_area > 3)		//removed by Erwin
	if (++config_area > 2)		//Erwin
	 	return;
	if (config_area == 1)
		pTbl = hwmib_table;
	else
		pTbl = mib_table;

	while (pTbl[idx].id) {
		if ( pTbl[idx].id == MIB_WLAN_AC_ADDR)
			APMIB_GET(MIB_WLAN_AC_NUM, (void *)&num);
#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_ )
		else if ( pTbl[idx].id == MIB_WLAN_AC2_ADDR)
			APMIB_GET(MIB_WLAN_AC2_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_WLAN_AC3_ADDR)
			APMIB_GET(MIB_WLAN_AC3_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_WLAN_AC4_ADDR)
			APMIB_GET(MIB_WLAN_AC4_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_WLAN_AC5_ADDR)
			APMIB_GET(MIB_WLAN_AC5_NUM, (void *)&num);
#endif
                else if ( pTbl[idx].id == MIB_WDS_LIST)
                        APMIB_GET(MIB_WDS_NUM, (void *)&num);
#if defined(_WIRELESS_SCHEDULE_)
		else if ( pTbl[idx].id == MIB_WLAN_SCH)
					APMIB_GET(MIB_WLAN_SCH_NUM, (void *)&num);
#endif

#if defined(_PROFILE_LIST_)

		else if ( pTbl[idx].id == MIB_PROFILE_LIST)
//					APMIB_GET(MIB_PROFILE_LIST_NUM, (void *)&num);
num=0;
#endif

// Static DHCP Lease
		else if ( pTbl[idx].id == MIB_STATIC_DHCP)
			APMIB_GET(MIB_STATIC_DHCP_NUM, (void *)&num);
//-------------------
#ifdef WLAN_WDS
                else if ( pTbl[idx].id == MIB_WLAN_WDS)
                        APMIB_GET(MIB_WLAN_WDS_NUM, (void *)&num);
#endif

#ifdef _EZVIEW_
		else if ( pTbl[idx].id == MIB_EZVIEW)
			APMIB_GET(MIB_EZVIEW_NUM, (void *)&num);
#endif

#ifdef _STARCRAFT_
		else if ( pTbl[idx].id == MIB_STARCRAFT)
			APMIB_GET(MIB_STARCRAFT_NUM, (void *)&num);
#endif

#ifdef _QoS_MODE_SELECT_
		else if ( pTbl[idx].id == MIB_WANCON_QOS)
			APMIB_GET(MIB_WANCON_QOS_NUM, (void *)&num);
#endif

#ifdef _VLAN_CONFIG_SUPPORTED_				
		else if ( pTbl[idx].id == MIB_VLANCONFIG){
			APMIB_GET(MIB_VLANCONFIG_NUM, (void *)&num);
		}
#endif	

#ifdef HOME_GATEWAY
		else if ( pTbl[idx].id == MIB_VSER)
			APMIB_GET(MIB_VSER_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_DMZ)
			APMIB_GET(MIB_DMZ_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_URLB)
			APMIB_GET(MIB_URLB_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_ACPC)
			APMIB_GET(MIB_ACPC_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_SROUT)
			APMIB_GET(MIB_SROUT_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_PORTFW)
			APMIB_GET(MIB_PORTFW_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_PORTFILTER)
			APMIB_GET(MIB_PORTFILTER_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_IPFILTER)
			APMIB_GET(MIB_IPFILTER_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_MACFILTER)
			APMIB_GET(MIB_MACFILTER_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_TRIGGERPORT)
			APMIB_GET(MIB_TRIGGERPORT_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_WAN1_QOS)
			APMIB_GET(MIB_WAN1_QOS_NUM, (void *)&num);
#if defined(_Customer_) | defined(_MPPPOE_)
		else if ( pTbl[idx].id == MIB_WAN2_QOS)
			APMIB_GET(MIB_WAN2_QOS_NUM, (void *)&num);
#endif
#else
		else if ( pTbl[idx].id == MIB_RSER_USR)
			APMIB_GET(MIB_RSER_USR_NUM, (void *)&num);
		else if ( pTbl[idx].id == MIB_RSER_CLT)
			APMIB_GET(MIB_RSER_CLT_NUM, (void *)&num);

#endif
		else
			num = 1;
		if (num >0) {
			getMIB(pTbl[idx].name, pTbl[idx].id, pTbl[idx].type, num, 1, NULL);
		}
		idx++;
	}

	idx = 0;
	goto next_tbl;
}

//////////////////////////////////////////////////////////////////////////////////
static void showHelp(void)
{
	printf("Usage: flash cmd\n");
	printf("option:\n");
	printf("cmd:\n");
	printf("      default -- write flash parameters to default.\n");
	printf("      get mib-name -- get a specific mib from flash memory.\n");
	printf("      set mib-name mib-value -- set a specific mib into flash memory.\n");
	printf("      all -- dump all flash parameters.\n");
	printf("      reset -- reset current setting to default.\n\n");
}


//////////////////////////////////////////////////////////////////////////////////
static void showAllMibName(void)
{
	int idx=0;

	if (config_area == 1) {
		printf("H/W mib-name:\n");
		while (hwmib_table[idx].id) {
			printf("%s\n", mib_table[idx].name);
			idx++;
		}
	}
	else {
		printf("S/W mib-name:\n");
		while (mib_table[idx].id) {
			printf("%s\n", mib_table[idx].name);
			idx++;
		}
	}
}

#ifdef _VLAN_CONFIG_SUPPORTED_
////////////////////////////////////////////////////////////////////////////////////
static void showSetVlanConfigHelp(void)
{
#if 0
	printf("flash set VLAN CONFIG  cmd\n");
	printf("cmd:\n");
	printf("      add enable iface -- update vlan config for specific iface.\n");
	printf("      del entry-number -- delete a vlan config entry.\n");
	printf("      delall -- delete all vlan config entry\n");
#endif		
}
#endif

///////////////////////////////////////////////////////////////////////////////////
static void showSetACHelp(void)
{
	printf("flash set WLAN_MACAC_ADDR cmd\n");
	printf("cmd:\n");
	printf("      add mac-addr, comment -- append a filter mac address.\n");
	printf("      del mac-addr -- delete a filter mac address.\n");
	printf("      delall -- delete all filter mac address.\n");
}

static void showSetwdsHelp(void)
{
        printf("flash set WDS_LIST cmd\n");
        printf("cmd:\n");
        printf("      add mac-addr, comment -- append a filter mac address.\n");
        printf("      del mac-addr -- delete a filter mac address.\n");
        printf("      delall -- delete all filter mac address.\n");
}

//---------------------

#ifdef WLAN_WDS
///////////////////////////////////////////////////////////////////////////////////
static void showSetWdsHelp(void)
{
        printf("flash set WLAN_WDS cmd\n");
        printf("cmd:\n");
        printf("      add mac-addr comment -- append a WDS mac address.\n");
        printf("      del entry-number -- delete a WDS entry.\n");
        printf("      delall -- delete all WDS mac address.\n");
}
#endif

#ifdef _EZVIEW_
static void showSetEZviewHelp(void)
{
	printf("flash set EZVIEW_TBL cmd\n");
	printf("cmd:\n");
	printf("      add DeviceName IP MAC BlockStartTime BlockStopTime UL*8+DL Type*4+Block-- add a record.\n");
	printf("      del entry-number -- delete a record.\n");
	printf("      delall -- delete all records.\n");
}
#endif

#ifdef _STARCRAFT_
static void showSetSTARCRAFTHelp(void)
{
	printf("flash set STARCRAFT_TBL cmd\n");
	printf("cmd:\n");
	printf("      add ---\n");
	printf("      del ---\n");
	printf("      delall ---\n");
}
#endif

#ifdef _QoS_MODE_SELECT_
static void showSetConHelp(void)
{
	printf("flash set QoS_Connection_TBL cmd\n");
	printf("cmd:\n");
	printf("      add rule name connection size and band width in a record.\n");
	printf("      del entry-number -- delete a record.\n");
	printf("      delall -- delete all records.\n");
}
#endif
//////////////////////////////////////////////////////////////////////////////////////////
#ifdef _WIRELESS_SCHEDULE_ //vance 2008.11.17
static void showSetSchHelp(void)
{
	printf("flash set WLAN_Sch cmd\n");
	printf("cmd:\n");
	printf("      add minute hour day month week command comment -- append a schedule.\n");
	printf("      del location  -- delete a schedule.\n");
	printf("      delall -- delete all schedules.\n");
}
#endif
//////////////////////////////////////////////////////////////////////////////////////
#ifdef HOME_GATEWAY
///////////////////////////////////////////////////////////////////////////////////
static void showSetPortFwHelp(void)
{
	printf("flash set PORTFW_TBL cmd\n");
	printf("cmd:\n");
	printf("      add ip from-port to-port protocol comment -- add a filter.\n");
	printf("      del entry-number -- delete a filter.\n");
	printf("      delall -- delete all filter.\n");
}


///////////////////////////////////////////////////////////////////////////////////
static void showSetPortFilterHelp(void)
{
	printf("flash set PORTFILTER_TBL cmd\n");
	printf("cmd:\n");
	printf("      add from-port to-port protocol comment -- add a filter.\n");
	printf("      del entry-number -- delete a filter.\n");
	printf("      delall -- delete all filter.\n");
}

///////////////////////////////////////////////////////////////////////////////////
static void showSetIpFilterHelp(void)
{
	printf("flash set IPFILTER_TBL cmd\n");
	printf("cmd:\n");
	printf("      add ip protocol comment -- add a filter.\n");
	printf("      del entry-number -- delete a filter.\n");
	printf("      delall -- delete all filter.\n");
}

///////////////////////////////////////////////////////////////////////////////////
static void showSetMacFilterHelp(void)
{
	printf("flash set MACFILTER_TBL cmd\n");
	printf("cmd:\n");
	printf("      add mac-addr comment -- add a filter.\n");
	printf("      del entry-number -- delete a filter.\n");
	printf("      delall -- delete all filter.\n");
}
///////////////////////////////////////////////////////////////////////////////////
static void showSetTriggerPortHelp(void)
{
	printf("flash set TRIGGER_PORT cmd\n");
	printf("cmd:\n");
	printf("   add trigger-from trigger-to trigger-proto incoming-from incoming-to incoming-proto comment -- add a trigger-port.\n");
	printf("   del entry-number -- delete a trigger-port.\n");
	printf("   delall -- delete all trigger-port.\n");
}
#endif

#ifdef PARSE_TXT_FILE
////////////////////////////////////////////////////////////////////////////////
static int parseTxtConfig(char *filename, APMIB_Tp pConfig)
{
	char line[100], value[100];
	FILE *fp;
	int id;

	fp = fopen(filename, "r");
	if ( fp == NULL )
		return -1;

	acNum = 0;
// Static DHCP Lease
	SDHCPNum = 0;
//------------------
#ifdef WLAN_WDS
        wdsNum = 0;
#endif

#ifdef _EZVIEW_
	EZVIEWNum=0;
#endif

#ifdef _STARCRAFT_
	STARCRAFTNum=0;
#endif

#ifdef _QoS_MODE_SELECT_
	QosCONNum=0;
#endif

#ifdef HOME_GATEWAY
	portFilterNum = ipFilterNum = macFilterNum = portFwNum = vserNum = DMZNum = URLBNum =ACPCNum =sroutNum = wan1QosNum= 0;
#else
	rserUsrNum = rserCltNum = 0;
#endif
#ifdef _VLAN_CONFIG_SUPPORTED_
 vlanConfigNum=0;	
#endif

	while ( fgets(line, 100, fp) ) {
		id = getToken(line, value);
		if ( id == 0 )
			continue;
		if ( set_mib(pConfig, id, value) < 0) {
			printf("Parse MIB [%d] error!\n", id );
			fclose(fp);
			return -1;
		}
	}

	fclose(fp);
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
static int getToken(char *line, char *value)
{
	char *ptr=line, *p1;
	char token[100]={0};
	int len=0, idx;

	if ( *ptr == ';' )	// comments
		return 0;

	// get token
	while (*ptr && *ptr!=EOL) {
		if ( *ptr == '=' ) {
			memcpy(token, line, len);

			// delete ending space
			for (idx=len-1; idx>=0; idx--) {
				if (token[idx]!= SPACE )
					break;
			}
			token[idx+1] = '\0';
			ptr++;
			break;
		}
		ptr++;
		len++;
	}
	if ( !token[0] )
		return 0;

	// get value
	len=0;
	while (*ptr == SPACE ) ptr++; // delete space

	p1 = ptr;
	while ( *ptr && *ptr!=EOL) {
		ptr++;
		len++;
	}
	memcpy(value, p1, len );
	value[len] = '\0';

	idx = 0;
	while (mib_table[idx].id) {
		if (!strcmp(mib_table[idx].name, token))
			return mib_table[idx].id;
		idx++;
	}
	return 0;
}


////////////////////////////////////////////////////////////////////////////////
static int set_mib(APMIB_Tp pMib, int id, void *value)
{
	unsigned char key[100];
	char *p1, *p2;
#ifdef HOME_GATEWAY
	char *p3, *p4, *p5;
	#if defined(GW_QOS_ENGINE) || ifdef _VLAN_CONFIG_SUPPORTED_ && !defined(QOS_BY_BANDWIDTH)
	char *p6, *p7, *p8, *p9, *p10, *p11, *p12;
	#endif

#if defined(QOS_BY_BANDWIDTH) && !defined(GW_QOS_ENGINE)

#ifdef _VLAN_CONFIG_SUPPORTED_
	char *p6, *p7, *p8;
#else
	char *p6, *p7;
#endif

#endif

#else

#ifdef _VLAN_CONFIG_SUPPORTED_
	char *p3, *p4, *p5, *p6, *p7, *p8;
#endif

#endif

	struct in_addr inAddr, inAddr1, inAddr2;
	int i;
	MACFILTER_Tp pWlAc;
// Static DHCP Lease
	SDHCP_Tp pSDHCP;
//-------------------
#ifdef WLAN_WDS
        WDS_Tp pWds;
#endif

#ifdef _EZVIEW_
        EZVIEW_Tp pEZVIEW;
#endif

#ifdef _STARCRAFT_
        STARCRAFT_Tp pSTARCRAFT;
#endif

#ifdef _QoS_MODE_SELECT_
        QOSCON_Tp pQOSCON;
#endif

#ifdef HOME_GATEWAY
	PORTFW_Tp pPortFw;
	PORTFILTER_Tp pPortFilter;
	IPFILTER_Tp pIpFilter;
	DMZ_Tp pDMZ;//////////////////////////////////////////////////////
	ACPC_Tp pACPC;//////////////////////////////////////////////////////
	URLB_Tp pURLB;//////////////////////////////////////////////////////
	SROUT_Tp pSROUT;//////////////////////////////////////////////////////
	MACFILTER_Tp pMacFilter;
#else
	RSER_USR_T pRserUsr;
	RSER_CLT_T pRserClt;
#endif
#ifdef _VLAN_CONFIG_SUPPORTED_
	VLAN_CONFIG_Tp pVlanConfig;
	int j;
#endif	

	for (i=0; mib_table[i].id; i++) {
		if ( mib_table[i].id == id )
			break;
	}
	if ( mib_table[i].id == 0 )
		return -1;

	switch (mib_table[i].type) {
	case BYTE_T:
		*((unsigned char *)(((long)pMib) + mib_table[i].offset)) = (unsigned char)atoi(value);
		break;

	case WORD_T:
		*((unsigned short *)(((long)pMib) + mib_table[i].offset)) = (unsigned short)atoi(value);
		break;

	case STRING_T:
		if ( strlen(value)+1 > mib_table[i].size )
			return 0;
		strcpy((char *)(((long)pMib) + mib_table[i].offset), (char *)value);
		break;

	case BYTE5_T:
		if (strlen(value)!=10 || !string_to_hex(value, key, 10))
			return -1;
		memcpy((unsigned char *)(((long)pMib) + mib_table[i].offset), key, 5);
		break;

	case BYTE6_T:
		if (strlen(value)!=12 || !string_to_hex(value, key, 12))
			return -1;
		memcpy((unsigned char *)(((long)pMib) + mib_table[i].offset), key, 6);
		break;
	case BYTE8_T:
		if (strlen(value)!=16 || !string_to_hex(value, key, 16))
			return -1;
		memcpy((unsigned char *)(((long)pMib) + mib_table[i].offset), key, 8);
		break;

	case BYTE13_T:
		if (strlen(value)!=26 || !string_to_hex(value, key, 26))
			return -1;
		memcpy((unsigned char *)(((long)pMib) + mib_table[i].offset), key, 13);
		break;

	case BYTE16_T:
		if (strlen(value)!=32 || !string_to_hex(value, key, 32))
			return -1;
		memcpy((unsigned char *)(((long)pMib) + mib_table[i].offset), key, 16);
		break;

	case DWORD_T:
		*((unsigned long *)(((long)pMib) + mib_table[i].offset)) = value;
		break;

#ifdef HOME_GATEWAY
	case INDEX_T://////////////////////////////////////////////////////////////////////////////
		*((unsigned long *)(((long)pMib) + mib_table[i].offset)) = value;
		break;
	case INDEXDOS_T://////////////////////////////////////////////////////////////
		*((unsigned long *)(((long)pMib) + mib_table[i].offset)) = value;
		break;
#endif
#ifdef WLAN_WPA
#if 0
	case BYTE32_T:
		if (strlen(value)!=64 || !string_to_hex(value, key, 64))
			return -1;
		memcpy((unsigned char *)(((long)pMib) + mib_table[i].offset), key, 32);
		break;
#endif
#endif

	case IA_T:
		if ( !inet_aton(value, &inAddr) )
			return -1;
		memcpy((unsigned char *)(((long)pMib) + mib_table[i].offset), (unsigned char *)&inAddr,  4);
		break;

	case WLAC_ARRAY_T:
#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_ )
	case WLAC_ARRAY2_T:
	case WLAC_ARRAY3_T:
	case WLAC_ARRAY4_T:
	case WLAC_ARRAY5_T:
#endif

//vance 2008.11.17
#if defined(_WIRELESS_SCHEDULE_)
	case WLAN_SCH_ARRAY_T:
#endif

//Felix 21011222
#if defined(_PROFILE_LIST_)
	case PROFILE_LIST_ARRAY_T:
#endif

		getVal2((char *)value, &p1, &p2);
		if (p1 == NULL) {
			printf("Invalid WLAC in argument!\n");
			break;
		}
		if (strlen(p1)!=12 || !string_to_hex(p1, key, 12))
			return -1;

		pWlAc = (MACFILTER_Tp)(((long)pMib)+mib_table[i].offset+acNum*sizeof(MACFILTER_T));
		memcpy(pWlAc->macAddr, key, 6);
		if (p2 != NULL )
			strcpy(pWlAc->comment, p2);
		acNum++;
		break;
// Static DHCP Lease -------------------------------------------------------------
	case SDHCP_ARRAY_T:
//vance 2008.11.03
#ifdef _PC_DATABASE_
		getVal4((char *)value, &p1, &p2, &p3, &p4);
#else
		getVal2((char *)value, &p1, &p2);
#endif
		if (p1 == NULL || p2 == NULL) {
			printf("Invalid Static DHCP in argument!\n");
			break;
		}
		if (strlen(p1)!=12 || !string_to_hex(p1, key, 12))
			return -1;

		if ( !inet_aton(p2, &inAddr) )
			return -1;

		pSDHCP = (SDHCP_Tp)(((long)pMib)+mib_table[i].offset+SDHCPNum*sizeof(SDHCP_T));
		memcpy(pSDHCP->macAddr, key, 6);
		memcpy(pSDHCP->ipAddr, (unsigned char *)&inAddr, 4);

//vance 2008.11.03
#ifdef _PC_DATABASE_
		if (p3 != NULL )
			strcpy(pSDHCP->hostName, p3);

		if (p4 != NULL )
			pSDHCP->groupIdx = (unsigned char)atoi(p4);
#endif
		//if (p2 != NULL )
		//	strcpy(pSDHCP->comment, p2);
		SDHCPNum++;
		break;
//--------------------------------------------------------------------------------
#ifdef WLAN_WDS
        case WDS_ARRAY_T:
                getVal2((char *)value, &p1, &p2);
                if (p1 == NULL) {
                        printf("Invalid WDS in argument!\n");
                        break;
                }
                if (strlen(p1)!=12 || !string_to_hex(p1, key, 12))
                        return -1;
                pWds = (WDS_Tp)(((long)pMib)+mib_table[i].offset+wdsNum*sizeof(WDS_T));
                memcpy(pWds->macAddr, key, 6);
                if (p2 != NULL )
                        strcpy(pWds->comment, p2);
                wdsNum++;
                break;
#endif

#ifdef _EZVIEW_
	case EZVIEW_ARRAY_T:
		getVal2((char *)value, &p1, &p2);
		pEZVIEW = (EZVIEW_Tp)(((long)pMib)+mib_table[i].offset+EZVIEWNum*sizeof(EZVIEW_T));
		strcpy(pEZVIEW->DeviceName, p1);
		EZVIEWNum++;
		break;
#endif

#ifdef _STARCRAFT_
	case STARCRAFT_ARRAY_T:
		getVal2((char *)value, &p1, &p2);
		pSTARCRAFT = (STARCRAFT_Tp)(((long)pMib)+mib_table[i].offset+STARCRAFTNum*sizeof(STARCRAFT_T));
		//strcpy(pSTARCRAFT->IP[0], p1);
		STARCRAFTNum++;
		break;
#endif

#ifdef HOME_GATEWAY
	case MACFILTER_ARRAY_T:
		getVal2((char *)value, &p1, &p2);
		if (p1 == NULL) {
			printf("Invalid MACFILTER in argument!\n");
			break;
		}
		if (strlen(p1)!=12 || !string_to_hex(p1, key, 12))
			return -1;

		pMacFilter = (MACFILTER_Tp)(((long)pMib)+mib_table[i].offset+macFilterNum*sizeof(MACFILTER_T));
		memcpy(pMacFilter->macAddr, key, 6);
		if (p2 != NULL )
			strcpy(pMacFilter->comment, p2);
		macFilterNum++;
		break;

//////////////////////////////////////////////////////////////////////////////////////////////
	case VSER_ARRAY_T:
		getVal5((char *)value, &p1, &p2, &p3, &p4, &p5);
		if (p1 == NULL || p2 == NULL || p3 == NULL || p4 == NULL ) {
			printf("Invalid PORTFW arguments!\n");
			break;
		}
		if ( !inet_aton(p1, &inAddr) )
			return -1;

		pPortFw = (PORTFW_Tp)(((long)pMib)+mib_table[i].offset+vserNum*sizeof(PORTFW_T));
		memcpy(pPortFw->ipAddr, (unsigned char *)&inAddr, 4);
		pPortFw->fromPort = (unsigned short)atoi(p2);
		pPortFw->toPort = (unsigned short)atoi(p3);
		pPortFw->protoType = (unsigned char)atoi(p4);
		if ( p5 )
			strcpy( pPortFw->comment, p5 );
		vserNum++;
		break;


	case DMZ_ARRAY_T:
		getVal2((char *)value, &p1, &p2);
		if (p1 == NULL || p2 == NULL) {
			printf("Invalid IPFILTER arguments!\n");
			break;
		}
		if ( !inet_aton(p1, &inAddr) )
			return -1;
		if ( !inet_aton(p2, &inAddr1) )
			return -1;
		pDMZ = (DMZ_Tp)(((long)pMib)+mib_table[i].offset+DMZNum*sizeof(DMZ_T));
		memcpy(pDMZ->pipAddr, (unsigned char *)&inAddr, 4);
		memcpy(pDMZ->cipAddr, (unsigned char *)&inAddr1, 4);
		DMZNum++;
		break;

	case SROUT_ARRAY_T:
		getVal5((char *)value, &p1, &p2, &p3, &p4, &p5);
		if (p1 == NULL || p2 == NULL || p3 == NULL || p4 == NULL || p5 == NULL) {
			printf("Invalid Static Routing arguments!\n");
			break;
		}
		if ( !inet_aton(p1, &inAddr) )
			return -1;
		if ( !inet_aton(p2, &inAddr1) )
			return -1;
		if ( !inet_aton(p3, &inAddr2) )
			return -1;
		pSROUT = (SROUT_Tp)(((long)pMib)+mib_table[i].offset+sroutNum*sizeof(SROUT_T));
		memcpy(pDMZ->ipAddr, (unsigned char *)&inAddr, 4);
		memcpy(pDMZ->maskAddr, (unsigned char *)&inAddr1, 4);
		memcpy(pDMZ->gatewayAddr, (unsigned char *)&inAddr2, 4);
		pSROUT->hopCount = (unsigned short)atoi(p4);
		pSROUT->interface = (unsigned short)atoi(p5);
		sroutNum++;
		break;
//----------------------------------------------------------------------------------------
	case URLB_ARRAY_T:
		getVal1((char *)value, &p1);
		if (p1 == NULL) {
			printf("Invalid URLB in argument!\n");
			break;
		}

		pURLB = (URLB_Tp)(((long)pMib)+mib_table[i].offset+URLBNum*sizeof(URLB_T));
			strcpy(pURLB->URLBAddr, p1);
		URLBNum++;
		break;
//------------------------------------------------------------------------------------------
	case ACPC_ARRAY_T:
/*		getVal15((char *)value, &p1, &p2, &p3, &p4, &p5, &p6, &p7, &p8, &p9, &p10, &p11, &p12, &p13, &p14, &p15);
		if (p1 == NULL || p2 == NULL) {
			printf("Invalid ACPC arguments!\n");
			break;
		}
		if ( !inet_aton(p1, &inAddr) )
			return -1;
		pACPC = (ACPC_Tp)(((long)pMib)+mib_table[i].offset+ACPCNum*sizeof(ACPC_T));
		memcpy(pACPC->sipAddr, (unsigned char *)&inAddr, 4);
		if ( !inet_aton(p2, &inAddr) )
			return -1;
		memcpy(pACPC->eipAddr, (unsigned char *)&inAddr, 4);


	pACPC->Port[0][0] = (unsigned long)atoi(p3);	pACPC->Port[0][1] = (unsigned long)atoi(p4);
	pACPC->Port[1][0] = (unsigned long)atoi(p5);	pACPC->Port[1][1] = (unsigned long)atoi(p6);
	pACPC->Port[2][0] = (unsigned long)atoi(p7);	pACPC->Port[2][1] = (unsigned long)atoi(p8);
	pACPC->Port[3][0] = (unsigned long)atoi(p9);	pACPC->Port[3][1] = (unsigned long)atoi(p10);
	pACPC->Port[4][0] = (unsigned long)atoi(p11);	pACPC->Port[4][1] = (unsigned long)atoi(p12);

		pACPC->serindex = (unsigned char)atol(p13);
		pACPC->protoType = (unsigned char)atoi(p14);
		if ( p15 )
			strcpy( pACPC->comment, p15 );
		ACPCNum++;
*/
		break;
//////////////////////////////////////////////////////////////////////////////////////////////

	case PORTFW_ARRAY_T:
		getVal5((char *)value, &p1, &p2, &p3, &p4, &p5);
		if (p1 == NULL || p2 == NULL || p3 == NULL || p4 == NULL ) {
			printf("Invalid PORTFW arguments!\n");
			break;
		}
		if ( !inet_aton(p1, &inAddr) )
			return -1;

		pPortFw = (PORTFW_Tp)(((long)pMib)+mib_table[i].offset+portFwNum*sizeof(PORTFW_T));
		memcpy(pPortFw->ipAddr, (unsigned char *)&inAddr, 4);
		pPortFw->fromPort = (unsigned short)atoi(p2);
		pPortFw->toPort = (unsigned short)atoi(p3);
		pPortFw->protoType = (unsigned char)atoi(p4);
		if ( p5 )
			strcpy( pPortFw->comment, p5 );
		portFwNum++;
		break;

	case IPFILTER_ARRAY_T:
		getVal3((char *)value, &p1, &p2, &p3);
		if (p1 == NULL || p2 == NULL) {
			printf("Invalid IPFILTER arguments!\n");
			break;
		}
		if ( !inet_aton(p1, &inAddr) )
			return -1;
		pIpFilter = (IPFILTER_Tp)(((long)pMib)+mib_table[i].offset+ipFilterNum*sizeof(IPFILTER_T));
		memcpy(pIpFilter->ipAddr, (unsigned char *)&inAddr, 4);
		pIpFilter->protoType = (unsigned char)atoi(p2);
		if ( p3 )
			strcpy( pIpFilter->comment, p3 );
		ipFilterNum++;
		break;

	case PORTFILTER_ARRAY_T:
		getVal4((char *)value, &p1, &p2, &p3, &p4);
		if (p1 == NULL || p2 == NULL || p3 == NULL) {
			printf("Invalid PORTFILTER arguments!\n");
			break;
		}
		if ( !inet_aton(p1, &inAddr) )
			return -1;
		pPortFilter = (PORTFILTER_Tp)(((long)pMib)+mib_table[i].offset+portFilterNum*sizeof(PORTFILTER_T));
		pPortFilter->fromPort = (unsigned short)atoi(p1);
		pPortFilter->toPort = (unsigned short)atoi(p2);
		pPortFilter->protoType = (unsigned char)atoi(p3);
		if ( p4 )
			strcpy( pPortFilter->comment, p4 );
		portFilterNum++;
		break;
///////////////////////////////////////////////////////////////////////////////////////////
#else
	case RSER_CLT_ARRAY_T:
		getVal2((char *)value, &p1, &p2);
		if (p1 == NULL || p2 == NULL) {
			printf("Invalid RADIUS Server Client arguments!\n");
			break;
		}
		if ( !inet_aton(p1, &inAddr) )
			return -1;

		pRserClt = (RSER_CLT_Tp)(((long)pMib)+mib_table[i].offset+rserCltNum*sizeof(RSER_CLT_T));
		memcpy(pRserClt->ipAddr, (unsigned char *)&inAddr, 4);
		strcpy( pRserClt->password, p2 );
		rserCltNum++;
		break;

	case RSER_USR_ARRAY_T:
		getVal2((char *)value, &p1, &p2);
		if (p1 == NULL || p2 == NULL) {
			printf("Invalid RADIUS Server User arguments!\n");
			break;
		}
		pRserusr = (RSER_USR_Tp)(((long)pMib)+mib_table[i].offset+rserUsrNum*sizeof(RSER_USR_T));
		strcpy( pRserUsr->username, p1 );
		strcpy( pRserUsr->password, p2 );
		rserUsrNum++;
		break;
#endif
#ifdef _VLAN_CONFIG_SUPPORTED_
	case VLANCONFIG_ARRAY_T:
	getVal6((char *)value, &p1, &p2, &p3, &p4, &p5, &p6);
		if (p1 == NULL || p2 == NULL || p3 == NULL || p4 == NULL || p5 == NULL || p6 == NULL ) {
			printf("Invalid VLAN Config arguments!\n");
			break;
		}	
		if (p2){
			pVlanConfig = (VLAN_CONFIG_Tp)(((long)pMib)+mib_table[i].offset);
			for(j=0;j<vlanConfigNum;j++){
			if(!strcmp((pVlanConfig+(j*sizeof(VLAN_CONFIG_T)))->netIface, p2){
				pVlanConfig =  (VLAN_CONFIG_Tp)(((long)pMib)+mib_table[i].offset+(j*sizeof(VLAN_CONFIG_T)));
				pVlanConfig->enabled = (unsigned char)atoi(p1);
				pVlanConfig->tagged = (unsigned char)atoi(p3);
				pVlanConfig->priority = (unsigned char)atoi(p4);
				pVlanConfig->cfi = (unsigned char)atoi(p5);
				pVlanConfig->vlanId = (unsigned short)atoi(p6);
	        	}
	        }
        	}
	break;
#endif	

	default:
		return -1;
	}
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
static void getVal1(char *value, char **p1)
{
	value = getVal(value, p1);
}
//////////////////////////////////////////////////////////////////////////////////
static void getVal2(char *value, char **p1, char **p2)
{
	value = getVal(value, p1);
	if ( value )
		getVal(value, p2);
	else
		*p2 = NULL;
}

#ifdef HOME_GATEWAY
////////////////////////////////////////////////////////////////////////////////
static void getVal3(char *value, char **p1, char **p2, char **p3)
{
	*p1 = *p2 = *p3 = NULL;

	value = getVal(value, p1);
	if ( !value )
		return;
	value = getVal(value, p2);
	if ( !value )
		return;
	getVal(value, p3);
}

////////////////////////////////////////////////////////////////////////////////
static void getVal4(char *value, char **p1, char **p2, char **p3, char **p4)
{
	*p1 = *p2 = *p3 = *p4 = NULL;

	value = getVal(value, p1);
	if ( !value )
		return;
	value = getVal(value, p2);
	if ( !value )
		return;
	value = getVal(value, p3);
	if ( !value )
		return;
	getVal(value, p4);
}
///////////////////////////////////////////////////////////////////////////////////
static void getVal5(char *value, char **p1, char **p2, char **p3, char **p4, char **p5)
{
	*p1 = *p2 = *p3 = *p4 = *p5 = NULL;

	value = getVal(value, p1);
	if ( !value )
		return;
	value = getVal(value, p2);
	if ( !value )
		return;
	value = getVal(value, p3);
	if ( !value )
		return;
	value = getVal(value, p4);
	if ( !value )
		return;
	getVal(value, p5);
}

////////////////////////////////////////////////////////////////////////////////
/*static void getVal5(char *value, char **p1, char **p2, char **p3, char **p4, char **p5)
{
	*p1 = *p2 = *p3 = *p4 = *p5 = NULL;

	value = getVal(value, p1);
	if ( !value )
		return;
	value = getVal(value, p2);
	if ( !value )
		return;
	value = getVal(value, p3);
	if ( !value )
		return;
	value = getVal(value, p4);
	if ( !value )
		return;
	getVal(value, p5);
}*/
////////////////////////////////////////////////////////////////////////////////
/*static void getVal15(char *value, char **p1, char **p2, char **p3, char **p4, char **p5, char **p6, char **p7, char **p8, char **p9, char **p10, char **p11, char **p12, char **p13, char **p14, char **p15)
{
	*p1 = *p2 = *p3 = *p4 = *p5 = *p6 = *p7 = *p8 = *p9 = *p10= *p11 = *p12 = *p13 = *p14 = *p15 = NULL;

	value = getVal(value, p1);
	if ( !value )
		return;
	value = getVal(value, p2);
	if ( !value )
		return;
	value = getVal(value, p3);
	if ( !value )
		return;
	value = getVal(value, p4);
	if ( !value )
		return;
	value = getVal(value, p5);
	if ( !value )
		return;
	value = getVal(value, p6);
	if ( !value )
		return;
	value = getVal(value, p7);
	if ( !value )
		return;
	value = getVal(value, p8);
	if ( !value )
		return;
	value = getVal(value, p9);
	if ( !value )
		return;
	value = getVal(value, p10);
	if ( !value )
		return;
	value = getVal(value, p11);
	if ( !value )
		return;
	value = getVal(value, p12);
	if ( !value )
		return;
	value = getVal(value, p13);
	if ( !value )
		return;
	value = getVal(value, p14);
	if ( !value )
		return;
	getVal(value, p15);
}
*/
#endif // HOME_GATEWAY

#endif // PARSE_TXT_FILE

////////////////////////////////////////////////////////////////////////////////
static int getdir(char *fullname, char *path, int loop)
{
	char tmpBuf[100], *p, *p1;

	strcpy(tmpBuf, fullname);
	path[0] = '\0';

	p1 = tmpBuf;
	while (1) {
		if ((p=strchr(p1, '/'))) {
			if (--loop == 0) {
				*p = '\0';
				strcpy(path, tmpBuf);
				return 0;
			}
			p1 = ++p;
		}
		else
			break;
	}
	return -1;
}

////////////////////////////////////////////////////////////////////////////////
static int read_flash_webpage(char *prefix, char *webfile)
{
	WEB_HEADER_T header;
	char *buf, tmpFile[100], tmpFile1[100], tmpBuf[100];
	int fh=0, i, loop, size;
	FILE_ENTRY_Tp pEntry;
	struct stat sbuf;
	char *file;

	if (webfile[0])
		file = webfile;
	else
		file = NULL;

	if (!file) {
		if ( flash_read((char *)&header, WEB_PAGE_OFFSET, sizeof(header)) == 0) {
			printf("Read web header failed!\n");
			return -1;
		}
	}
	else {
		if ((fh = open(file, O_RDONLY)) < 0) {
			printf("Can't open file %s\n", file);
			return -1;
		}
		lseek(fh, 0L, SEEK_SET);
		if (read(fh, &header, sizeof(header)) != sizeof(header)) {
			printf("Read web header failed %s!\n", file);
			close(fh);
			return -1;
		}
	}
//#ifndef _LITTLE_ENDIAN_
	header.len = DWORD_SWAP(header.len);
//#endif

	if (memcmp(header.signature, _WEB_HEADER_, SIGNATURE_LEN)) {
		printf("Invalid web image!\n");
		return -1;
	}

	buf = malloc(header.len);
	if (buf == NULL) {
		sprintf(tmpBuf, "Allocate buffer failed %ld!\n", header.len);
		printf(tmpBuf);
		return -1;
	}

	if (!file) {
		if ( flash_read(buf, WEB_PAGE_OFFSET+sizeof(header), header.len) == 0) {
			printf("Read web image failed!\n");
			return -1;
		}
	}
	else {
		if (read(fh, buf, header.len) != header.len) {
			printf("Read web image failed!\n");
			return -1;
		}
		close(fh);
	}

	if ( !CHECKSUM_OK(buf, header.len) ) {
		printf("Web image invalid!\n");
		free(buf);
		return -1;
	}
	// save to a file
	strcpy(tmpFile, "flashweb.gz");
	fh = open(tmpFile, O_RDWR|O_CREAT|O_TRUNC);
	if (fh == -1) {
		printf("Create output file error %s!\n", tmpFile );
		return -1;
	}
	if ( write(fh, buf, header.len-1) != header.len -1) {
		printf("write file error %s!\n", tmpFile);
		return -1;
	}
	close(fh);
	sync();

	// decompress file
	sprintf(tmpFile1, "%sXXXXXX", tmpFile);
	mkstemp(tmpFile1);

	sprintf(tmpBuf, "gunzip -c %s > %s", tmpFile, tmpFile1);
	system(tmpBuf);

	unlink(tmpFile);
	sync();

	if (stat(tmpFile1, &sbuf) != 0) {
		printf("Stat file error %s!\n", tmpFile1);
		return -1;
	}
	if (sbuf.st_size < sizeof(FILE_ENTRY_T) ) {
		sprintf(tmpBuf, "Invalid decompress file size %ld!\n", sbuf.st_size);
		printf(tmpBuf);
		unlink(tmpFile1);
		return -1;
	}

	buf = malloc(sbuf.st_size);
	if (buf == NULL) {
		sprintf(tmpBuf,"Allocate buffer failed %ld!\n", sbuf.st_size);
		printf(tmpBuf);
		return -1;
	}
	if ((fh = open(tmpFile1, O_RDONLY)) < 0) {
		printf("Can't open file %s\n", tmpFile1);
		return -1;
	}
	lseek(fh, 0L, SEEK_SET);
	if ( read(fh, buf, sbuf.st_size) != sbuf.st_size) {
		printf("Read file error %ld!\n", sbuf.st_size);
		return -1;
	}
	close(fh);
	unlink(tmpFile1);
	sync();
	size = sbuf.st_size;
	for (i=0; i<size; ) {
		pEntry = (FILE_ENTRY_Tp)&buf[i];

//#ifndef _LITTLE_ENDIAN_
		pEntry->size = DWORD_SWAP(pEntry->size);
//#endif

		strcpy(tmpFile, prefix);
		strcat(tmpFile, "/");
		strcat(tmpFile, pEntry->name);

		loop = 0;
		while (1) {
			if (getdir(tmpFile, tmpBuf, ++loop) < 0)
				break;
			if (tmpBuf[0] && stat(tmpBuf, &sbuf) < 0) { // not exist
 				if ( mkdir(tmpBuf, S_IREAD|S_IWRITE) < 0) {
					printf("Create directory %s failed!\n", tmpBuf);
					return -1;
				}
			}
		}

		fh = open(tmpFile, O_RDWR|O_CREAT|O_TRUNC);
		if (fh == -1) {
			printf("Create output file error %s!\n", tmpFile );
			return -1;
		}

		if ( write(fh, &buf[i+sizeof(FILE_ENTRY_T)], pEntry->size) != pEntry->size ) {
			printf("Write file error %s, len=%ld!\n", tmpFile, pEntry->size);
			return -1;
		}
		close(fh);
		// always set execuatble for script file
		chmod(tmpFile,  S_IXUSR);

		i += (pEntry->size + sizeof(FILE_ENTRY_T));
	}

	return 0;
}


static void generateWdsList(void)
{
    unsigned char buf1[1024];
	int i;

    if ( !apmib_init())
	{
    	printf("Initialize AP MIB failed!\n");
        return;
   	}

	printf("WdsList=");

	for(i=0; i<6; i++)
		if(pMib->wlLinkMac1[i] != '\x00')
			break;
	if(i != 6)
		printf("%02x:%02x:%02x:%02x:%02x:%02x;",
				pMib->wlLinkMac1[0], pMib->wlLinkMac1[1], pMib->wlLinkMac1[2],
				pMib->wlLinkMac1[3], pMib->wlLinkMac1[4], pMib->wlLinkMac1[5]);

	for(i=0; i<6; i++)
		if(pMib->wlLinkMac2[i] != '\x00')
				break;
	if(i != 6)
		printf("%02x:%02x:%02x:%02x:%02x:%02x;",
				pMib->wlLinkMac2[0], pMib->wlLinkMac2[1], pMib->wlLinkMac2[2],
				pMib->wlLinkMac2[3], pMib->wlLinkMac2[4], pMib->wlLinkMac2[5]);

	for(i=0; i<6; i++)
		if(pMib->wlLinkMac3[i] != '\x00')
			break;
	if(i != 6)
		printf("%02x:%02x:%02x:%02x:%02x:%02x;",
				pMib->wlLinkMac3[0], pMib->wlLinkMac3[1], pMib->wlLinkMac3[2],
				pMib->wlLinkMac3[3], pMib->wlLinkMac3[4], pMib->wlLinkMac3[5]);

	for(i=0; i<6; i++)
		if(pMib->wlLinkMac4[i] != '\x00')
			break;
	if(i != 6)
		printf("%02x:%02x:%02x:%02x:%02x:%02x;",
						pMib->wlLinkMac4[0], pMib->wlLinkMac4[1], pMib->wlLinkMac4[2],
						pMib->wlLinkMac4[3], pMib->wlLinkMac4[4], pMib->wlLinkMac4[5]);
	printf("\n");
}

static void generateAclList(void)	// RexHua
{
	int i;
	if ( !apmib_init()) {
		printf("Initialize AP MIB failed!\n");
		return;
	}

	printf("AccessControlList0=");

	for (i=0; i< pMib->acNum; i++)
	{
		 printf("%02x:%02x:%02x:%02x:%02x:%02x;",
						 pMib->acAddrArray[i].macAddr[0], pMib->acAddrArray[i].macAddr[1],
						 pMib->acAddrArray[i].macAddr[2], pMib->acAddrArray[i].macAddr[3],
						 pMib->acAddrArray[i].macAddr[4], pMib->acAddrArray[i].macAddr[5]);
	}
	printf("\n");
}

#if defined(_MULTIPLE_WLAN_ACCESS_CONTROL_)
	static void generateAclList2(void)	// Kyle
	{
		int i;
		if ( !apmib_init()) {
			printf("Initialize AP MIB failed!\n");
			return;
		}

		printf("AccessControlList1=");

		for (i=0; i< pMib->acNum2; i++)
		{
			 printf("%02x:%02x:%02x:%02x:%02x:%02x;",
							 pMib->acAddrArray2[i].macAddr[0], pMib->acAddrArray2[i].macAddr[1],
							 pMib->acAddrArray2[i].macAddr[2], pMib->acAddrArray2[i].macAddr[3],
							 pMib->acAddrArray2[i].macAddr[4], pMib->acAddrArray2[i].macAddr[5]);
		}
		printf("\n");
	}
	static void generateAclList3(void)	// Kyle
	{
		int i;
		if ( !apmib_init()) {
			printf("Initialize AP MIB failed!\n");
			return;
		}

		printf("AccessControlList2=");

		for (i=0; i< pMib->acNum3; i++)
		{
			 printf("%02x:%02x:%02x:%02x:%02x:%02x;",
							 pMib->acAddrArray3[i].macAddr[0], pMib->acAddrArray3[i].macAddr[1],
							 pMib->acAddrArray3[i].macAddr[2], pMib->acAddrArray3[i].macAddr[3],
							 pMib->acAddrArray3[i].macAddr[4], pMib->acAddrArray3[i].macAddr[5]);
		}
		printf("\n");
	}
	static void generateAclList4(void)	// Kyle
	{
		int i;
		if ( !apmib_init()) {
			printf("Initialize AP MIB failed!\n");
			return;
		}

		printf("AccessControlList3=");

		for (i=0; i< pMib->acNum4; i++)
		{
			 printf("%02x:%02x:%02x:%02x:%02x:%02x;",
							 pMib->acAddrArray4[i].macAddr[0], pMib->acAddrArray4[i].macAddr[1],
							 pMib->acAddrArray4[i].macAddr[2], pMib->acAddrArray4[i].macAddr[3],
							 pMib->acAddrArray4[i].macAddr[4], pMib->acAddrArray4[i].macAddr[5]);
		}
		printf("\n");
	}
	static void generateAclList5(void)	// Kyle
	{
		int i;
		if ( !apmib_init()) {
			printf("Initialize AP MIB failed!\n");
			return;
		}

		printf("AccessControlList5=");

		for (i=0; i< pMib->acNum4; i++)
		{
			 printf("%02x:%02x:%02x:%02x:%02x:%02x;",
							 pMib->acAddrArray5[i].macAddr[0], pMib->acAddrArray5[i].macAddr[1],
							 pMib->acAddrArray5[i].macAddr[2], pMib->acAddrArray5[i].macAddr[3],
							 pMib->acAddrArray5[i].macAddr[4], pMib->acAddrArray5[i].macAddr[5]);
		}
		printf("\n");
	}


#endif
        static void generatewdsList(void)
        {
                int i;
                if ( !apmib_init()) {
                        printf("Initialize AP MIB failed!\n");
                        return;
                }

                printf("WDSList=");

                for (i=0; i< pMib->WDSNum; i++)
                {
                         printf("%02x:%02x:%02x:%02x:%02x:%02x;",
                                                         pMib->WDSList[i].macAddr[0], pMib->WDSList[i].macAddr[1],
                                                         pMib->WDSList[i].macAddr[2], pMib->WDSList[i].macAddr[3],
                                                         pMib->WDSList[i].macAddr[4], pMib->WDSList[i].macAddr[5]);
                }
                printf("\n");
        }

#if defined(_WIRELESS_SCHEDULE_)
	// vance 2008.11.20
	static void generateSchList(char *sel)
	{
		int i;
		int isel = atoi(sel);
		unsigned char buf[1024];
		unsigned char commandList[6][40]={"setUpWireless 0","setUpWireless 1","SetupPortForwarding 1","SetupPortForwarding 0","SetupUrlBlocking 1","SetupUrlBlocking 0"};
		if ( !apmib_init()) {
			printf("Initialize AP MIB failed!\n");
			return;
		}
			//printf("run getmib\n");
		SCHEDULE_Tp pEntry=&pMib->WlanSchArray[isel];
		//sprintf(buf, "%d %d %d %d", pEntry->minute,pEntry->hour,pEntry->day,pEntry->month);

		//if(pEntry->minute == 100)
		//	strcpy(buf,"*");
		//else
			sprintf(buf, "%d", pEntry->minute);

		//if(pEntry->hour == 100)
		//	strcat(buf," *");
		//else
			sprintf(buf, "%s %d",buf, pEntry->hour);

		//if(pEntry->day == 100)
		//	strcat(buf," *");
		//else
			sprintf(buf, "%s %d",buf, pEntry->day);

		//if(pEntry->month == 100)
		//	strcat(buf," *");
		//else
			sprintf(buf, "%s %d",buf, pEntry->month);


		//use binary type to decide that weekday be selected
		int first=1;
		int indexVal=0x0001;
		int j;
		for( j=0; j<=6 ;j++ )
		{
			if (pEntry->week & indexVal)
			{
				if (first == 0)
					strcat(buf,",");
				else
					strcat(buf," ");
				//strcat(buf,WEEKArray[j]);
				sprintf(buf,"%s%d",buf,j);
				first = 0;
			}
			indexVal *=2;
		}

		sprintf(buf, "%s;/bin/scriptlib_util.sh %s",buf , commandList[pEntry->command]);

		//printf("run getmib next\n");
		if ( strlen(pEntry->comment) ) {
			sprintf(buf, "%s|\"%s\"",buf, pEntry->comment);
		}

		printf("WirelessScheduleList=%s\n",buf);
	}
#endif

////////////////////////////////////////////////////////////////////////////////
static void __inline__ WRITE_WPA_FILE(int fh, unsigned char *buf)
{
	if ( write(fh, buf, strlen(buf)) != strlen(buf) ) {
		printf("Write WPA config file error!\n");
		close(fh);
		exit(1);
	}
}

////////////////////////////////////////////////////////////////////////////////
static void generateWpaConf(char *outputFile, int isWds)
{
	int fh, intVal, encrypt, enable1x, wep;
	unsigned char buf1[1024], buf2[1024];

#if 0
//#ifdef UNIVERSAL_REPEATER
	int isVxd = 0;

	if (strstr(outputFile, "-vxd"))
		isVxd = 1;
#endif

	if ( !apmib_init()) {
		printf("Initialize AP MIB failed!\n");
		return;
	}

	fh = open(outputFile, O_RDWR|O_CREAT|O_TRUNC);
	if (fh == -1) {
		printf("Create WPA config file error!\n");
		return;
	}
	if (!isWds) {
	

#if 0
//#ifdef UNIVERSAL_REPEATER
	if (isVxd && (encrypt == ENCRYPT_WPA2_MIXED)) {
		apmib_get( MIB_AP_MODE, (void *)&intVal);
		if (intVal == AP_MODE || intVal == AP_WDS_MODE)
			encrypt = ENCRYPT_WPA;
	}
#endif
#if defined(_MULTIPLE_WLAN_RS_SETTING_)
	if ( wlan_va_idx != 0 ) {
		switch(wlan_va_idx)
		{
			case 1: apmib_get( MIB_WLAN_ENCRYPT_1, (void *)&encrypt); break;
			case 2: apmib_get( MIB_WLAN_ENCRYPT_2, (void *)&encrypt); break;
			case 3: apmib_get( MIB_WLAN_ENCRYPT_3, (void *)&encrypt); break;
			case 4: apmib_get( MIB_WLAN_ENCRYPT_4, (void *)&encrypt); break;
			case 5: apmib_get( MIB_WLAN_ENCRYPT_5, (void *)&encrypt); break;
			case 6: apmib_get( MIB_WLAN_ENCRYPT_6, (void *)&encrypt); break;
			case 7: apmib_get( MIB_WLAN_ENCRYPT_7, (void *)&encrypt); break;
			default: break;
		}
	}
	else {
		apmib_get( MIB_WLAN_ENCRYPT, (void *)&encrypt);
	}
	sprintf(buf2, "encryption = %d\n", encrypt);
	WRITE_WPA_FILE(fh, buf2);
#else
	apmib_get( MIB_WLAN_ENCRYPT, (void *)&encrypt);
	sprintf(buf2, "encryption = %d\n", encrypt);
	WRITE_WPA_FILE(fh, buf2);
#endif
#if 0
//#ifdef UNIVERSAL_REPEATER
	if (isVxd) {
		if (strstr(outputFile, "wlan0-vxd"))
			apmib_get( MIB_REPEATER_SSID1, (void *)buf1);
		else
			apmib_get( MIB_REPEATER_SSID2, (void *)buf1);
	}
	else
#endif
#if defined(_MULTIPLE_WLAN_RS_SETTING_)
	if ( wlan_va_idx != 0 ) {
		switch(wlan_va_idx)
		{
			case 1: apmib_get( MIB_WLAN_SSID_1, (void *)buf1); break;
			case 2: apmib_get( MIB_WLAN_SSID_2, (void *)buf1); break;
			case 3: apmib_get( MIB_WLAN_SSID_3, (void *)buf1); break;
			case 4: apmib_get( MIB_WLAN_SSID_4, (void *)buf1); break;
			case 5: apmib_get( MIB_WLAN_SSID_5, (void *)buf1); break;
			case 6: apmib_get( MIB_WLAN_SSID_6, (void *)buf1); break;
			case 7: apmib_get( MIB_WLAN_SSID_7, (void *)buf1); break;
			default: break;
		}
	}
	else {
		apmib_get( MIB_WLAN_SSID,  (void *)buf1);
	}
	sprintf(buf2, "ssid = \"%s\"\n", buf1);
	WRITE_WPA_FILE(fh, buf2);
#else
	apmib_get( MIB_WLAN_SSID,  (void *)buf1);
	sprintf(buf2, "ssid = \"%s\"\n", buf1);
	WRITE_WPA_FILE(fh, buf2);
#endif
#if defined(_MULTIPLE_WLAN_RS_SETTING_)
	if ( wlan_va_idx != 0 ) {
		switch(wlan_va_idx)
		{
			case 1: apmib_get( MIB_WLAN_ENABLE_1X_1, (void *)&enable1x); break;
			case 2: apmib_get( MIB_WLAN_ENABLE_1X_2, (void *)&enable1x); break;
			case 3: apmib_get( MIB_WLAN_ENABLE_1X_3, (void *)&enable1x); break;
			case 4: apmib_get( MIB_WLAN_ENABLE_1X_4, (void *)&enable1x); break;
			case 5: apmib_get( MIB_WLAN_ENABLE_1X_5, (void *)&enable1x); break;
			case 6: apmib_get( MIB_WLAN_ENABLE_1X_6, (void *)&enable1x); break;
			case 7: apmib_get( MIB_WLAN_ENABLE_1X_7, (void *)&enable1x); break;
			default: break;
		}
	}
	else {
		apmib_get( MIB_WLAN_ENABLE_1X, (void *)&enable1x);
	}
	sprintf(buf2, "enable1x = %d\n", enable1x);
	WRITE_WPA_FILE(fh, buf2);
#else
	apmib_get( MIB_WLAN_ENABLE_1X, (void *)&enable1x);
	sprintf(buf2, "enable1x = %d\n", enable1x);
	WRITE_WPA_FILE(fh, buf2);
#endif
	apmib_get( MIB_WLAN_ENABLE_MAC_AUTH, (void *)&intVal);
	sprintf(buf2, "enableMacAuth = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_ENABLE_SUPP_NONWPA, (void *)&intVal);
	if (intVal)
		apmib_get( MIB_WLAN_SUPP_NONWPA, (void *)&intVal);

	sprintf(buf2, "supportNonWpaClient = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_WEP, (void *)&wep);
	sprintf(buf2, "wepKey = 1\n");
	WRITE_WPA_FILE(fh, buf2);

	if ( encrypt==1 && enable1x ) {
		if (wep == 1) {
			apmib_get( MIB_WLAN_WEP64_KEY1, (void *)buf1);
			sprintf(buf2, "wepGroupKey = \"%02x%02x%02x%02x%02x\"\n", buf1[0],buf1[1],buf1[2],buf1[3],buf1[4]);
		}
		else {
			apmib_get( MIB_WLAN_WEP128_KEY1, (void *)buf1);
			sprintf(buf2, "wepGroupKey = \"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x\"\n",
				buf1[0],buf1[1],buf1[2],buf1[3],buf1[4],
				buf1[5],buf1[6],buf1[7],buf1[8],buf1[9],
				buf1[10],buf1[11],buf1[12]);
		}
	}
	else
		strcpy(buf2, "wepGroupKey = \"\"\n");
	WRITE_WPA_FILE(fh, buf2);
#if defined(_MULTIPLE_WLAN_RS_SETTING_)
	if ( wlan_va_idx != 0 ) {
		switch(wlan_va_idx)
		{
			case 1: apmib_get( MIB_WLAN_WPA_AUTH_1, (void *)&intVal); break;
			case 2: apmib_get( MIB_WLAN_WPA_AUTH_2, (void *)&intVal); break;
			case 3: apmib_get( MIB_WLAN_WPA_AUTH_3, (void *)&intVal); break;
			case 4: apmib_get( MIB_WLAN_WPA_AUTH_4, (void *)&intVal); break;
			case 5: apmib_get( MIB_WLAN_WPA_AUTH_5, (void *)&intVal); break;
			case 6: apmib_get( MIB_WLAN_WPA_AUTH_6, (void *)&intVal); break;
			case 7: apmib_get( MIB_WLAN_WPA_AUTH_7, (void *)&intVal); break;
			default: break;
		}
	}
	else {
		apmib_get( MIB_WLAN_WPA_AUTH, (void *)&intVal);
	}
#else
	apmib_get( MIB_WLAN_WPA_AUTH, (void *)&intVal);
#endif
	sprintf(buf2, "authentication = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);
#if defined(_MULTIPLE_WLAN_RS_SETTING_)
	if ( wlan_va_idx != 0 ) {
		switch(wlan_va_idx)
		{
			case 1: apmib_get( MIB_WLAN_WPA_CIPHER_SUITE_1, (void *)&intVal); break;
			case 2: apmib_get( MIB_WLAN_WPA_CIPHER_SUITE_2, (void *)&intVal); break;
			case 3: apmib_get( MIB_WLAN_WPA_CIPHER_SUITE_3, (void *)&intVal); break;
			case 4: apmib_get( MIB_WLAN_WPA_CIPHER_SUITE_4, (void *)&intVal); break;
			case 5: apmib_get( MIB_WLAN_WPA_CIPHER_SUITE_5, (void *)&intVal); break;
			case 6: apmib_get( MIB_WLAN_WPA_CIPHER_SUITE_6, (void *)&intVal); break;
			case 7: apmib_get( MIB_WLAN_WPA_CIPHER_SUITE_7, (void *)&intVal); break;
			default: break;
		}
	}
	else {
		apmib_get( MIB_WLAN_WPA_CIPHER_SUITE, (void *)&intVal);
	}
#else
	apmib_get( MIB_WLAN_WPA_CIPHER_SUITE, (void *)&intVal);
#endif
	sprintf(buf2, "unicastCipher = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);
//EDX
	if(encrypt == 6)
		sprintf(buf2, "wpa2UnicastCipher = 3\n");
	else {
//EDX
#if defined(_MULTIPLE_WLAN_RS_SETTING_)
	if ( wlan_va_idx != 0 ) {
		switch(wlan_va_idx)
		{
			case 1: apmib_get( MIB_WLAN_WPA2_CIPHER_SUITE_1, (void *)&intVal); break;
			case 2: apmib_get( MIB_WLAN_WPA2_CIPHER_SUITE_2, (void *)&intVal); break;
			case 3: apmib_get( MIB_WLAN_WPA2_CIPHER_SUITE_3, (void *)&intVal); break;
			case 4: apmib_get( MIB_WLAN_WPA2_CIPHER_SUITE_4, (void *)&intVal); break;
			case 5: apmib_get( MIB_WLAN_WPA2_CIPHER_SUITE_5, (void *)&intVal); break;
			case 6: apmib_get( MIB_WLAN_WPA2_CIPHER_SUITE_6, (void *)&intVal); break;
			case 7: apmib_get( MIB_WLAN_WPA2_CIPHER_SUITE_7, (void *)&intVal); break;
			default: break;
		}
	}
	else {
		apmib_get( MIB_WLAN_WPA2_CIPHER_SUITE, (void *)&intVal);
	}
#else
	apmib_get( MIB_WLAN_WPA2_CIPHER_SUITE, (void *)&intVal);
#endif
	sprintf(buf2, "wpa2UnicastCipher = %d\n", intVal);
	}//EDX
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_WPA2_PRE_AUTH, (void *)&intVal);
	sprintf(buf2, "enablePreAuth = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_WPA_PSK_FORMAT, (void *)&intVal);
	if (intVal==0)
		sprintf(buf2, "usePassphrase = 1\n");
	else
		sprintf(buf2, "usePassphrase = 0\n");
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_WPA_PSK, (void *)buf1);
	sprintf(buf2, "psk = \"%s\"\n", buf1);
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_WPA_GROUP_REKEY_TIME, (void *)&intVal);
	sprintf(buf2, "groupRekeyTime = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);
#if defined(_MULTIPLE_WLAN_RS_SETTING_)
	if ( wlan_va_idx != 0 ) {
		switch(wlan_va_idx)
		{
			case 1: apmib_get( MIB_WLAN_RS_PORT_1, (void *)&intVal); break;
			case 2: apmib_get( MIB_WLAN_RS_PORT_2, (void *)&intVal); break;
			case 3: apmib_get( MIB_WLAN_RS_PORT_3, (void *)&intVal); break;
			case 4: apmib_get( MIB_WLAN_RS_PORT_4, (void *)&intVal); break;
			case 5: apmib_get( MIB_WLAN_RS_PORT_5, (void *)&intVal); break;
			case 6: apmib_get( MIB_WLAN_RS_PORT_6, (void *)&intVal); break;
			case 7: apmib_get( MIB_WLAN_RS_PORT_7, (void *)&intVal); break;
			default: break;
		}
	}
	else {
		apmib_get( MIB_WLAN_RS_PORT, (void *)&intVal);
	}
	sprintf(buf2, "rsPort = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);
#else
	apmib_get( MIB_WLAN_RS_PORT, (void *)&intVal);
	sprintf(buf2, "rsPort = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);
#endif
#if defined(_MULTIPLE_WLAN_RS_SETTING_)
	if ( wlan_va_idx != 0 ) {
		switch(wlan_va_idx)
		{
			case 1: apmib_get( MIB_WLAN_RS_IP_1, (void *)buf1); break;
			case 2: apmib_get( MIB_WLAN_RS_IP_2, (void *)buf1); break;
			case 3: apmib_get( MIB_WLAN_RS_IP_3, (void *)buf1); break;
			case 4: apmib_get( MIB_WLAN_RS_IP_4, (void *)buf1); break;
			case 5: apmib_get( MIB_WLAN_RS_IP_5, (void *)buf1); break;
			case 6: apmib_get( MIB_WLAN_RS_IP_6, (void *)buf1); break;
			case 7: apmib_get( MIB_WLAN_RS_IP_7, (void *)buf1); break;
			default: break;
		}
	}
	else {
		apmib_get( MIB_WLAN_RS_IP, (void *)buf1);
	}
	sprintf(buf2, "rsIP = %s\n", inet_ntoa(*((struct in_addr *)buf1)));
	WRITE_WPA_FILE(fh, buf2);
#else
	apmib_get( MIB_WLAN_RS_IP, (void *)buf1);
	sprintf(buf2, "rsIP = %s\n", inet_ntoa(*((struct in_addr *)buf1)));
	WRITE_WPA_FILE(fh, buf2);
#endif
#if defined(_MULTIPLE_WLAN_RS_SETTING_)
	if ( wlan_va_idx != 0 ) {
		switch(wlan_va_idx)
		{
			case 1: apmib_get( MIB_WLAN_RS_PASSWORD_1, (void *)buf1); break;
			case 2: apmib_get( MIB_WLAN_RS_PASSWORD_2, (void *)buf1); break;
			case 3: apmib_get( MIB_WLAN_RS_PASSWORD_3, (void *)buf1); break;
			case 4: apmib_get( MIB_WLAN_RS_PASSWORD_4, (void *)buf1); break;
			case 5: apmib_get( MIB_WLAN_RS_PASSWORD_5, (void *)buf1); break;
			case 6: apmib_get( MIB_WLAN_RS_PASSWORD_6, (void *)buf1); break;
			case 7: apmib_get( MIB_WLAN_RS_PASSWORD_7, (void *)buf1); break;
			default: break;
		}
	}
	else {
		apmib_get( MIB_WLAN_RS_PASSWORD, (void *)buf1);
	}
	sprintf(buf2, "rsPassword = \"%s\"\n", buf1);
	WRITE_WPA_FILE(fh, buf2);
#else
	apmib_get( MIB_WLAN_RS_PASSWORD, (void *)buf1);
	sprintf(buf2, "rsPassword = \"%s\"\n", buf1);
	WRITE_WPA_FILE(fh, buf2);
#endif
	apmib_get( MIB_WLAN_RS_RETRY, (void *)&intVal);
	sprintf(buf2, "rsMaxReq = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);
#if defined(_MULTIPLE_WLAN_RS_SETTING_)
	if ( wlan_va_idx != 0 ) {
		switch(wlan_va_idx)
		{
			case 1: apmib_get( MIB_WLAN_RS_INTERVAL_TIME_1, (void *)&intVal); break;
			case 2: apmib_get( MIB_WLAN_RS_INTERVAL_TIME_2, (void *)&intVal); break;
			case 3: apmib_get( MIB_WLAN_RS_INTERVAL_TIME_3, (void *)&intVal); break;
			case 4: apmib_get( MIB_WLAN_RS_INTERVAL_TIME_4, (void *)&intVal); break;
			case 5: apmib_get( MIB_WLAN_RS_INTERVAL_TIME_5, (void *)&intVal); break;
			case 6: apmib_get( MIB_WLAN_RS_INTERVAL_TIME_6, (void *)&intVal); break;
			case 7: apmib_get( MIB_WLAN_RS_INTERVAL_TIME_7, (void *)&intVal); break;
			default: break;
		}
	}
	else {
		apmib_get( MIB_WLAN_RS_INTERVAL_TIME, (void *)&intVal);
	}
	sprintf(buf2, "rsAWhile = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);
#else
	apmib_get( MIB_WLAN_RS_INTERVAL_TIME, (void *)&intVal);
	sprintf(buf2, "rsAWhile = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);
#endif
	apmib_get( MIB_WLAN_ACCOUNT_RS_ENABLED, (void *)&intVal);
	sprintf(buf2, "accountRsEnabled = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_ACCOUNT_RS_PORT, (void *)&intVal);
	sprintf(buf2, "accountRsPort = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_ACCOUNT_RS_IP, (void *)buf1);
	sprintf(buf2, "accountRsIP = %s\n", inet_ntoa(*((struct in_addr *)buf1)));
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_ACCOUNT_RS_PASSWORD, (void *)buf1);
	sprintf(buf2, "accountRsPassword = \"%s\"\n", buf1);
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_ACCOUNT_UPDATE_ENABLED, (void *)&intVal);
	sprintf(buf2, "accountRsUpdateEnabled = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_ACCOUNT_UPDATE_DELAY, (void *)&intVal);
	sprintf(buf2, "accountRsUpdateTime = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);

	apmib_get( MIB_WLAN_ACCOUNT_RS_RETRY, (void *)&intVal);
	sprintf(buf2, "accountRsMaxReq = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);
#if defined(_MULTIPLE_WLAN_RS_SETTING_)
	if ( wlan_va_idx != 0 ) {
		switch(wlan_va_idx)
		{
			case 1: apmib_get( MIB_WLAN_ACCOUNT_RS_INTERVAL_TIME_1, (void *)&intVal); break;
			case 2: apmib_get( MIB_WLAN_ACCOUNT_RS_INTERVAL_TIME_2, (void *)&intVal); break;
			case 3: apmib_get( MIB_WLAN_ACCOUNT_RS_INTERVAL_TIME_3, (void *)&intVal); break;
			case 4: apmib_get( MIB_WLAN_ACCOUNT_RS_INTERVAL_TIME_4, (void *)&intVal); break;
			case 5: apmib_get( MIB_WLAN_ACCOUNT_RS_INTERVAL_TIME_5, (void *)&intVal); break;
			case 6: apmib_get( MIB_WLAN_ACCOUNT_RS_INTERVAL_TIME_6, (void *)&intVal); break;
			case 7: apmib_get( MIB_WLAN_ACCOUNT_RS_INTERVAL_TIME_7, (void *)&intVal); break;
			default: break;
		}
	}
	else {
		apmib_get( MIB_WLAN_ACCOUNT_RS_INTERVAL_TIME, (void *)&intVal);
	}
	sprintf(buf2, "accountRsAWhile = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);
#else
	apmib_get( MIB_WLAN_ACCOUNT_RS_INTERVAL_TIME, (void *)&intVal);
	sprintf(buf2, "accountRsAWhile = %d\n", intVal);
	WRITE_WPA_FILE(fh, buf2);
#endif
	}

#ifdef CONFIG_RTK_MESH
	else if (isWds==7) {

		apmib_get( MIB_MESH_ENCRYPT, (void *)&encrypt);
		sprintf(buf2, "encryption = %d\n", encrypt);
		WRITE_WPA_FILE(fh, buf2);

		apmib_get( MIB_MESH_ID,  (void *)buf1);
		sprintf(buf2, "ssid = \"%s\"\n", buf1);
		WRITE_WPA_FILE(fh, buf2);

		WRITE_WPA_FILE(fh, "enable1x = 0\n");
		WRITE_WPA_FILE(fh, "enableMacAuth = 0\n");
		WRITE_WPA_FILE(fh, "supportNonWpaClient = 0\n");
		WRITE_WPA_FILE(fh, "wepKey = 0\n");
		WRITE_WPA_FILE(fh,  "wepGroupKey = \"\"\n");

		apmib_get( MIB_MESH_WPA_AUTH, (void *)&intVal);
		sprintf(buf2, "authentication = %d\n", intVal);
		WRITE_WPA_FILE(fh, buf2);

		//sprintf(buf2, "unicastCipher = %d\n", intVal);
		sprintf(buf2, "unicastCipher = 1\n");
		WRITE_WPA_FILE(fh, buf2);

		apmib_get( MIB_MESH_WPA2_CIPHER_SUITE, (void *)&intVal);

		sprintf(buf2, "wpa2UnicastCipher = %d\n", intVal);
		WRITE_WPA_FILE(fh, buf2);

		WRITE_WPA_FILE(fh, "enablePreAuth = 0\n");

		apmib_get( MIB_MESH_WPA_PSK_FORMAT, (void *)&intVal);
		if (intVal==0)
			sprintf(buf2, "usePassphrase = 1\n");
		else
			sprintf(buf2, "usePassphrase = 0\n");
		WRITE_WPA_FILE(fh, buf2);

		apmib_get( MIB_MESH_WPA_PSK, (void *)buf1);
		sprintf(buf2, "psk = \"%s\"\n", buf1);
		WRITE_WPA_FILE(fh, buf2);

		WRITE_WPA_FILE(fh, "groupRekeyTime = 86400\n");
		WRITE_WPA_FILE(fh, "rsPort = 1812\n");
		WRITE_WPA_FILE(fh, "rsIP = 0.0.0.0\n");
		WRITE_WPA_FILE(fh, "rsPassword = \"\"\n");
		WRITE_WPA_FILE(fh, "rsMaxReq = 3\n");
		WRITE_WPA_FILE(fh, "rsAWhile = 5\n");
		WRITE_WPA_FILE(fh, "accountRsEnabled = 0\n");
		WRITE_WPA_FILE(fh, "accountRsPort = 1813\n");
		WRITE_WPA_FILE(fh, "accountRsIP = 0.0.0.0\n");
		WRITE_WPA_FILE(fh, "accountRsPassword = \"\"\n");
		WRITE_WPA_FILE(fh, "accountRsUpdateEnabled = 0\n");
		WRITE_WPA_FILE(fh, "accountRsUpdateTime = 60\n");
		WRITE_WPA_FILE(fh, "accountRsMaxReq = 3\n");
		WRITE_WPA_FILE(fh, "accountRsAWhile = 5\n");
	}
#endif // CONFIG_RTK_MESH

	else {
		apmib_get( MIB_WLAN_WDS_ENCRYPT, (void *)&encrypt);
/*
		if (encrypt == WDS_ENCRYPT_TKIP)
			encrypt = ENCRYPT_WPA;
		else if (encrypt == WDS_ENCRYPT_AES)
			encrypt = ENCRYPT_WPA2;
		else
			encrypt = 0;
*/
		if (encrypt == 2) //0: disable; 1: wep; 2:wpa psk
		{
			apmib_get( MIB_WLAN_WPA2_CIPHER_SUITE, (void *)&encrypt);
			if (encrypt == 2)
			{
				encrypt = ENCRYPT_WPA2;
				intVal = WPA_CIPHER_AES;
			}
			else
			{
				encrypt = ENCRYPT_WPA;
				intVal = WPA_CIPHER_TKIP;
			}
		}
		sprintf(buf2, "encryption = %d\n", encrypt);
		WRITE_WPA_FILE(fh, buf2);
		WRITE_WPA_FILE(fh, "ssid = \"REALTEK\"\n");
		WRITE_WPA_FILE(fh, "enable1x = 1\n");
		WRITE_WPA_FILE(fh, "enableMacAuth = 0\n");
		WRITE_WPA_FILE(fh, "supportNonWpaClient = 0\n");
		WRITE_WPA_FILE(fh, "wepKey = 0\n");
		WRITE_WPA_FILE(fh,  "wepGroupKey = \"\"\n");
		WRITE_WPA_FILE(fh,  "authentication = 2\n");

		if (encrypt == ENCRYPT_WPA)
			intVal = WPA_CIPHER_TKIP;
		else
			intVal = WPA_CIPHER_AES;

		sprintf(buf2, "unicastCipher = %d\n", intVal);
		WRITE_WPA_FILE(fh, buf2);

		sprintf(buf2, "wpa2UnicastCipher = %d\n", intVal);
		WRITE_WPA_FILE(fh, buf2);

		WRITE_WPA_FILE(fh, "enablePreAuth = 0\n");

		apmib_get( MIB_WLAN_WDS_WPA_PSK_FORMAT, (void *)&intVal);
		if (intVal==0)
			sprintf(buf2, "usePassphrase = 1\n");
		else
			sprintf(buf2, "usePassphrase = 0\n");
		WRITE_WPA_FILE(fh, buf2);

		apmib_get( MIB_WLAN_WDS_WPA_PSK, (void *)buf1);
		sprintf(buf2, "psk = \"%s\"\n", buf1);
		WRITE_WPA_FILE(fh, buf2);

		WRITE_WPA_FILE(fh, "groupRekeyTime = 0\n");
		WRITE_WPA_FILE(fh, "rsPort = 1812\n");
		WRITE_WPA_FILE(fh, "rsIP = 192.168.1.1\n");
		WRITE_WPA_FILE(fh, "rsPassword = \"\"\n");
		WRITE_WPA_FILE(fh, "rsMaxReq = 3\n");
		WRITE_WPA_FILE(fh, "rsAWhile = 10\n");
		WRITE_WPA_FILE(fh, "accountRsEnabled = 0\n");
		WRITE_WPA_FILE(fh, "accountRsPort = 1813\n");
		WRITE_WPA_FILE(fh, "accountRsIP = 192.168.1.1\n");
		WRITE_WPA_FILE(fh, "accountRsPassword = \"\"\n");
		WRITE_WPA_FILE(fh, "accountRsUpdateEnabled = 0\n");
		WRITE_WPA_FILE(fh, "accountRsUpdateTime = 1000\n");
		WRITE_WPA_FILE(fh, "accountRsMaxReq = 3\n");
		WRITE_WPA_FILE(fh, "accountRsAWhile = 1\n");
	}

	close(fh);
}

#ifdef WIFI_SIMPLE_CONFIG
enum {
	MODE_AP_UNCONFIG=1, 			// AP unconfigured (enrollee)
	MODE_CLIENT_UNCONFIG=2, 		// client unconfigured (enrollee)
	MODE_CLIENT_CONFIG=3,			// client configured (registrar)
	MODE_AP_PROXY=4, 			// AP configured (proxy)
	MODE_AP_PROXY_REGISTRAR=5,		// AP configured (proxy and registrar)
	MODE_CLIENT_UNCONFIG_REGISTRAR=6		// client unconfigured (registrar)
};

#define WRITE_WSC_PARAM(dst, tmp, str, val) {	\
	sprintf(tmp, str, val); \
	memcpy(dst, tmp, strlen(tmp)); \
	dst += strlen(tmp); \
}

static void convert_bin_to_str(unsigned char *bin, int len, char *out)
{
	int i;
	char tmpbuf[10];

	out[0] = '\0';

	for (i=0; i<len; i++) {
		sprintf(tmpbuf, "%02x", bin[i]);
		strcat(out, tmpbuf);
	}
}

static void convert_hex_to_ascii(unsigned long code, char *out)
{
	*out++ = '0' + ((code / 10000000) % 10);
	*out++ = '0' + ((code / 1000000) % 10);
	*out++ = '0' + ((code / 100000) % 10);
	*out++ = '0' + ((code / 10000) % 10);
	*out++ = '0' + ((code / 1000) % 10);
	*out++ = '0' + ((code / 100) % 10);
	*out++ = '0' + ((code / 10) % 10);
	*out++ = '0' + ((code / 1) % 10);
	*out = '\0';
}

static int compute_pin_checksum(unsigned long int PIN)
{
	unsigned long int accum = 0;
	int digit;

	PIN *= 10;
	accum += 3 * ((PIN / 10000000) % 10);
	accum += 1 * ((PIN / 1000000) % 10);
	accum += 3 * ((PIN / 100000) % 10);
	accum += 1 * ((PIN / 10000) % 10);
	accum += 3 * ((PIN / 1000) % 10);
	accum += 1 * ((PIN / 100) % 10);
	accum += 3 * ((PIN / 10) % 10);

	digit = (accum % 10);
	return (10 - digit) % 10;
}


static int updateWscConf(char *in, char *out, int genpin)
{
	int fh;
	struct stat status;
	char *buf, *ptr;
	int intVal, intVal2, intVal3, intVal4, is_client, is_config, is_registrar, len, is_wep=0, wep_key_type=0, wep_transmit_key=0;
	char tmpbuf[100], tmp1[100];
	unsigned char gen_pincode_tempbuf1[20];
	char gen_pincode_tempbuf2[20];

	if ( !apmib_init()) {
		printf("Initialize AP MIB failed!\n");
		return -1;
	}

//	apmib_get(MIB_WSC_PIN, (void *)tmpbuf);
	apmib_get(MIB_WPS_PIN_CODE, (void *)tmpbuf);
	if (genpin || !memcmp(tmpbuf, "\x0\x0\x0\x0\x0\x0\x0\x0", PIN_LEN)) {
		#include <sys/time.h>
		struct timeval tod;
		unsigned long num,num2;

		gettimeofday(&tod , NULL);
#if defined(_PINCODE_BY_MAC_)
  		apmib_get(MIB_HW_NIC0_ADDR, (void *)gen_pincode_tempbuf1);
  		
		sprintf(gen_pincode_tempbuf2, "%02X%02X%02X%02X",
			gen_pincode_tempbuf1[2],
			gen_pincode_tempbuf1[3],
			gen_pincode_tempbuf1[4],
			gen_pincode_tempbuf1[5]);
		printf("Get last  8 characters of HW_NIC0_ADDR = %s\n",gen_pincode_tempbuf2);
		num2 = strtoul(&gen_pincode_tempbuf2[0],NULL,16);
		printf("Generate MAC = %lu\n", num2);
#endif
		apmib_get(MIB_HW_NIC0_ADDR, (void *)&tmp1);
		tod.tv_sec += tmp1[4]+tmp1[5];
		srand(tod.tv_sec);
#if defined(_PINCODE_BY_MAC_)
		printf("Generate 10 PIN numbers= %lu\n", num2);
		 num2 =  num2% 10000000;
		 num2 = num2*10 + compute_pin_checksum(num2);
		convert_hex_to_ascii((unsigned long)num2, tmpbuf);
		printf("ping code %s\n", tmpbuf);
		apmib_set(MIB_WPS_PIN_CODE, (void *)tmpbuf);
#else
		num = rand() % 10000000;
		num = num*10 + compute_pin_checksum(num);
		convert_hex_to_ascii((unsigned long)num, tmpbuf);
		apmib_set(MIB_WPS_PIN_CODE, (void *)tmpbuf);
#endif
		apmib_update(CURRENT_SETTING);
		system("/bin/scriptlib_util.sh reloadFlash");
                system("killall -SIGUSR1 webs");


		printf("Generated PIN = %s\n", tmpbuf);

		if (genpin)
			return 0;
	}

	if (stat(in, &status) < 0) {
		printf("stat() error [%s]!\n", in);
		return -1;
	}

	buf = malloc(status.st_size+2048);
	if (buf == NULL) {
		printf("malloc() error [%d]!\n", (int)status.st_size+2048);
		return -1;
	}

	ptr = buf;
//EDX	apmib_get(MIB_AP_MODE, (void *)&is_client);
	apmib_get(MIB_AP_MODE, (void *)&is_client);
//EDX	apmib_get(MIB_WSC_CONFIGURED, (void *)&is_config);
	//apmib_get(MIB_WPS_CONFIG_STATUS, (void *)&is_config); // 1=configured 0=unconfigured
	apmib_get(MIB_WPS_CONFIG_TYPE, (void *)&is_config);
//EDX	apmib_get(MIB_WSC_REGISTRAR_ENABLED, (void *)&is_registrar);
	apmib_get(MIB_WPS_CONFIG_MODE, (void *)&is_registrar);	// EDX, register(0x1) proxy(0x2) enrolee(0x4)
	if(is_registrar != 0x1)
		is_registrar = 0;

#ifdef CONFIG_RTL8186_KLD_REPEATER
	int is_repeater_enabled;
	int wps_vxdAP_enabled=0;
	apmib_get(MIB_REPEATER_ENABLED1, (void *)&is_repeater_enabled);
#endif
	if (is_client == CLIENT_MODE) {
#ifdef CONFIG_RTL8186_KLD_REPEATER
		if (is_repeater_enabled && is_config) {
			intVal = MODE_AP_PROXY_REGISTRAR;
			wps_vxdAP_enabled = 1;
			WRITE_WSC_PARAM(ptr, tmpbuf, "disable_configured_by_exReg = %d\n", 1);
		}
		else
#endif
		{
			if (is_registrar) {
				if (!is_config)
					intVal = MODE_CLIENT_UNCONFIG_REGISTRAR;
				else
					intVal = MODE_CLIENT_CONFIG;
			}
			else
				intVal = MODE_CLIENT_UNCONFIG;
		}
	}
	else {
		if (!is_config)
			intVal = MODE_AP_UNCONFIG;
		else
			intVal = MODE_AP_PROXY_REGISTRAR;
	}
	WRITE_WSC_PARAM(ptr, tmpbuf, "mode = %d\n", intVal);

	if (is_client) {
#ifdef CONFIG_RTL8186_KLD_REPEATER
		if (wps_vxdAP_enabled)
			apmib_get(MIB_WSC_UPNP_ENABLED, (void *)&intVal);
		else
#endif
			intVal = 0;
	}
	else
//EDX		apmib_get(MIB_WSC_UPNP_ENABLED, (void *)&intVal);
	WRITE_WSC_PARAM(ptr, tmpbuf, "upnp = 1\n", intVal);

	intVal = 0;

// EDX	apmib_get(MIB_WSC_METHOD, (void *)&intVal);
	intVal = 3; // metheod, 1: PIN, 2: PBC, 3: both

	//Ethernet(0x2)+Label(0x4)+PushButton(0x80) Bitwise OR
	if (intVal == 1) //Pin+Ethernet
		intVal = (CONFIG_METHOD_ETH | CONFIG_METHOD_PIN);
	else if (intVal == 2) //PBC+Ethernet
		intVal = (CONFIG_METHOD_ETH | CONFIG_METHOD_PBC);
	if (intVal == 3) //Pin+PBC+Ethernet
		intVal = (CONFIG_METHOD_ETH | CONFIG_METHOD_PIN | CONFIG_METHOD_PBC);
	WRITE_WSC_PARAM(ptr, tmpbuf, "config_method = %d\n", intVal);

// EDX	apmib_get(MIB_WSC_AUTH, (void *)&intVal2);
	{
		int intSECURITY, intENCRYPT, intWPACIPHER, intWPA2CIPHER;

		intVal2 = 0;

		apmib_get(MIB_SECURITY_MODE, (void *)&intSECURITY); // 0: none 1:wep, 2: WPA_PSK, 3:WPA-Radius
		apmib_get(MIB_WLAN_ENCRYPT, (void *)&intENCRYPT); // 0 - Encrypt disabled, 1 - WEP, 2 - WPA, 4 - WPA2, 6 - WPA2 mixed
		apmib_get(MIB_WLAN_WPA_CIPHER_SUITE, (void *)&intWPACIPHER);   // WPA Cipher suite, 1 - TKIP, 2 - AES
		apmib_get(MIB_WLAN_WPA2_CIPHER_SUITE, (void *)&intWPA2CIPHER); // WPA Cipher suite, 1 - TKIP, 2 - AES
		if (( intSECURITY == 0 ) || (intSECURITY == 1 ))
			intVal2 = 1;
		else if ( intSECURITY == 2 )
		{
			// ignore WPA-RADIUS
			if( intENCRYPT & 2 )
				intVal2 += 2;
			if( intENCRYPT & 4 )
			intVal2 += 0x20;
		}

		// # AUTH_OPEN=1, AUTH_WPAPSK=2, AUTH_SHARED=4, AUTH_WPA=8, AUTH_WPA2=0x10, AUTH_WPA2PSK=0x20, ex:34
		WRITE_WSC_PARAM(ptr, tmpbuf, "auth_type = %d\n", intVal2);

		intVal = 0;

		if (( intSECURITY <= 1 ))
			intVal = intSECURITY;
		else
		{
			if((intWPACIPHER == 1) || (intWPA2CIPHER == 1))
				intVal += 4;
			if((intWPACIPHER == 2) || (intWPA2CIPHER == 2))
				intVal += 8;
		}


		// # ENCRYPT_NONE=1, ENCRYPT_WEP=2, ENCRYPT_TKIP=4, ENCRYPT_AES=8, ex:12
		WRITE_WSC_PARAM(ptr, tmpbuf, "encrypt_type = %d\n", intVal);
	}

	if (intVal == WSC_ENCRYPT_WEP)
		is_wep = 1;

	if (is_client) {
#ifdef CONFIG_RTL8186_KLD_REPEATER
		if (wps_vxdAP_enabled)
			intVal = 1;
		else
#endif
		{
			apmib_get(MIB_WLAN_NETWORK_TYPE, (void *)&intVal);
			if (intVal == 0)
				intVal = 1;
			else
				intVal = 2;
		}
	}
	else
		intVal = 1;
	WRITE_WSC_PARAM(ptr, tmpbuf, "connection_type = %d\n", intVal);

// EDX	apmib_get(MIB_WSC_MANUAL_ENABLED, (void *)&intVal);
	intVal = 0;

	WRITE_WSC_PARAM(ptr, tmpbuf, "manual_config = %d\n", intVal);

	if (is_wep) { // only allow WEP in none-MANUAL mode (configured by external registrar)
		apmib_get(MIB_WLAN_ENCRYPT, (void *)&intVal);
		if (intVal != ENCRYPT_WEP) {
			printf("WEP mismatched between WPS and host system\n");
			free(buf);
			return -1;
		}
		apmib_get(MIB_WLAN_WEP, (void *)&intVal);
		if (intVal <= WEP_DISABLED || intVal > WEP128) {
			printf("WEP encrypt length error\n");
			free(buf);
			return -1;
		}
		apmib_get(MIB_WLAN_WEP_KEY_TYPE, (void *)&wep_key_type);
		apmib_get(MIB_WLAN_WEP_DEFAULT_KEY, (void *)&wep_transmit_key);
		wep_transmit_key++;
		WRITE_WSC_PARAM(ptr, tmpbuf, "wep_transmit_key = %d\n", wep_transmit_key);
		if (intVal == WEP64) {
			apmib_get(MIB_WLAN_WEP64_KEY1, (void *)&tmpbuf);
			if (wep_key_type == KEY_ASCII) {
				memcpy(tmp1, tmpbuf, 5);
				tmp1[5] = '\0';
			}
			else {
				convert_bin_to_str(tmpbuf, 5, tmp1);
				tmp1[10] = '\0';
			}
			WRITE_WSC_PARAM(ptr, tmpbuf, "network_key = %s\n", tmp1);

			apmib_get(MIB_WLAN_WEP64_KEY2, (void *)&tmpbuf);
			if (wep_key_type == KEY_ASCII) {
				memcpy(tmp1, tmpbuf, 5);
				tmp1[5] = '\0';
			}
			else {
				convert_bin_to_str(tmpbuf, 5, tmp1);
				tmp1[10] = '\0';
			}
			WRITE_WSC_PARAM(ptr, tmpbuf, "wep_key2 = %s\n", tmp1);

			apmib_get(MIB_WLAN_WEP64_KEY3, (void *)&tmpbuf);
			if (wep_key_type == KEY_ASCII) {
				memcpy(tmp1, tmpbuf, 5);
				tmp1[5] = '\0';
			}
			else {
				convert_bin_to_str(tmpbuf, 5, tmp1);
				tmp1[10] = '\0';
			}
			WRITE_WSC_PARAM(ptr, tmpbuf, "wep_key3 = %s\n", tmp1);


			apmib_get(MIB_WLAN_WEP64_KEY4, (void *)&tmpbuf);
			if (wep_key_type == KEY_ASCII) {
				memcpy(tmp1, tmpbuf, 5);
				tmp1[5] = '\0';
			}
			else {
				convert_bin_to_str(tmpbuf, 5, tmp1);
				tmp1[10] = '\0';
			}
			WRITE_WSC_PARAM(ptr, tmpbuf, "wep_key4 = %s\n", tmp1);
		}
		else {
			apmib_get(MIB_WLAN_WEP128_KEY1, (void *)&tmpbuf);
			if (wep_key_type == KEY_ASCII) {
				memcpy(tmp1, tmpbuf, 13);
				tmp1[13] = '\0';
			}
			else {
				convert_bin_to_str(tmpbuf, 13, tmp1);
				tmp1[26] = '\0';
			}
			WRITE_WSC_PARAM(ptr, tmpbuf, "network_key = %s\n", tmp1);


			apmib_get(MIB_WLAN_WEP128_KEY2, (void *)&tmpbuf);
			if (wep_key_type == KEY_ASCII) {
				memcpy(tmp1, tmpbuf, 13);
				tmp1[13] = '\0';
			}
			else {
				convert_bin_to_str(tmpbuf, 13, tmp1);
				tmp1[26] = '\0';
			}
			WRITE_WSC_PARAM(ptr, tmpbuf, "wep_key2 = %s\n", tmp1);

			apmib_get(MIB_WLAN_WEP128_KEY3, (void *)&tmpbuf);
			if (wep_key_type == KEY_ASCII) {
				memcpy(tmp1, tmpbuf, 13);
				tmp1[13] = '\0';
			}
			else {
				convert_bin_to_str(tmpbuf, 13, tmp1);
				tmp1[26] = '\0';
			}
			WRITE_WSC_PARAM(ptr, tmpbuf, "wep_key3 = %s\n", tmp1);

			apmib_get(MIB_WLAN_WEP128_KEY4, (void *)&tmpbuf);
			if (wep_key_type == KEY_ASCII) {
				memcpy(tmp1, tmpbuf, 13);
				tmp1[13] = '\0';
			}
			else {
				convert_bin_to_str(tmpbuf, 13, tmp1);
				tmp1[26] = '\0';
			}
			WRITE_WSC_PARAM(ptr, tmpbuf, "wep_key4 = %s\n", tmp1);
		}
	}
	else {
		apmib_get(MIB_WLAN_WPA_PSK, (void *)&tmp1);
		WRITE_WSC_PARAM(ptr, tmpbuf, "network_key = %s\n", tmp1);
	}

#ifdef CONFIG_RTL8186_KLD_REPEATER
	if (wps_vxdAP_enabled)
		apmib_get(MIB_REPEATER_SSID1, (void *)&tmp1);
	else
#endif
		apmib_get(MIB_WLAN_SSID, (void *)&tmp1);
	WRITE_WSC_PARAM(ptr, tmpbuf, "ssid = %s\n", tmp1);

#if 0
//	}
//	else {
		apmib_get(MIB_WSC_PSK, (void *)&tmp1);
		WRITE_WSC_PARAM(ptr, tmpbuf, "network_key = %s\n", tmp1);

		apmib_get(MIB_WSC_SSID, (void *)&tmp1);
		WRITE_WSC_PARAM(ptr, tmpbuf, "ssid = %s\n", tmp1);
//	}
#endif

	apmib_get(MIB_WPS_PIN_CODE, (void *)&tmp1);
//	apmib_get(MIB_HW_WSC_PIN, (void *)&tmp1);
	WRITE_WSC_PARAM(ptr, tmpbuf, "pin_code = %s\n", tmp1);

	apmib_get(MIB_WLAN_CHAN_NUM, (void *)&intVal);
	if (intVal > 14)
		intVal = 2;
	else
		intVal = 1;
	WRITE_WSC_PARAM(ptr, tmpbuf, "rf_band = %d\n", intVal);

/*
	apmib_get(MIB_HW_MODEL_NUM, (void *)&tmp1);
	WRITE_WSC_PARAM(ptr, tmpbuf, "model_num = \"%s\"\n", tmp1);

	apmib_get(MIB_HW_SERIAL_NUM, (void *)&tmp1);
	WRITE_WSC_PARAM(ptr, tmpbuf, "serial_num = \"%s\"\n", tmp1);
*/
// EDX	apmib_get(MIB_DEVICE_NAME, (void *)&tmp1);
	WRITE_WSC_PARAM(ptr, tmpbuf, "device_name = \"%s\"\n", _PRODUCT_NAME_);

	len = (int)(((long)ptr)-((long)buf));

	fh = open(in, O_RDONLY);
	if (fh == -1) {
		printf("open() error [%s]!\n", in);
		return -1;
	}

	lseek(fh, 0L, SEEK_SET);
	if (read(fh, ptr, status.st_size) != status.st_size) {
		printf("read() error [%s]!\n", in);
		return -1;
	}
	close(fh);

	// search UUID field, replace last 12 char with hw mac address
	ptr = strstr(ptr, "uuid =");
	if (ptr) {
		char tmp2[100];
		apmib_get(MIB_HW_NIC0_ADDR, (void *)&tmp1);
		convert_bin_to_str(tmp1, 6, tmp2);
		memcpy(ptr+27, tmp2, 12);
	}

	fh = open(out, O_RDWR|O_CREAT|O_TRUNC);
	if (fh == -1) {
		printf("open() error [%s]!\n", out);
		return -1;
	}

	if (write(fh, buf, len+status.st_size) != len+status.st_size ) {
		printf("Write() file error [%s]!\n", out);
		return -1;
	}
	close(fh);
	free(buf);

	return 0;
}

#endif

/*
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/wireless.h>
#include "../../../linux-2.4.18_1400/drivers/net/rtl8190/ieee802_mib.h"
#include "../../linux-2.4.18_1400/drivers/net/rtl8190/ieee802_mib.h"

static int initWlan(char *ifname)
{
	struct wifi_mib *pmib;
	int i, intVal, intVal2, encrypt, enable1x, wep, mode;// enable1xVxd;
	unsigned char buf1[1024], buf2[1024], mac[6];
	int skfd;
	struct iwreq wrq, wrq_root;
	int wlan_band=0, channel_bound=0, aggregation=0;
	MACFILTER_T *pAcl=NULL;
	struct wdsEntry *wds_Entry=NULL;
	//WDS_Tp pwds_EntryUI;
#ifdef MSSID
	int v_previous=0;
#ifdef CONFIG_RTL8196B
	int vap_enable=0, intVal4=0;
#endif
#endif
	
	skfd = socket(AF_INET, SOCK_DGRAM, 0);
	strncpy(wrq.ifr_name, ifname, IFNAMSIZ);
	if (ioctl(skfd, SIOCGIWNAME, &wrq) < 0) {
		printf("Interface open failed!\n");
		return -1;
	}

	if ((pmib = (struct wifi_mib *)malloc(sizeof(struct wifi_mib))) == NULL) {
		printf("MIB buffer allocation failed!\n");
		return -1;
	}

	if (!apmib_init()) {
		printf("Initialize AP MIB failed!\n");
		return -1;
	}

	// Disable WLAN MAC driver and shutdown interface first
	sprintf(buf1, "ifconfig %s down", ifname);
	system(buf1);

	if (vwlan_idxx == 0) {
		// shutdown all WDS interface
		for (i=0; i<8; i++) {
			sprintf(buf1, "ifconfig %s-wds%d down", ifname, i);
			system(buf1);
		}
	
		// kill wlan application daemon
		sprintf(buf1, "wlanapp.sh kill %s", ifname);
		system(buf1);
	}
	else { // virtual interface
		sprintf(buf1, "wlan%d", wlan_idx);		
		strncpy(wrq_root.ifr_name, buf1, IFNAMSIZ);
		if (ioctl(skfd, SIOCGIWNAME, &wrq_root) < 0) {
			printf("Root Interface open failed!\n");
			return -1;
		}		
	}

	if (vwlan_idxx == 0) {
		apmib_get(MIB_HW_RF_TYPE, (void *)&intVal);
		if (intVal == 0) {
			printf("RF type is NULL!\n");
			return 0;
		}
	}
	apmib_get(MIB_WLAN_DISABLED, (void *)&intVal);

	if (intVal == 1) {
		free(pmib);
		close(skfd);		
		return 0;
	}

	// get mib from driver
	wrq.u.data.pointer = (caddr_t)pmib;
	wrq.u.data.length = sizeof(struct wifi_mib);

	if (vwlan_idxx == 0) {
		if (ioctl(skfd, 0x8B42, &wrq) < 0) {
			printf("Get WLAN MIB failed!\n");
			return -1;
		}
	}
	else {
		wrq_root.u.data.pointer = (caddr_t)pmib;
		wrq_root.u.data.length = sizeof(struct wifi_mib);				
		if (ioctl(skfd, 0x8B42, &wrq_root) < 0) {
			printf("Get WLAN MIB failed!\n");
			return -1;
		}		
	}

	// check mib version
	if (pmib->mib_version != MIB_VERSION) {
		printf("WLAN MIB version mismatch!\n");
		return -1;
	}

	if (vwlan_idxx > 0) {	//if not root interface, clone root mib to virtual interface
		wrq.u.data.pointer = (caddr_t)pmib;
		wrq.u.data.length = sizeof(struct wifi_mib);
		if (ioctl(skfd, 0x8B43, &wrq) < 0) {
			printf("Set WLAN MIB failed!\n");
			return -1;
		}	
	}

	// Set parameters to driver
	if (vwlan_idxx == 0) {	
		apmib_get(MIB_HW_REG_DOMAIN, (void *)&intVal);
		pmib->dot11StationConfigEntry.dot11RegDomain = intVal;
	}

	apmib_get(MIB_WLAN_MAC_ADDR, (void *)mac);
	if (!memcmp(mac, "\x00\x00\x00\x00\x00\x00", 6)) {
#ifdef MSSID
		if (vwlan_idxx > 0 && vwlan_idxx != NUM_VWLAN_INTERFACE) {
			switch (vwlan_idxx)
			{
				case 1:
					apmib_get(MIB_HW_WLAN_ADDR1, (void *)mac);
					break;
				case 2:
					apmib_get(MIB_HW_WLAN_ADDR2, (void *)mac);
					break;
				case 3:
					apmib_get(MIB_HW_WLAN_ADDR3, (void *)mac);
					break;
				case 4:
					apmib_get(MIB_HW_WLAN_ADDR4, (void *)mac);
					break;
				default:
					printf("Fail to get MAC address of VAP%d!\n", vwlan_idxx-1);
					return 0;
			}
		}
		else
#endif
		apmib_get(MIB_HW_WLAN_ADDR, (void *)mac);
	}

	// ifconfig all wlan interface when not in WISP
	// ifconfig wlan1 later interface when in WISP mode, the wlan0  will be setup in WAN interface
	apmib_get(MIB_OP_MODE, (void *)&intVal);
	apmib_get(MIB_WISP_WAN_ID, (void *)&intVal2);
	sprintf(buf1, "wlan%d", intVal2);
	if ((intVal != 2) ||
#ifdef MSSID
		vwlan_idxx > 0 ||
#endif
		strcmp(ifname, buf1)) {
		sprintf(buf2, "ifconfig %s hw ether %02x%02x%02x%02x%02x%02x", ifname, mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
		system(buf2);
		memcpy(&(pmib->dot11OperationEntry.hwaddr[0]), mac, 6);
	}

#ifdef BR_SHORTCUT	
	if (intVal == 2
#ifdef MSSID
		&& vwlan_idxx == 0
#endif
	) 
		pmib->dot11OperationEntry.disable_brsc = 1;
#endif
	
	apmib_get(MIB_HW_LED_TYPE, (void *)&intVal);
	pmib->dot11OperationEntry.ledtype = intVal;

	// set AP/client/WDS mode
	apmib_get(MIB_WLAN_SSID, (void *)buf1);
	intVal2 = strlen(buf1);
	pmib->dot11StationConfigEntry.dot11DesiredSSIDLen = intVal2;
	memset(pmib->dot11StationConfigEntry.dot11DesiredSSID, 0, 32);
	memcpy(pmib->dot11StationConfigEntry.dot11DesiredSSID, buf1, intVal2);
	if ((pmib->dot11StationConfigEntry.dot11DesiredSSIDLen == 3) &&
		((pmib->dot11StationConfigEntry.dot11DesiredSSID[0] == 'A') || (pmib->dot11StationConfigEntry.dot11DesiredSSID[0] == 'a')) &&
		((pmib->dot11StationConfigEntry.dot11DesiredSSID[1] == 'N') || (pmib->dot11StationConfigEntry.dot11DesiredSSID[1] == 'n')) &&
		((pmib->dot11StationConfigEntry.dot11DesiredSSID[2] == 'Y') || (pmib->dot11StationConfigEntry.dot11DesiredSSID[2] == 'y'))) {
		pmib->dot11StationConfigEntry.dot11SSIDtoScanLen = 0;
		memset(pmib->dot11StationConfigEntry.dot11SSIDtoScan, 0, 32);
	}
	else {
		pmib->dot11StationConfigEntry.dot11SSIDtoScanLen = intVal2;
		memset(pmib->dot11StationConfigEntry.dot11SSIDtoScan, 0, 32);
		memcpy(pmib->dot11StationConfigEntry.dot11SSIDtoScan, buf1, intVal2);
	}

	apmib_get(MIB_AP_MODE, (void *)&mode);
	if (mode == 1) {
		// client mode
		apmib_get(MIB_WLAN_NETWORK_TYPE, (void *)&intVal2);
		if (intVal2 == 0)
			pmib->dot11OperationEntry.opmode = 8;
		else {
			pmib->dot11OperationEntry.opmode = 32;
			apmib_get(MIB_WLAN_SSID, (void *)buf1);
			intVal2 = strlen(buf1);
			pmib->dot11StationConfigEntry.dot11DefaultSSIDLen = intVal2;
			memset(pmib->dot11StationConfigEntry.dot11DefaultSSID, 0, 32);
			memcpy(pmib->dot11StationConfigEntry.dot11DefaultSSID, buf1, intVal2);
		}
	}
	else
		pmib->dot11OperationEntry.opmode = 16;

	if (mode == 2)	// WDS only
		pmib->dot11WdsInfo.wdsPure = 1;
	else
		pmib->dot11WdsInfo.wdsPure = 0;

	if (vwlan_idxx == 0) { // root interface	
		// set RF parameters
		apmib_get(MIB_HW_RF_TYPE, (void *)&intVal);
		pmib->dot11RFEntry.dot11RFType = intVal;
		
#if defined(CONFIG_RTL8196B)

		apmib_get(MIB_HW_BOARD_VER, (void *)&intVal);
		if (intVal == 1)
			pmib->dot11RFEntry.MIMO_TR_mode = 3;	// 2T2R
		else if(intVal == 2)
                        pmib->dot11RFEntry.MIMO_TR_mode = 4; // 1T1R
		else
			pmib->dot11RFEntry.MIMO_TR_mode = 1;	// 1T2R

		apmib_get(MIB_HW_TX_POWER_CCK, (void *)buf1);
		memcpy(pmib->dot11RFEntry.pwrlevelCCK, buf1, 14);
		
		apmib_get(MIB_HW_TX_POWER_OFDM_1S, (void *)buf1);
		memcpy(pmib->dot11RFEntry.pwrlevelOFDM_1SS, buf1, 162);
		
		apmib_get(MIB_HW_TX_POWER_OFDM_2S, (void *)buf1);
		memcpy(pmib->dot11RFEntry.pwrlevelOFDM_2SS, buf1, 162);

		// not used for RTL8192SE
		//apmib_get(MIB_HW_11N_XCAP, (void *)&intVal);
		//pmib->dot11RFEntry.crystalCap = intVal;
		
		apmib_get(MIB_HW_11N_LOFDMPWDA, (void *)&intVal);
		pmib->dot11RFEntry.LOFDM_pwd_A = intVal;

		apmib_get(MIB_HW_11N_LOFDMPWDB, (void *)&intVal);
		pmib->dot11RFEntry.LOFDM_pwd_B = intVal;

		apmib_get(MIB_HW_11N_TSSI1, (void *)&intVal);
		pmib->dot11RFEntry.tssi1 = intVal;

		apmib_get(MIB_HW_11N_TSSI2, (void *)&intVal);
		pmib->dot11RFEntry.tssi2 = intVal;

		apmib_get(MIB_HW_11N_THER, (void *)&intVal);
		pmib->dot11RFEntry.ther = intVal;

		if (pmib->dot11RFEntry.dot11RFType == 10) { // Zebra
			apmib_get(MIB_WLAN_RF_POWER, (void *)&intVal);
			if(intVal == 1)
				intVal = 3;
			else if(intVal == 2)
					intVal = 6;
				else if(intVal == 3)
						intVal = 9;
					else if(intVal == 4)
							intVal = 17;
			if (intVal) {
				for (i=0; i<14; i++) {
					if(pmib->dot11RFEntry.pwrlevelCCK[i] != 0){ 
						if ((pmib->dot11RFEntry.pwrlevelCCK[i] - intVal) >= 1)
							pmib->dot11RFEntry.pwrlevelCCK[i] -= intVal;
						else
							pmib->dot11RFEntry.pwrlevelCCK[i] = 1;
					}
				}
				for (i=0; i<162; i++) {
					if (pmib->dot11RFEntry.pwrlevelOFDM_1SS[i] != 0){
						if((pmib->dot11RFEntry.pwrlevelOFDM_1SS[i] - intVal) >= 1)
							pmib->dot11RFEntry.pwrlevelOFDM_1SS[i] -= intVal;
						else
							pmib->dot11RFEntry.pwrlevelOFDM_1SS[i] = 1;
					}
					if (pmib->dot11RFEntry.pwrlevelOFDM_2SS[i] != 0){
						if((pmib->dot11RFEntry.pwrlevelOFDM_2SS[i] - intVal) >= 1)
							pmib->dot11RFEntry.pwrlevelOFDM_2SS[i] -= intVal;
						else
							pmib->dot11RFEntry.pwrlevelOFDM_2SS[i] = 1;
					}
				}		
			}
		}

#else
//!CONFIG_RTL8196B => rtl8651c+rtl8190
		apmib_get(MIB_HW_ANT_DIVERSITY, (void *)&intVal);
		pmib->dot11RFEntry.dot11DiversitySupport = intVal;

		apmib_get(MIB_HW_TX_ANT, (void *)&intVal);
		pmib->dot11RFEntry.defaultAntennaB = intVal;

#if 0
		apmib_get(MIB_HW_INIT_GAIN, (void *)&intVal);
		pmib->dot11RFEntry.initialGain = intVal;
#endif

		apmib_get(MIB_HW_TX_POWER_CCK, (void *)buf1);
		memcpy(pmib->dot11RFEntry.pwrlevelCCK, buf1, 14);
		
		apmib_get(MIB_HW_TX_POWER_OFDM, (void *)buf1);
		memcpy(pmib->dot11RFEntry.pwrlevelOFDM, buf1, 162);

		apmib_get(MIB_HW_11N_LOFDMPWD, (void *)&intVal);
		pmib->dot11RFEntry.legacyOFDM_pwrdiff = intVal;
		
		apmib_get(MIB_HW_11N_ANTPWD_C, (void *)&intVal);
		pmib->dot11RFEntry.antC_pwrdiff = intVal;
		
		apmib_get(MIB_HW_11N_THER_RFIC, (void *)&intVal);
		pmib->dot11RFEntry.ther_rfic = intVal;
		
		apmib_get(MIB_HW_11N_XCAP, (void *)&intVal);
		pmib->dot11RFEntry.crystalCap = intVal;

		// set output power scale
		//if (pmib->dot11RFEntry.dot11RFType == 7) { // Zebra
			if (pmib->dot11RFEntry.dot11RFType == 10) { // Zebra
			apmib_get(MIB_HW_RF_TYPE, (void *)&intVal);
			if(intVal == 1)
				intVal = 3;
			else if(intVal == 2)
					intVal = 6;
				else if(intVal == 3)
						intVal = 9;
					else if(intVal == 4)
							intVal = 17;
			if (intVal) {
				for (i=0; i<14; i++) {
					if(pmib->dot11RFEntry.pwrlevelCCK[i] != 0){ 
						if ((pmib->dot11RFEntry.pwrlevelCCK[i] - intVal) >= 1)
							pmib->dot11RFEntry.pwrlevelCCK[i] -= intVal;
						else
							pmib->dot11RFEntry.pwrlevelCCK[i] = 1;
					}
				}
				for (i=0; i<162; i++) {
					if (pmib->dot11RFEntry.pwrlevelOFDM[i] != 0){
						if((pmib->dot11RFEntry.pwrlevelOFDM[i] - intVal) >= 1)
							pmib->dot11RFEntry.pwrlevelOFDM[i] -= intVal;
						else
							pmib->dot11RFEntry.pwrlevelOFDM[i] = 1;
					}
				}		
			}
		}
#endif  //For Check CONFIG_RTL8196B
		
		apmib_get(MIB_WLAN_BEACON_INTERVAL, (void *)&intVal);
		pmib->dot11StationConfigEntry.dot11BeaconPeriod = intVal;

		apmib_get(MIB_WLAN_CHAN_NUM, (void *)&intVal);
		pmib->dot11RFEntry.dot11channel = intVal;

		apmib_get(MIB_WLAN_RTS_THRESHOLD, (void *)&intVal);
		pmib->dot11OperationEntry.dot11RTSThreshold = intVal;

		apmib_get(MIB_WLAN_FRAG_THRESHOLD, (void *)&intVal);
		pmib->dot11OperationEntry.dot11FragmentationThreshold = intVal;

		apmib_get(MIB_WLAN_INACTIVITY_TIME, (void *)&intVal);
		pmib->dot11OperationEntry.expiretime = intVal;

		apmib_get(MIB_WLAN_PREAMBLE_TYPE, (void *)&intVal);
		pmib->dot11RFEntry.shortpreamble = intVal;

		apmib_get(MIB_WLAN_DTIM_PERIOD, (void *)&intVal);
		pmib->dot11StationConfigEntry.dot11DTIMPeriod = intVal;

		// enable/disable the notification for IAPP
		apmib_get(MIB_WLAN_IAPP_DISABLED, (void *)&intVal);
		if (intVal == 0)
			pmib->dot11OperationEntry.iapp_enable = 1;
		else
			pmib->dot11OperationEntry.iapp_enable = 0;

		// set 11g protection mode
		apmib_get(MIB_WLAN_PROTECTION_DISABLED, (void *)&intVal);
		pmib->dot11StationConfigEntry.protectionDisabled = intVal;

		// set block relay
		apmib_get(MIB_WLAN_BLOCK_RELAY, (void *)&intVal);
		pmib->dot11OperationEntry.block_relay = intVal;

		// set WiFi specific mode
		apmib_get(MIB_WIFI_TEST, (void *)&intVal);
		pmib->dot11OperationEntry.wifi_specific = intVal;

		// Set WDS
		apmib_get(MIB_WLAN_WDS_ENABLED, (void *)&intVal);
		apmib_get(MIB_WLAN_WDS_NUM, (void *)&intVal2);
		pmib->dot11WdsInfo.wdsNum = 0;
#ifdef MSSID 
		if (v_previous > 0) 
			intVal = 0;
#endif
		if (((mode == 2) || (mode == 3)) &&
			(intVal != 0) &&
			(intVal2 != 0)) {
//			for (i=0; i<intVal2; i++) {
//				buf1[0] = i+1;
//				apmib_get(MIB_WLAN_WDS, (void *)buf1);
//				pwds_EntryUI = (WDS_Tp)buf1;
//				wds_Entry = &(pmib->dot11WdsInfo.entry[i]);
//				memcpy(wds_Entry->macAddr, &(pwds_EntryUI->macAddr[0]), 6);
//				wds_Entry->txRate = pwds_EntryUI->fixedTxRate;
//				pmib->dot11WdsInfo.wdsNum++;
//				sprintf(buf2, "ifconfig %s-wds%d hw ether %02x%02x%02x%02x%02x%02x", ifname, i, mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
//				system(buf2);
//			}
			pmib->dot11WdsInfo.wdsEnabled = intVal;
		}
		else
			pmib->dot11WdsInfo.wdsEnabled = 0;

		if (((mode == 2) || (mode == 3)) &&
			(intVal != 0)) {
			apmib_get(MIB_WLAN_WDS_ENCRYPT, (void *)&intVal);
			if (intVal == 0)
				pmib->dot11WdsInfo.wdsPrivacy = 0;
			else if (intVal == 1) {
				apmib_get(MIB_WLAN_WDS_WPA_PSK, (void *)buf1);
				pmib->dot11WdsInfo.wdsPrivacy = 1;
				string_to_hex(buf1, &(pmib->dot11WdsInfo.wdsWepKey[0]), 10);
			}
			else if (intVal == 2) {
				apmib_get(MIB_WLAN_WDS_WPA_PSK, (void *)buf1);
				pmib->dot11WdsInfo.wdsPrivacy = 5;
				string_to_hex(buf1, &(pmib->dot11WdsInfo.wdsWepKey[0]), 26);
			}
			else if (intVal == 3)
				pmib->dot11WdsInfo.wdsPrivacy = 2;
			else
				pmib->dot11WdsInfo.wdsPrivacy = 4;
		}

		// enable/disable the notification for IAPP
		apmib_get(MIB_WLAN_IAPP_DISABLED, (void *)&intVal);
		if (intVal == 0)
			pmib->dot11OperationEntry.iapp_enable = 1;
		else
			pmib->dot11OperationEntry.iapp_enable = 0;

		pmib->dot11StationConfigEntry.dot11AclNum = 0;
		apmib_get(MIB_WLAN_AC_ENABLED, (void *)&intVal);
		pmib->dot11StationConfigEntry.dot11AclMode = intVal;
		if (intVal != 0) {
			apmib_get(MIB_WLAN_AC_NUM, (void *)&intVal);
			if (intVal != 0) {
				for (i=0; i<intVal; i++) {
					buf1[0] = i+1;
					apmib_get(MIB_WLAN_AC_ADDR, (void *)buf1);
					pAcl = (MACFILTER_T *)buf1;
					memcpy(&(pmib->dot11StationConfigEntry.dot11AclAddr[i][0]), &(pAcl->macAddr[0]), 6);
					pmib->dot11StationConfigEntry.dot11AclNum++;
				}
			}
		}

#if defined(CONFIG_RTK_MESH) && defined(_MESH_ACL_ENABLE_) // below code copy above ACL code
		// Copy Webpage setting to userspace MIB struct table
		pmib->dot1180211sInfo.mesh_acl_num = 0;
		apmib_get(MIB_MESH_ACL_ENABLED, (void *)&intVal);
		pmib->dot1180211sInfo.mesh_acl_mode = intVal;

		if (intVal != 0) {
			apmib_get(MIB_MESH_ACL_NUM, (void *)&intVal);
			if (intVal != 0) {
				for (i=0; i<intVal; i++) {
					buf1[0] = i+1;
					apmib_get(MIB_MESH_ACL_ADDR, (void *)buf1);
					pAcl = (MACFILTER_T *)buf1;
					memcpy(&(pmib->dot1180211sInfo.mesh_acl_addr[i][0]), &(pAcl->macAddr[0]), 6);
					pmib->dot1180211sInfo.mesh_acl_num++;
				}
			}
		}
#endif

		// set nat2.5 disable when client and mac clone is set
		apmib_get(MIB_WLAN_NAT25_MAC_CLONE, (void *)&intVal);
		if ((intVal == 1) && (mode == 1)) {
			pmib->ethBrExtInfo.nat25_disable = 1;
			pmib->ethBrExtInfo.macclone_enable = 1;
		}
		else {
			pmib->ethBrExtInfo.nat25_disable = 0;
			pmib->ethBrExtInfo.macclone_enable = 0;
		}		

		// set nat2.5 disable and macclone disable when wireless isp mode
		apmib_get(MIB_OP_MODE, (void *)&intVal);
		if (intVal == 2) {
			pmib->ethBrExtInfo.nat25_disable = 1;
			pmib->ethBrExtInfo.macclone_enable = 0;
		}

#ifdef WIFI_SIMPLE_CONFIG
		pmib->wscEntry.wsc_enable = 0;
#endif

	// for 11n
		apmib_get(MIB_ABAND_CHANNEL, &channel_bound);
		pmib->dot11nConfigEntry.dot11nUse40M = channel_bound;
		//apmib_get(MIB_WLAN_CONTROL_SIDEBAND, &intVal);
		apmib_get(MIB_ABAND_MODE, &intVal);
		
		if(channel_bound ==0){
			pmib->dot11nConfigEntry.dot11n2ndChOffset = 0;
		}else {
			if(intVal == 0 )
				pmib->dot11nConfigEntry.dot11n2ndChOffset = 1;
			if(intVal == 1 )
				pmib->dot11nConfigEntry.dot11n2ndChOffset = 2;	
		}
		//apmib_get(MIB_WLAN_SHORT_GI, &intVal);
		//pmib->dot11nConfigEntry.dot11nShortGIfor20M = intVal;
		//pmib->dot11nConfigEntry.dot11nShortGIfor40M = intVal;
		pmib->dot11nConfigEntry.dot11nShortGIfor20M = 1;
		pmib->dot11nConfigEntry.dot11nShortGIfor40M = 1;
		
		//apmib_get(MIB_WLAN_AGGREGATION, &aggregation);
		aggregation = 0;
		if(aggregation ==0){
			pmib->dot11nConfigEntry.dot11nAMPDU = 0;
			pmib->dot11nConfigEntry.dot11nAMSDU = 0;
		}else if(aggregation ==1){
			pmib->dot11nConfigEntry.dot11nAMPDU = 1;
			pmib->dot11nConfigEntry.dot11nAMSDU = 0;
		}else if(aggregation ==2){
			pmib->dot11nConfigEntry.dot11nAMPDU = 0;
			pmib->dot11nConfigEntry.dot11nAMSDU = 1;
		}
		else if(aggregation ==3){
			pmib->dot11nConfigEntry.dot11nAMPDU = 1;
			pmib->dot11nConfigEntry.dot11nAMSDU = 1;
		}

#if defined(CONFIG_RTL8196B) && defined(MSSID)
		if(pmib->dot11OperationEntry.opmode & 0x00000010){// AP mode
			for (vwlan_idxx = 1; vwlan_idxx < 5; vwlan_idxx++) {
				apmib_get(MIB_WLAN_DISABLED, (void *)&intVal4);
				if (intVal4 == 0)
					vap_enable++;
				intVal4=0;
			}
			vwlan_idxx = 0;
		}
		if (vap_enable && (mode ==  AP_MODE || mode ==  AP_WDS_MODE))	
			pmib->miscEntry.vap_enable=1;
		else
			pmib->miscEntry.vap_enable=0;
#endif
	}

	if (vwlan_idxx != NUM_VWLAN_INTERFACE) { // not repeater interface
		apmib_get(MIB_WLAN_BASIC_RATE, (void *)&intVal);
		pmib->dot11StationConfigEntry.dot11BasicRates = intVal;

		apmib_get(MIB_WLAN_SUPPORTED_RATE, (void *)&intVal);
		pmib->dot11StationConfigEntry.dot11SupportedRates = intVal;

		apmib_get(MIB_WLAN_RATE_ADAPTIVE_ENABLED, (void *)&intVal);
		if (intVal == 0) {
			pmib->dot11StationConfigEntry.autoRate = 0;
			apmib_get(MIB_WLAN_FIX_RATE, (void *)&intVal);
			pmib->dot11StationConfigEntry.fixedTxRate = intVal;
		}
		else
			pmib->dot11StationConfigEntry.autoRate = 1;

		apmib_get(MIB_WLAN_HIDDEN_SSID, (void *)&intVal);
		pmib->dot11OperationEntry.hiddenAP = intVal;

	// set band
		apmib_get(MIB_WLAN_BAND, (void *)&intVal);
		wlan_band = intVal;
		if ((mode != 1) && (pmib->dot11OperationEntry.wifi_specific == 1) && (wlan_band == 2))
			wlan_band = 3;

		if (wlan_band == 8) { // pure-11n
			wlan_band += 3; // b+g+n
			pmib->dot11StationConfigEntry.legacySTADeny = 3;
		}
		else if (wlan_band == 2) { // pure-11g
			wlan_band += 1; // b+g
			pmib->dot11StationConfigEntry.legacySTADeny = 1;
		}
		else if (wlan_band == 10) { // g+n
			wlan_band += 1; // b+g+n
			pmib->dot11StationConfigEntry.legacySTADeny = 1;
		}
		else
			pmib->dot11StationConfigEntry.legacySTADeny = 0;	

		pmib->dot11BssType.net_work_type = wlan_band;

		// set guest access
		//apmib_get(MIB_WLAN_ACCESS, (void *)&intVal);
		//pmib->dot11OperationEntry.guest_access = intVal;
		pmib->dot11OperationEntry.guest_access = 0;
		
		
		// set WMM
		//apmib_get(MIB_WLAN_WMM_ENABLED, (void *)&intVal);
		apmib_get(MIB_WLAN_WMM, (void *)&intVal);
		
		pmib->dot11QosEntry.dot11QosEnable = intVal;		
	}

	apmib_get(MIB_WLAN_AUTH_TYPE, (void *)&intVal);
	apmib_get(MIB_WLAN_ENCRYPT, (void *)&encrypt);
#ifdef CONFIG_RTL_WAPI_SUPPORT
	//wapi is independed. disable WAPI first if not WAPI
	if(7 !=encrypt)
	{
		pmib->wapiInfo.wapiType=0;	
	}
#endif
	if ((intVal == 1) && (encrypt != 1)) {
		// shared-key and not WEP enabled, force to open-system
		intVal = 0;
	}
	pmib->dot1180211AuthEntry.dot11AuthAlgrthm = intVal;

	if (encrypt == 0)
		pmib->dot1180211AuthEntry.dot11PrivacyAlgrthm = 0;
	else if (encrypt == 1) {
		// WEP mode
		apmib_get(MIB_WLAN_WEP, (void *)&wep);
		if (wep == 1) {
			pmib->dot1180211AuthEntry.dot11PrivacyAlgrthm = 1;
			apmib_get(MIB_WLAN_WEP64_KEY1, (void *)buf1);
			memcpy(&(pmib->dot11DefaultKeysTable.keytype[0]), buf1, 5);
			apmib_get(MIB_WLAN_WEP64_KEY2, (void *)buf1);
			memcpy(&(pmib->dot11DefaultKeysTable.keytype[1]), buf1, 5);
			apmib_get(MIB_WLAN_WEP64_KEY3, (void *)buf1);
			memcpy(&(pmib->dot11DefaultKeysTable.keytype[2]), buf1, 5);
			apmib_get(MIB_WLAN_WEP64_KEY4, (void *)buf1);
			memcpy(&(pmib->dot11DefaultKeysTable.keytype[3]), buf1, 5);
			apmib_get(MIB_WLAN_WEP_DEFAULT_KEY, (void *)&intVal);
			pmib->dot1180211AuthEntry.dot11PrivacyKeyIndex = intVal;
		}
		else {
			pmib->dot1180211AuthEntry.dot11PrivacyAlgrthm = 5;
			apmib_get(MIB_WLAN_WEP128_KEY1, (void *)buf1);
			memcpy(&(pmib->dot11DefaultKeysTable.keytype[0]), buf1, 13);
			apmib_get(MIB_WLAN_WEP128_KEY2, (void *)buf1);
			memcpy(&(pmib->dot11DefaultKeysTable.keytype[1]), buf1, 13);
			apmib_get(MIB_WLAN_WEP128_KEY3, (void *)buf1);
			memcpy(&(pmib->dot11DefaultKeysTable.keytype[2]), buf1, 13);
			apmib_get(MIB_WLAN_WEP128_KEY4, (void *)buf1);
			memcpy(&(pmib->dot11DefaultKeysTable.keytype[3]), buf1, 13);
			apmib_get(MIB_WLAN_WEP_DEFAULT_KEY, (void *)&intVal);
			pmib->dot1180211AuthEntry.dot11PrivacyKeyIndex = intVal;
		}
	}
#ifdef CONFIG_RTL_WAPI_SUPPORT	
	else if(7 == encrypt)
	{
		pmib->dot1180211AuthEntry.dot11PrivacyAlgrthm = 0;
		pmib->dot1180211AuthEntry.dot11AuthAlgrthm = 0;
	}
#endif	
	else {
		// WPA mode
		pmib->dot1180211AuthEntry.dot11PrivacyAlgrthm = 2;
	}

	// Set 802.1x flag
	enable1x = 0;
	if (encrypt < 2) {
		apmib_get(MIB_WLAN_ENABLE_1X, (void *)&intVal);
		apmib_get(MIB_WLAN_ENABLE_MAC_AUTH, (void *)&intVal2);
		if ((intVal != 0) || (intVal2 != 0))
			enable1x = 1;
	}
#ifdef CONFIG_RTL_WAPI_SUPPORT
	else if(encrypt == 7)
	{
		//wapi
		enable1x = 0;
	}
#endif	
	else
		enable1x = 1;
	pmib->dot118021xAuthEntry.dot118021xAlgrthm = enable1x;

#ifdef CONFIG_RTL_WAPI_SUPPORT
	if(7 == encrypt)
	{
		//apmib_get(MIB_WLAN_WAPI_ASIPADDR,);
		apmib_get(MIB_WLAN_WAPI_AUTH,(void *)&intVal);
		pmib->wapiInfo.wapiType=intVal;

		apmib_get(MIB_WLAN_WAPI_MCAST_PACKETS,(void *)&intVal);
		pmib->wapiInfo.wapiUpdateMCastKeyPktNum=intVal;
		
		apmib_get(MIB_WLAN_WAPI_MCAST_REKEYTYPE,(void *)&intVal);
		pmib->wapiInfo.wapiUpdateMCastKeyType=intVal;

		apmib_get(MIB_WLAN_WAPI_MCAST_TIME,(void *)&intVal);
		pmib->wapiInfo.wapiUpdateMCastKeyTimeout=intVal;

		apmib_get(MIB_WLAN_WAPI_UCAST_PACKETS,(void *)&intVal);
		pmib->wapiInfo.wapiUpdateUCastKeyPktNum=intVal;
		
		apmib_get(MIB_WLAN_WAPI_UCAST_REKETTYPE,(void *)&intVal);
		pmib->wapiInfo.wapiUpdateUCastKeyType=intVal;

		apmib_get(MIB_WLAN_WAPI_UCAST_TIME,(void *)&intVal);
		pmib->wapiInfo.wapiUpdateUCastKeyTimeout=intVal;

		//1: hex  -else passthru
		apmib_get(MIB_WLAN_WAPI_PSK_FORMAT,(void *)&intVal2);
		apmib_get(MIB_WLAN_WAPI_PSKLEN,(void *)&intVal);
		apmib_get(MIB_WLAN_WAPI_PSK,(void *)buf1);
		pmib->wapiInfo.wapiPsk.len=intVal;
		if(1 == intVal2 )
		{
			//hex
			string_to_hex(buf1, buf2, pmib->wapiInfo.wapiPsk.len*2);
		}else
		{
			//passthru
			strcpy(buf2,buf1);
		}
		memcpy(pmib->wapiInfo.wapiPsk.octet,buf2,pmib->wapiInfo.wapiPsk.len);
	}
#endif

#ifdef CONFIG_RTK_MESH

#ifdef CONFIG_NEW_MESH_UI
	//new feature:Mesh enable/disable
	//brian add new key:MIB_WLAN_MESH_ENABLE
	apmib_get(MIB_WLAN_MESH_ENABLE,(void *)&intVal);
	if (mode == AP_MESH_MODE || mode == MESH_MODE)
	{
		if( intVal )
			pmib->dot1180211sInfo.mesh_enable = 1;
		else
			pmib->dot1180211sInfo.mesh_enable = 0;
	}
	else
		pmib->dot1180211sInfo.mesh_enable = 0;
	// set mesh argument
	// brian change to shutdown portal/root as default
	if (mode == AP_MESH_MODE)
	{
		pmib->dot1180211sInfo.mesh_ap_enable = 1;
		pmib->dot1180211sInfo.mesh_portal_enable = 0;
	}
	else if (mode == MESH_MODE)
	{
		pmib->dot1180211sInfo.mesh_ap_enable = 0;
		pmib->dot1180211sInfo.mesh_portal_enable = 0;
	}
	else
	{
		pmib->dot1180211sInfo.mesh_ap_enable = 0;
		pmib->dot1180211sInfo.mesh_portal_enable = 0;	
	}
	#if 0	//by brian, dont enable root by default
	apmib_get(MIB_MESH_ROOT_ENABLE, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_root_enable = intVal;
	#else
	pmib->dot1180211sInfo.mesh_root_enable = 0;
	#endif
#else
	if (mode == AP_MPP_MODE)
	{
		pmib->dot1180211sInfo.mesh_enable = 1;
		pmib->dot1180211sInfo.mesh_ap_enable = 1;
		pmib->dot1180211sInfo.mesh_portal_enable = 1;	
	}
	else if (mode == MPP_MODE)
	{
		pmib->dot1180211sInfo.mesh_enable = 1;
		pmib->dot1180211sInfo.mesh_ap_enable = 0;
		pmib->dot1180211sInfo.mesh_portal_enable = 1;
	}
	else if (mode == MAP_MODE)
	{
		pmib->dot1180211sInfo.mesh_enable = 1;
		pmib->dot1180211sInfo.mesh_ap_enable = 1;
		pmib->dot1180211sInfo.mesh_portal_enable = 0;
	}		
	else if (mode == MP_MODE)
	{
		pmib->dot1180211sInfo.mesh_enable = 1;
		pmib->dot1180211sInfo.mesh_ap_enable = 0;
		pmib->dot1180211sInfo.mesh_portal_enable = 0;		
	}
	else
	{
		pmib->dot1180211sInfo.mesh_enable = 0;
		pmib->dot1180211sInfo.mesh_ap_enable = 0;
		pmib->dot1180211sInfo.mesh_portal_enable = 0;	
	}

	apmib_get(MIB_MESH_ROOT_ENABLE, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_root_enable = intVal;
#endif
	apmib_get(MIB_MESH_MAX_NEIGHTBOR, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_max_neightbor = intVal;

	apmib_get(MIB_LOG_ENABLED, (void *)&intVal);
	pmib->dot1180211sInfo.log_enabled = intVal;

	apmib_get(MIB_MESH_ID, (void *)buf1);
	intVal2 = strlen(buf1);
	memset(pmib->dot1180211sInfo.mesh_id, 0, 32);
	memcpy(pmib->dot1180211sInfo.mesh_id, buf1, intVal2);

	apmib_get(MIB_MESH_ENCRYPT, (void *)&intVal);
	apmib_get(MIB_MESH_WPA_AUTH, (void *)&intVal2);

	if( intVal2 == 2 && intVal)
		pmib->dot11sKeysTable.dot11Privacy  = 2;
	else
		pmib->dot11sKeysTable.dot11Privacy  = 0;
	
#ifdef 	_11s_TEST_MODE_	

	apmib_get(MIB_MESH_TEST_PARAM1, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reserved1 = intVal;

	apmib_get(MIB_MESH_TEST_PARAM2, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reserved2 = intVal;

	apmib_get(MIB_MESH_TEST_PARAM3, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reserved3 = intVal;

	apmib_get(MIB_MESH_TEST_PARAM4, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reserved4 = intVal;

	apmib_get(MIB_MESH_TEST_PARAM5, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reserved5 = intVal;

	apmib_get(MIB_MESH_TEST_PARAM6, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reserved6 = intVal;

	apmib_get(MIB_MESH_TEST_PARAM7, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reserved7 = intVal;

	apmib_get(MIB_MESH_TEST_PARAM8, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reserved8 = intVal;
	
	apmib_get(MIB_MESH_TEST_PARAM9, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reserved9 = intVal;

	apmib_get(MIB_MESH_TEST_PARAMA, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reserveda = intVal;

	apmib_get(MIB_MESH_TEST_PARAMB, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reservedb = intVal;
	
	apmib_get(MIB_MESH_TEST_PARAMC, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reservedc = intVal;

	apmib_get(MIB_MESH_TEST_PARAMD, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reservedd = intVal;

	apmib_get(MIB_MESH_TEST_PARAME, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reservede = intVal;

	apmib_get(MIB_MESH_TEST_PARAMF, (void *)&intVal);
	pmib->dot1180211sInfo.mesh_reservedf = intVal;
	
	apmib_get(MIB_MESH_TEST_PARAMSTR1, (void *)buf1);
	intVal2 = strlen(buf1)<15 ? strlen(buf1) : 15;
	memset(pmib->dot1180211sInfo.mesh_reservedstr1, 0, 16);
	memcpy(pmib->dot1180211sInfo.mesh_reservedstr1, buf1, intVal2);
	
#endif
	
#endif // CONFIG_RTK_MESH

	// When using driver base WPA, set wpa setting to driver
#if 1
	int intVal3;
	apmib_get(MIB_WLAN_WPA_AUTH, (void *)&intVal3);
//#ifdef CONFIG_RTL8196B
// button 2009.05.21
#if 1
	if ((intVal3 & WPA_AUTH_PSK) && encrypt >= 2 
#ifdef CONFIG_RTL_WAPI_SUPPORT
&& encrypt < 7
#endif
)
#else
	if (mode != 1 && (intVal3 & WPA_AUTH_PSK) && encrypt >= 2 
#ifdef CONFIG_RTL_WAPI_SUPPORT
&& encrypt < 7
#endif
)
#endif
	{
		if (encrypt == 2)
			intVal = 1;
		else if (encrypt == 4)
			intVal = 2;
		else if (encrypt == 6)
			intVal = 3;
		else {
			printf("invalid ENCRYPT value!\n");
			return -1;
		}
		pmib->dot1180211AuthEntry.dot11EnablePSK = intVal;

		if (encrypt == 2 || encrypt == 6) {
			apmib_get(MIB_WLAN_WPA_CIPHER_SUITE, (void *)&intVal2);
			if (intVal2 == 1)
				intVal = 2;
			else if (intVal2 == 2)
				intVal = 8;
			else if (intVal2 == 3)
				intVal = 10;
			else {
				printf("invalid WPA_CIPHER_SUITE value!\n");
				return -1;
			}
			pmib->dot1180211AuthEntry.dot11WPACipher = intVal;			
		}
		
		if (encrypt == 4 || encrypt == 6) {
			apmib_get(MIB_WLAN_WPA2_CIPHER_SUITE, (void *)&intVal2);
			if (intVal2 == 1)
				intVal = 2;
			else if (intVal2 == 2)
				intVal = 8;
			else if (intVal2 == 3)
				intVal = 10;
			else {
				printf("invalid WPA2_CIPHER_SUITE value!\n");
				return -1;
			}
			pmib->dot1180211AuthEntry.dot11WPA2Cipher = intVal;			
		}

		apmib_get(MIB_WLAN_WPA_PSK, (void *)buf1);
		strcpy(pmib->dot1180211AuthEntry.dot11PassPhrase, buf1);

		apmib_get(MIB_WLAN_WPA_GROUP_REKEY_TIME, (void *)&intVal);
		pmib->dot1180211AuthEntry.dot11GKRekeyTime = intVal;			
	}
	else		
		pmib->dot1180211AuthEntry.dot11EnablePSK = 0;
#endif

	wrq.u.data.pointer = (caddr_t)pmib;
	wrq.u.data.length = sizeof(struct wifi_mib);
	if (ioctl(skfd, 0x8B43, &wrq) < 0) {
		printf("Set WLAN MIB failed!\n");
		return -1;
	}
	close(skfd);



	free(pmib);
	return 0;
}
*/

